/* ------------------------------------------------------------------------------
  File: CHR_GPS.c
  Author: CH Robotics
  Version: 1.0
  
  Description:
  
  Functions for processing and storing GPS data.
  
------------------------------------------------------------------------------ */ 
#include <string.h>
#include "stm32f10x.h"
#include "CHR_GPS.h"

// Global data for parsing functions
char g_GPS_packet[MAX_GPS_PACKET_SIZE];

uint8_t g_GPS_clock_synch;

/*******************************************************************************
* Function Name  : parseGPSPacket()
* Input          : 
* Output         : 
* Return         : 1 if successful, 0 otherwise
* Description    : 

Parses the GPS packet found in the array 'packet'
This function should be called when a new packet has been received

*******************************************************************************/
int parseGPSPacket( char* packet, GPS_Struct* GPS_data )
{
	int returnval;
	char token[MAX_GPS_PACKET_SIZE];

	// Determine what kind of packet this is
	returnval = getNextGPSToken( packet, token, 1 );
	
	if( returnval != 1 )
	{
		return 0;
	}

	if( strcmp( token, "$GPGGA" ) == 0 )
	{
		return parseGGA( packet, GPS_data );
	}
	else if( strcmp( token, "$GPGLL" ) == 0 )
	{
		return parseGLL( packet, GPS_data );
	}
	else if( strcmp( token, "$GPGSA" ) == 0 )
	{
		return parseGSA( packet, GPS_data );
	}
	else if( strcmp( token, "$GPGSV" ) == 0 )
	{
		return parseGSV( packet, GPS_data );
	}
	else if( strcmp( token, "$GPRMC" ) == 0 )
	{
		return parseRMC( packet, GPS_data );
	}
	else if( strcmp( token, "$GPVTG" ) == 0 )
	{
		return parseVTG( packet, GPS_data );
	}
	else
	{
		return 0;
	}
}

/*******************************************************************************
* Function Name  : parseGGA( char* token )
* Input          : 
* Output         : 
* Return         : 1 if successful, 0 otherwise
* Description    : 

Parses data in the packet[] buffer as a GGA NMEA message.
This function is designed to be called from within parseGPSPacket().  The
first token in the packet[] buffer should have already been retrieved to
identify the packet type.

FIELDS IN GGA PACKET
1. UTC Time				(hhmmss.sss)
2. Latitude  			(ddmm.mmmm)
3. N/S indicator 		{N,S}
4. Longitude			(dddmm.mmmm)
5. E/W indicator		{E,W}
6. GPS Quality			{0 = position fix unavailable,
						1 = valid position fix, SPS mode,
						2 = valid position fix, differential GPS mode,
						3 = GPS PPS Mode, fix valid
						4 = Real-time Kinematic.  System used in RTK mode.  Floating integers
						5 = Float RTK.  Satellite system used in RTK mode.  Floating integers
						6 = Estimated (dead reckoning) Mode
						7 = Maual Input Mode
						8 = Simulator Mode}
7. Satellites Used		Number of satellites used in solution (00 ~ 12)
8. HDOP					Horizontal Dilution of Precision (00.0 ~ 99.9)
9. Altitude				Altitude above mean sea level (-9999.9 ~ 17999.9)
10 - 13. 				NULL or unsused
14. DGPS Station ID		(0000 ~ 1023, NULL when not used)
15. Checksum

*******************************************************************************/
int parseGGA( char* packet, GPS_Struct* GPS_data )
{
	int returnval;
	char token[MAX_GPS_PACKET_SIZE];
	
	// 1. Get UTC Time
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}

	GPS_data->UTC_hour = ConvertAtoI_n( token, 2 );
	GPS_data->UTC_minute = ConvertAtoI_n( &token[2], 2 );
	GPS_data->UTC_second = ConvertAtoI_n( &token[4], 2 );
	GPS_data->UTC_fractional_second = ConvertAtoI_n( &token[7], 3 );
	GPS_data->UTC_time = 3600.0f*(float)GPS_data->UTC_hour + 60.0f*(float)GPS_data->UTC_minute + (float)GPS_data->UTC_second + (float)GPS_data->UTC_fractional_second/1000.0f;

	// 2. Get latitude
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}

	// Latitude is in the format ddmm.mmmm where d is degree and m is arcminutes.  Need to convert to fractional
	// degrees alone.
	GPS_data->latitude = (float)ConvertAtoI_n( token, 2 ) + (float)ConvertAtoF(&token[2])/60.0f;

	// 3. Get N/S indicator and add to latitude (S = negative, N = positive)
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}
	
	if( token[0] == 'N' )
	{
		GPS_data->NS_indicator = token[0];
		
	}
	else if( token[0] == 'S' )
	{
		GPS_data->NS_indicator = token[0];
		GPS_data->latitude = -GPS_data->latitude;
	}
	else
	{
		return 0;
	}

	// 4. Get longitude
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}

	// Longitude is in the format dddmm.mmmm, where d = degrees and m = arcminutes.  We want fractional degrees only.
	GPS_data->longitude = (float)ConvertAtoI_n( token, 3 ) + (float)ConvertAtoF(&token[3])/60.0f;

	// 5. Get E/W indicator and add to longitude (W = negative, E = positive)
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}
	
	if( token[0] == 'E' )
	{
		GPS_data->EW_indicator = token[0];
		
	}
	else if( token[0] == 'W' )
	{
		GPS_data->EW_indicator = token[0];
		GPS_data->longitude = -GPS_data->longitude;
	}
	else
	{
		return 0;
	}

	// 6. Get GPS quality indicator
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}

	GPS_data->quality_indicator = (char)ConvertAtoI( token );

	// 7. Get number of satellites used
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}

	GPS_data->satellites_used = (char)ConvertAtoI( token );

	// 8. Get HDOP
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}

	GPS_data->Hdop = (float)ConvertAtoF( token );

	// 9. Get Altitude
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}

	GPS_data->altitude = (float)ConvertAtoF( token );

	// 10. - 13. iterate through unused fields
	getNextGPSToken( packet, token, 0 );
	getNextGPSToken( packet, token, 0 );
	getNextGPSToken( packet, token, 0 );
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}

	// 14. Get the DGPS station ID
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}

	if( strlen( token ) > 0 )
	{
		GPS_data->DGPS_station_ID = (int16_t)ConvertAtoI( token );
	}
	else
	{
		GPS_data->DGPS_station_ID = 0;
	}

	// IGNORE CHECKSUM
	// WE MAY WANT TO CHANGE THIS EVENTUALLY TO AVOID ERRORS

	// Set flag indicating that a GGA packet was received and return
	GPS_data->GGA_packet_received = 1;

	return 1;
}

/*******************************************************************************
* Function Name  : parseGLL( char* token )
* Input          : 
* Output         : 
* Return         : 1 if successful, 0 otherwise
* Description    : 

Parses data in the packet[] buffer as a GLL NMEA message.
This function is designed to be called from within parseGPSPacket().  The
first token in the packet[] buffer should have already been retrieved to
identify the packet type.

*******************************************************************************/
int parseGLL( char* packet, GPS_Struct* GPS_data )
{
	int returnval;
	char token[MAX_GPS_PACKET_SIZE];
	
	// 1. Get latitude
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}

	// Latitude is in the format ddmm.mmmm where d is degree and m is arcminutes.  Need to convert to fractional
	// degrees alone.
	GPS_data->latitude = (float)ConvertAtoI_n( token, 2 ) + (float)ConvertAtoF(&token[2])/60.0f;

	// 3. Get N/S indicator and add to latitude (S = negative, N = positive)
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}
	
	if( token[0] == 'N' )
	{
		GPS_data->NS_indicator = token[0];
		
	}
	else if( token[0] == 'S' )
	{
		GPS_data->NS_indicator = token[0];
		GPS_data->latitude = -GPS_data->latitude;
	}
	else
	{
		return 0;
	}

	// Get longitude
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}

	// Longitude is in the format dddmm.mmmm, where d = degrees and m = arcminutes.  We want fractional degrees only.
	GPS_data->longitude = (float)ConvertAtoI_n( token, 3 ) + (float)ConvertAtoF(&token[3])/60.0f;

	// 5. Get E/W indicator and add to longitude (W = negative, E = positive)
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}
	
	if( token[0] == 'E' )
	{
		GPS_data->EW_indicator = token[0];
		
	}
	else if( token[0] == 'W' )
	{
		GPS_data->EW_indicator = token[0];
		GPS_data->longitude = -GPS_data->longitude;
	}
	else
	{
		return 0;
	}

	// Get UTC time
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}

	GPS_data->UTC_hour = ConvertAtoI_n( token, 2 );
	GPS_data->UTC_minute = ConvertAtoI_n( &token[2], 2 );
	GPS_data->UTC_second = ConvertAtoI_n( &token[4], 2 );
	GPS_data->UTC_fractional_second = ConvertAtoI_n( &token[7], 3 );
	GPS_data->UTC_time = 3600.0f*(float)GPS_data->UTC_hour + 60.0f*(float)GPS_data->UTC_minute + (float)GPS_data->UTC_second + (float)GPS_data->UTC_fractional_second/1000.0f;

	// Get data active indicator
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}

	if( token[0] == 'A' )
	{
		GPS_data->GLL_packet_received = 1;
	}
	
	return 1;
}

/*******************************************************************************
* Function Name  : parseGSA( char* token )
* Input          : 
* Output         : 
* Return         : 1 if successful, 0 otherwise
* Description    : 

Parses data in the packet[] buffer as a GSA NMEA message.
This function is designed to be called from within parseGPSPacket().  The
first token in the packet[] buffer should have already been retrieved to
identify the packet type.

*******************************************************************************/
int parseGSA( char* packet, GPS_Struct* GPS_data )
{
	int index;
	int returnval;
	char token[MAX_GPS_PACKET_SIZE];

	// Get item indicating auto/manual selection of 2D or 3D fix.
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}

	GPS_data->fix_type = token[0];

	// Get the GPS mode (no fix, 2D fix, 3D fix)
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}

	GPS_data->mode = ConvertAtoI( token );

	// Get the PRNs of all satellites used for the fix
	for( index = 0; index < 12; index++ )
	{
		returnval = getNextGPSToken( packet, token, 0 );
	
		if( returnval != 1 )
		{
			return 0;
		}
	
		// For now, we aren't using this data.  Just iterate through the fields and discard.
	}

	// Get PDOP
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}

	GPS_data->Pdop = ConvertAtoF( token );

	// Get HDOP
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}

	GPS_data->Hdop = ConvertAtoF( token );

	// Get VDOP
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}

	GPS_data->Vdop = ConvertAtoF( token );

	GPS_data->GSA_packet_received = 1;

	return 1;
}

/*******************************************************************************
* Function Name  : parseGSV( char* token )
* Input          : 
* Output         : 
* Return         : 1 if successful, 0 otherwise
* Description    : 

Parses data in the packet[] buffer as a GSV NMEA message.
This function is designed to be called from within parseGPSPacket().  The
first token in the packet[] buffer should have already been retrieved to
identify the packet type.

*******************************************************************************/
int parseGSV( char* packet, GPS_Struct* GPS_data )
{
	int returnval;
	char token[MAX_GPS_PACKET_SIZE];

	// Get number of sentences for full data
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}

	int16_t total_sentences = ConvertAtoI( token );

	if( total_sentences == 0 )
	{
		return 0;
	}

	// Get which sentence this happens to be
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}

	int16_t sentence_number = ConvertAtoI( token );

	if( sentence_number == 0 )
	{
		return 0;
	}

	// Get the number of satellites in view
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}

	GPS_data->satellites_in_view = ConvertAtoI( token );

	// Now get data for each satellite in the packet
	int sat_count = 0;
	while( sat_count < 4 )		// Maximum of four satellites per sentence
	{
		if( getNextGPSToken( packet, token, 0 ) != 1 )
		{
			break;
		}
		int16_t PRN_number = ConvertAtoI( token );
	
		if( getNextGPSToken( packet, token, 0 ) != 1 )
		{
			break;
		}
		int16_t elevation = ConvertAtoI( token );
		
		if( getNextGPSToken( packet, token, 0 ) != 1 )
		{
			break;
		}
		int16_t azimuth = ConvertAtoI( token );
		
		if( getNextGPSToken( packet, token, 0 ) != 1 )
		{
			break;
		}
		int16_t SNR = ConvertAtoI( token );
	
		// If we got here, then we (presumably) have data on a new satellite.
		// Add the satellite to the data array
		int16_t satellite_index = 4*(sentence_number-1)+sat_count;
		GPS_data->satellite_data[satellite_index][0] = PRN_number;
		GPS_data->satellite_data[satellite_index][1] = elevation;
		GPS_data->satellite_data[satellite_index][2] = azimuth;
		GPS_data->satellite_data[satellite_index][3] = SNR;
	
		// Increment satellite count
		sat_count++;
	}

	// If we've received all sentences, indicate as much
	if( sentence_number == total_sentences )
	{
		GPS_data->GSV_packet_received = 1;
	}

	return 1;
}

/*******************************************************************************
* Function Name  : parseRMC( char* token )
* Input          : 
* Output         : 
* Return         : 1 if successful, 0 otherwise
* Description    : 

Parses data in the packet[] buffer as a RMC NMEA message.
This function is designed to be called from within parseGPSPacket().  The
first token in the packet[] buffer should have already been retrieved to
identify the packet type.

*******************************************************************************/
int parseRMC( char* packet, GPS_Struct* GPS_data )
{
	int returnval;
	char token[MAX_GPS_PACKET_SIZE];

	// Get UTC time
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}

	GPS_data->UTC_hour = ConvertAtoI_n( token, 2 );
	GPS_data->UTC_minute = ConvertAtoI_n( &token[2], 2 );
	GPS_data->UTC_second = ConvertAtoI_n( &token[4], 2 );
	GPS_data->UTC_fractional_second = ConvertAtoI_n( &token[7], 3 );
	GPS_data->UTC_time = 3600.0f*(float)GPS_data->UTC_hour + 60.0f*(float)GPS_data->UTC_minute + (float)GPS_data->UTC_second + (float)GPS_data->UTC_fractional_second/1000.0f;

	// Get status (active or void)
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}

	if( token[0] != 'A' )
	{
		return 0;
	}

	// Get Latitude
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}

	// Latitude is in the format ddmm.mmmm where d is degree and m is arcminutes.  Need to convert to fractional
	// degrees alone.
	GPS_data->latitude = (float)ConvertAtoI_n( token, 2 ) + (float)ConvertAtoF(&token[2])/60.0f;

	// 3. Get N/S indicator and add to latitude (S = negative, N = positive)
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}
	
	if( token[0] == 'N' )
	{
		GPS_data->NS_indicator = token[0];
		
	}
	else if( token[0] == 'S' )
	{
		GPS_data->NS_indicator = token[0];
		GPS_data->latitude = -GPS_data->latitude;
	}
	else
	{
		return 0;
	}

	// Get longitude
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}

	// Longitude is in the format dddmm.mmmm, where d = degrees and m = arcminutes.  We want fractional degrees only.
	GPS_data->longitude = (float)ConvertAtoI_n( token, 3 ) + (float)ConvertAtoF(&token[3])/60.0f;

	// 5. Get E/W indicator and add to longitude (W = negative, E = positive)
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}
	
	if( token[0] == 'E' )
	{
		GPS_data->EW_indicator = token[0];
		
	}
	else if( token[0] == 'W' )
	{
		GPS_data->EW_indicator = token[0];
		GPS_data->longitude = -GPS_data->longitude;
	}
	else
	{
		return 0;
	}

	// Get ground speed in knots
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}

	// convert to meters per second
	GPS_data->speed = ConvertAtoF( token )*0.5144444f;

	// Get course in degrees
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}

	GPS_data->course = ConvertAtoF( token );

	// Get the UTC date
	returnval = getNextGPSToken( packet, token, 0 );
	
	if( returnval != 1 )
	{
		return 0;
	}
	
	GPS_data->UTC_day = ConvertAtoI_n( token, 2 );
	GPS_data->UTC_month = ConvertAtoI_n( &token[2], 2 );
	GPS_data->UTC_year = ConvertAtoI_n( &token[4], 2 );
	
	// Don't need final mode indicator.  Just return.
	GPS_data->RMC_packet_received = 1;

	return 1;
}

/*******************************************************************************
* Function Name  : parseVTG( char* token )
* Input          : 
* Output         : 
* Return         : 1 if successful, 0 otherwise
* Description    : 

Parses data in the packet[] buffer as a VTG NMEA message.
This function is designed to be called from within parseGPSPacket().  The
first token in the packet[] buffer should have already been retrieved to
identify the packet type.

*******************************************************************************/
int parseVTG( char* packet, GPS_Struct* GPS_data )
{
	int returnval;
	char token[MAX_GPS_PACKET_SIZE];

	// Get True track in degrees
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}

	GPS_data->course = ConvertAtoF( token );

	// Get true track indicator (we don't do anything with this)
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}

	// Get magnetic course (we don't do anything with this)
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}

	// Get magnetic track indicator (we do nothing with this)
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}

	// Get nautical ground speed (ignored...)
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}

	// Get knotical indicator (ignored...)
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}

	// Get ground speed in kph
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}

	// Convert to m/s and write it out.
	GPS_data->speed = ConvertAtoF(token)*0.2777778f;
	
	// Get kph indicator
	returnval = getNextGPSToken( packet, token, 0 );

	if( returnval != 1 )
	{
		return 0;
	}

	GPS_data->VTG_packet_received = 1;

	return 1;
}

/*******************************************************************************
* Function Name  : getNextGPSToken( char* token, int reset )
* Input          : 
* Output         : 
* Return         : 1 if successful, 0 otherwise
* Description    : 

This function is used to split a received NMEA string into comma-seperated tokens.
If reset != 0, then the tokenizer starts at the beginning and returns the first token.
This function keeps track of what tokens have been extracted, so that if reset == 0,
the next token is returned.

Tokens are seperated by commas, * characters, and <CR><LF> pairs.

Once all tokens have been returned by the function, all subsequent calls to
the function will return 0 until a call has been made with reset == 1

*******************************************************************************/
int getNextGPSToken( char* packet, char* token, int reset )
{
	static int token_start = 0;
	static int no_more_tokens = 0;

	int token_end;

	// If reset flag is set, start tokenizer from beginning
	if( reset )
	{
		token_start = 0;
		no_more_tokens = 0;
	}

	// If all tokens have been read, return 0
	if( no_more_tokens )
	{
		return 0;
	}

	// The token starts at the current index (the index is set so that it immediately follows the last comma)
	// Find the token stop location
	
	for( token_end = token_start; token_end < MAX_GPS_PACKET_SIZE; token_end++ )
	{
		// If this is end of the token, stop looping immediately (token_end will be used for parsing)
		if( (packet[token_end] == ',') || (packet[token_end] == '*') )
		{
			break;
		}
		
		// Check for final termination sequence
		if( token_end > 0 )
		{
			if( (packet[token_end] == '\n') && (packet[token_end-1] == '\r') )
			{
				token_end--;
				no_more_tokens = 1;
				break;
			}
		}
	}

	// If token_end is larger than the buffer size, it means that the GPS packet was not terminated correctly (no <CR><LF> pair)
	if( token_end == MAX_GPS_PACKET_SIZE )
	{
		return 0;
	}

	// Copy token to output
	int output_index = 0;
	int i;
	for( i = token_start; i < token_end; i++ )
	{
		token[output_index] = packet[i];
		output_index++;
	}

	// Add null terminating character
	token[output_index] = '\0';

	// Now set the start_index to the first character of the next token
	// token_end has the index of the most recent terminating character, so go 1 index beyond that
	token_start = token_end + 1;

	return 1;
}

/*******************************************************************************
* Function Name  : int ConvertAtoI( char* str )
* Input          : 
* Output         : 
* Return         : The integer equivalent of the ASCII number representation
* Description    : 

This function is used to split a received NMEA string into comma-seperated tokens.
If reset != 0, then the tokenizer starts at the beginning and returns the first token.
This function keeps track of what tokens have been extracted, so that if reset == 0,
the next token is returned.

Tokens are seperated by commas, * characters, and <CR><LF> pairs.

Once all tokens have been returned by the function, all subsequent calls to
the function will return 0 until a call has been made with reset == 1

*******************************************************************************/
int ConvertAtoI( char* str )
{
	int output = 0;
	int negative = 0;
	
	if( *str == '-' )
	{
		negative = 1;
		str++;
	}

	while( CHR_isdigit(*str) )
	{
		output = output*10 + (*str++ - '0');
	}

	if( negative )
	{
		return -output;
	}

	return output;
}

/*******************************************************************************
* Function Name  : int CHR_isdigit( char c )
* Input          : 
* Output         : 
* Return         : 1 if digit, 0 if not
* Description    : 

*******************************************************************************/
int CHR_isdigit( char c )
{
	if ((c>='0') && (c<='9')) 
	{
		return 1;
	}
			
    return 0;
}

/*******************************************************************************
* Function Name  : int ConvertAtoI_n( char* str, int n )
* Input          : 
* Output         : 
* Return         : The integer equivalent of the ASCII number representation
* Description    : 

*******************************************************************************/
int ConvertAtoI_n( char* str, int n )
{
	int index;
	int negative = 0;
	int output = 0;

	if( strlen(str) < n )
	{
		n = strlen(str);
	}

	if( *str == '-' )
	{
		negative = 1;
		str++;
	}

	for( index = 0; index < n; index++ )
	{
		output = output*10 + (*str++ - '0');
	}
	
	if( negative )
	{
		return -output;
	}

	return output;
}


/*******************************************************************************
* Function Name  : float ConvertAtoF( char* str )
* Input          : 
* Output         : 
* Return         : A float representing the value of the ASCII data
* Description    : 

This function is used as an alternative to the standard library function atof(.),
which is way to large to use in this application.

Converts the given string to a floating point number if possible.  If not possible,
returns 0.0

*******************************************************************************/
float ConvertAtoF( char* s )
{
	float a = 0.0;
	int e = 0;
	int c;

	int negative = 0;

	c = *s;

	if( c == '-' )
	{
		negative = 1;
		s++;
	}

	while ((c = *s++) != '\0' && CHR_isdigit(c)) {
			a = a*10.0 + (c - '0');
	}
	if (c == '.') {
			while ((c = *s++) != '\0' && CHR_isdigit(c)) {
					a = a*10.0 + (c - '0');
					e = e-1;
			}
	}
	if (c == 'e' || c == 'E') {
			int sign = 1;
			int i = 0;
			c = *s++;
			if (c == '+')
					c = *s++;
			else if (c == '-') {
					c = *s++;
					sign = -1;
			}
			while (CHR_isdigit(c)) {
					i = i*10 + (c - '0');
					c = *s++;
			}
			e += i*sign;
	}
	while (e > 0) {
			a *= 10.0;
			e--;
	}
	while (e < 0) {
			a *= 0.1;
			e++;
	}

	if( negative )
	{
		a = -a;
	}

	return a;
	
}


/*******************************************************************************
* Function Name  : uint32_t uint32Power( uint32_t x, uint32_t y )
* Input          : 
* Output         : 
* Return         : The value x^y
* Description    : 

This function is used instead of the math.h pow() function.  We only need powers
of unisgned integers, not of doubles.  This function does the job using much less
space.

*******************************************************************************/
uint32_t uint32Power( uint32_t x, uint32_t y )
{
	int index;
	uint32_t result;

	if( y == 0 )
	{
		return 1;
	}

	result = x;
	for( index = 0; index < (y-1); index++ )
	{
		result = result*x;
	}

	return result;
}