/* ------------------------------------------------------------------------------
  File: chr6dm_HMC.c
  Author: CH Robotics
  Version: 1.0
  
  Description: Functions for interacting with HMC5843 magnetic sensor
------------------------------------------------------------------------------ */ 

#include "stm32f10x.h"
#include "CHR_i2c.h"
#include "CHR_HMC.h"

uint8_t gHMC_Initialized = 0;

/*******************************************************************************
* Function Name  : initializeHMC
* Input          : None
* Output         : uint8_t* status_flag
* Return         : 1 if success, 0 if fail
* Description    : Sets the configuration registers of the HMC.  The registers
						 are set in an order that guarantees that the next register
						 accessed during a read will be 0x03 (i.e. getHMCData can be called
						 after initializeHMC is called.
						 
						 If the function succesfully initializes the HMC, then
						 the return value is 1.  If failed, then the function returns 0,
						 and status_flag is filled with one of the following error codes:
						 
						 HMC_REG0_FAILED => Couldn't set register 0
						 HMC_REG1_FAILED => Couldn't set register 1
						 HMC_REG2_FAILED => Couldn't set register 2
						 
*******************************************************************************/
int32_t initializeHMC( uint8_t* status_flag )
{	 
	 uint8_t returnval;
	 
	 // Set CRA register
	 g_i2cTxBuf[0] = HMC_REG_CRA;		// Register address
	 g_i2cTxBuf[1] = 0x98;				// 0b10011000 (Temperature sensor enabled, 75 Hz update rate)
	 
	 returnval = i2cBatchWrite( MAG_SLAVE_ADDRESS7, g_i2cTxBuf, 2 );
	 	 
	 // CRB register
	 g_i2cTxBuf[0] = HMC_REG_CRB;
	 g_i2cTxBuf[1] = 0x60;				// 0b01100000 (2.5 gauss dynamic range)
	 
	 returnval = i2cBatchWrite( MAG_SLAVE_ADDRESS7, g_i2cTxBuf, 2 );
	 
	 // Mode register
	 g_i2cTxBuf[0] = 0x02;
	 g_i2cTxBuf[1] = 0x00;				// 0b00000000 (Continuous conversion mode)

	 returnval = i2cBatchWrite( MAG_SLAVE_ADDRESS7, g_i2cTxBuf, 2 );
	 
	 gHMC_Initialized = 1;
	 
	 return 1;
}

/*******************************************************************************
* Function Name  : getHMCData
* Input          : None
* Output         : None
* Return         : 1 if success, 0 if fail
* Description    : Fills g_i2cRxBuf with 7 bytes of data - the first 6 bytes are
						 the magnetic readings from the x, z, and y axes (respectively)
						 of the HMC5843, while the last byte is the HMC status register.
						 This function call assumes that the internal register address
						 pointer of the HMC is set to register 0x03.
*******************************************************************************/
int32_t getHMCData( )
{
	 int i;
	 
	 i2cBatchRead( MAG_SLAVE_ADDRESS7, 0x03, g_i2cRxBuf, 6 );
	 
	 return 1;		  
}
