/*
 * CHR_matrix.c
 *
 * Created: 11/7/2012 11:44:42 AM
 *  Author: Caleb
 */ 

#include "CHR_matrix.h"

/*******************************************************************************
* Function Name  : mat_add( fMatrix* src1, fMatrix* src2, fMatrix* dest )
* Input          : fMatrix* src1, fMatrix* src2, fMatrix* dest
* Output         : fMatrix* dest
* Return         : 0 on success, 1 if failed
* Description    : 

Performs the operation
  dest = src1 + src2

src1 and src1 must have compatible dimensions for matrix addition.
src1, src2, and dest MUST have been preinitialized so that matrix->data points to
memory that has been allocated.
*******************************************************************************/
int mat_add( fMatrix* src1, fMatrix* src2, fMatrix* dest )
{
	int i, j;
	 
	// Make sure that rows and columns on all relevant matrices match
	if( (src1->rows != src2->rows) || (src1->columns != src2->columns) )
	{
		return 1;
	}
	 
	// Compute sum
	for( i=0; i < src1->rows; i++ )
	{
		int row_index = i*src1->columns;
		
		for( j = 0; j < src1->columns; j++ )
		{
			
			dest->data[row_index + j] = src1->data[row_index + j] + src2->data[row_index + j];
		}
	}
	
	dest->rows = src1->rows;
	dest->columns = src1->columns;
	 
	return 0;
}

/*******************************************************************************
* Function Name  : mat_mult( fMatrix* src1, fMatrix* src2, fMatrix* dest )
* Input          : fMatrix* src1, fMatrix* src2, fMatrix* dest
* Output         : fMatrix* dest
* Return         : 0 on success, 1 if failed
* Description    : 

Performs the operation
  dest = src1*src2

src1 and src1 must have compatible dimensions for matrix multiplication.
src1, src2, and dest MUST have been preinitialized so that matrix->data points to
memory that has been allocated.
*******************************************************************************/
int mat_mult( fMatrix* src1, fMatrix* src2, fMatrix* dest )
{
	int i,j,k;
	int src2_cols = src2->columns;
	int src2_colsx2 = src2_cols*2;
	int src1_cols = src1->columns;
	int src1_cols_half = src1_cols / 2; // For unrolling the inner loop: C division will truncate the 1/2 in case of an odd number of columns
	int odd = (src1_cols == src1_cols_half*2) ? 0 : 1;
	float dot_product;
	
	fMatrix result;
	
	float matrix_data[MAX_MATRIX_ROWS*MAX_MATRIX_COLS];

	// Make sure dimensions are consistent with matrix multiplication rules
	if( src1->columns != src2->rows )
	{
		return 1;
	}
	
	result.data = matrix_data;
	result.rows = src1->rows;
	result.columns = src2->columns;
	
	// Compute matrix product
	int src1_row_idx = 0;
	int dest_row_idx = 0;
	int src2_idx = 0;
	for (i = 0; i < src1->rows; i++)
	{
		for (j = 0; j < src2_cols; j++)
		{
			src2_idx = j;
			dot_product = 0.0f;
			for (k = 0; k < src1_cols-1; k+=2)
			{
				dot_product += src1->data[src1_row_idx+k] * src2->data[src2_idx];
				dot_product += src1->data[src1_row_idx+k+1] * src2->data[src2_idx+src2_cols];
				src2_idx += src2_colsx2;
			}
			// clean up unrolling: handle odd # of cols:
			if (odd)
			{
				dot_product += src1->data[src1_row_idx+k] * src2->data[src2_idx];
			}
			
			result.data[dest_row_idx+j] = dot_product;
		}
		src1_row_idx += src1_cols;
		dest_row_idx += src2_cols;
	}
	
	// Copy to destination
	// For some reason, removing this step causes the filter to crash...it seems
	// that writing directly to the final location should be no different than
	// writing to a temporary one and then doing the copy to destination.
	// [The "reason" is that if one of the sources is also the destination, the source will be corrupted
	// before the multiply operation is finished.]
	for( i=0; i < result.rows; i++ )
	{
		int row_index = i*result.columns;
		
		for( j = 0; j < result.columns; j++ )
		{
			dest->data[row_index + j] = result.data[row_index + j];
		}
	}
	dest->rows = result.rows;
	dest->columns = result.columns;
	 
	return 0;
}

/*******************************************************************************
* Function Name  : mat_mult_new_dest( fMatrix* src1, fMatrix* src2, fMatrix* dest )
* Input          : fMatrix* src1, fMatrix* src2, fMatrix* dest
* Output         : fMatrix* dest
* Return         : 0 on success, 1 if failed
* Description    : 

Performs the operation
  dest = src1*src2
Same functionality of mat_mult, but this assumes that the destination is different
than any source, which avoids an extra copy for a slight performance increase.

src1 and src1 must have compatible dimensions for matrix multiplication.
src1, src2, and dest MUST have been preinitialized so that matrix->data points to
memory that has been allocated.
*******************************************************************************/
int mat_mult_new_dest( const fMatrix* src1, const fMatrix* src2, fMatrix* dest )
{
	int i,j,k;
	int src2_cols = src2->columns;
	int src2_colsx2 = src2_cols*2;
	int src1_cols = src1->columns;
	int src1_cols_half = src1_cols / 2; // For unrolling the inner loop: C division will truncate the 1/2 in case of an odd number of columns
	int odd = (src1_cols == src1_cols_half*2) ? 0 : 1;
	float dot_product;

	// Make sure dimensions are consistent with matrix multiplication rules
	if( src1->columns != src2->rows )
	{
		return 1;
	}
	
	// Compute matrix product
	int src1_row_idx = 0;
	int dest_row_idx = 0;
	int src2_idx = 0;
	for (i = 0; i < src1->rows; i++)
	{
		for (j = 0; j < src2_cols; j++)
		{
			src2_idx = j;
			dot_product = 0.0f;
			for (k = 0; k < src1_cols-1; k+=2)
			{
				dot_product += src1->data[src1_row_idx+k] * src2->data[src2_idx];
				dot_product += src1->data[src1_row_idx+k+1] * src2->data[src2_idx+src2_cols];
				src2_idx += src2_colsx2;
			}
			// clean up unrolling: handle odd # of cols:
			if (odd)
			{
				dot_product += src1->data[src1_row_idx+k] * src2->data[src2_idx];
			}
			
			dest->data[dest_row_idx+j] = dot_product;
		}
		src1_row_idx += src1_cols;
		dest_row_idx += src2_cols;
	}
	
	dest->rows = src1->rows;
	dest->columns = src2_cols;
	 
	return 0;
}

/*******************************************************************************
* Function Name  : mat_scalar_mult( float scalar, fMatrix* src, fMatrix* dest )
* Input          : float scalar, fMatrix* src, fMatrix* dest
* Output         : fMatrix* dest
* Return         : 0 on success, 1 if failed
* Description    : 

Performs the operation
  dest = scalar*src

src and dest MUST have been preinitialized so that matrix->data points to
memory that has been allocated.
*******************************************************************************/
int mat_scalar_mult( float scalar, fMatrix* src, fMatrix* dest )
{
	int i, j;

	dest->rows = src->rows;
	dest->columns = src->columns;
	 
	for( i = 0; i < src->rows; i++ )
	{
		int row_index = i*src->columns;
		for( j = 0; j < src->columns; j++ )
		{
			dest->data[row_index + j] = scalar*src->data[row_index + j];
		}
	}
	 
	return 0;
}

/*******************************************************************************
* Function Name  : mat_determinant( fMatrix* src, float* det )
* Input          : fMatrix* src, float* det
* Output         : float* det
* Return         : 0 on success, 1 if failed
* Description    : 

Performs the operation
  det = det(src)

src must have been preinitialized so that src->data points to allocated memory
*******************************************************************************/
int mat_determinant( fMatrix* src, float* det )
{
	// TODO: Implement this when/if it is needed
	return 1;
}

/*******************************************************************************
* Function Name  : mat_transpose( fMatrix* src, fMatrix* dest )
* Input          : fMatrix* src, fMatrix* dest
* Output         : fMatrix* dest
* Return         : 0 on success, 1 if failed
* Description    : 

Performs the operation
  det = src^T

src and dest must have been preinitialized so that matrix->data points to 
allocated memory
*******************************************************************************/
int mat_transpose( fMatrix* src, fMatrix* dest )
{
	int i, j;
	fMatrix result;
	
	float matrix_data[MAX_MATRIX_ROWS*MAX_MATRIX_COLS];
	
	result.data = matrix_data;
	result.rows = src->columns;
	result.columns = src->rows;
	 
	// Compute transpose
	for( i = 0; i < src->rows; i++ )
	{
		int source_row_index = i*src->columns;
		
		for( j = 0; j < src->columns; j++ )
		{
			result.data[j*result.columns + i] = src->data[source_row_index + j];
		}
	}
	
	// Copy to destination
	for( i=0; i < result.rows; i++ )
	{
		int row_index = i*result.columns;
		
		for( j = 0; j < result.columns; j++ )
		{
			dest->data[row_index + j] = result.data[row_index + j];
		}
	}
	
	dest->rows = result.rows;
	dest->columns = result.columns;
		  
	return 0;
}

/*******************************************************************************
* Function Name  : mat_transpose_new_dest( fMatrix* src, fMatrix* dest )
* Input          : fMatrix* src, fMatrix* dest
* Output         : fMatrix* dest
* Return         : 0 on success, 1 if failed
* Description    : 

Performs the operation
  det = src^T
Same functionality as mat_transpose, but assumes the the destination is unique from
either source, which means no copy is needed at the end.

src and dest must have been preinitialized so that matrix->data points to 
allocated memory
*******************************************************************************/
int mat_transpose_new_dest( const fMatrix* src, fMatrix* dest )
{
	int i, j;
	 
	// Compute transpose
	for( i = 0; i < src->rows; i++ )
	{
		int source_row_index = i*src->columns;
		
		for( j = 0; j < src->columns; j++ )
		{
			dest->data[j*src->rows + i] = src->data[source_row_index + j];
		}
	}

	dest->rows = src->columns;
	dest->columns = src->rows;
		  
	return 0;
}

/*******************************************************************************
* Function Name  : mat_create_identity( fMatrix* dest, int rows, int columns )
* Input          : fMatrix* dest, int rows, int columns
* Output         : fMatrix* dest
* Return         : 0 on success, 1 if failed
* Description    : 

Creates the identity matrix with the given rows and columns

dest must have been preinitialized so that dest->data points to 
allocated memory.  This function does NOT allocate memory for the matrix.

The number of rows specified does not have to equal the number of columns
*******************************************************************************/
int mat_create_identity( fMatrix* dest, int rows, int columns )
{
	int i, j;
	 
	dest->rows = rows;
	dest->columns = columns;
	 
	for( i = 0; i < dest->rows; i++ )
	{
		int row_index = i*dest->columns;
		
		for( j = 0; j < dest->columns; j++ )
		{
			dest->data[row_index + j] = 0;
			
			if( i == j)
			{
				dest->data[row_index + j] = 1;
			}
		}
	}
	
	return 0;
}

/*******************************************************************************
* Function Name  : mat_zero( fMatrix* dest, int rows, int columns )
* Input          : fMatrix* dest, int rows, int columns
* Output         : fMatrix* dest
* Return         : 0 on success, 1 if failed
* Description    : 

Creates a matrix of all zeros with the size specified in rows,columns

dest must have been preinitialized so that dest->data points to 
allocated memory.  This function does NOT allocate memory for the matrix.
*******************************************************************************/
int mat_zero( fMatrix* dest, int rows, int columns )
{
	int i, j;
	 
	dest->rows = rows;
	dest->columns = columns;
	 
	for( i = 0; i < dest->rows; i++ )
	{
		int row_index = i*dest->columns;
		
		for( j = 0; j < dest->columns; j++ )
		{
			dest->data[row_index + j] = 0;
		}
	}
	
	return 0;
}

/*******************************************************************************
* Function Name  : mat_copy( fMatrix* src, fMatrix* dest )
* Input          : fMatrix* src, fMatrix* dest
* Output         : fMatrix* dest
* Return         : 0 on success, 1 if failed
* Description    : 

Copies the matrix in src to dest

src and dest must have been preinitialized so that matrix->data points to 
allocated memory.  This function does NOT allocate memory for the dest matrix.
*******************************************************************************/
int mat_copy( fMatrix* src, fMatrix* dest )
{
	int i, j;
	 
	dest->rows = src->rows;
	dest->columns = src->columns;
	 
	// Perform copy operation
	for( i = 0; i < dest->rows; i++ )
	{
		int row_index = i*dest->columns;
		
		for( j = 0; j < dest->columns; j++ )
		{
			dest->data[row_index + j] = src->data[row_index + j];
		}
	}
	
	return 0;
}

/*******************************************************************************
* Function Name  : mat_invert_2x2( fMatrix* src, fMatrix* dest )
* Input          : fMatrix* src, fMatrix* dest
* Output         : fMatrix* dest
* Return         : 0 on success, 1 if failed
* Description    : 

Inverts the 2x2 matrix src and puts it in dest.  Both src and dest must have been
allocated before this function is called.
*******************************************************************************/
int mat_invert_2x2( fMatrix* src, fMatrix* dest )
{	
	// Make sure that this is a 2x2 matrix
	if( src->rows != 2 || src->columns!= 2)
	{
		return 1;
	}
	
	float a,b,c,d;
	
	a = src->data[0];
	b = src->data[1];
	c = src->data[2];
	d = src->data[3];
	
	float det = a*d-b*c;
	
	dest->data[0] = d/det;
	dest->data[1] = -b/det;
	dest->data[2] = -c/det;
	dest->data[3] = a/det;
	
	dest->rows = 2;
	dest->columns = 2;
	
	return 0;
}
