/* ------------------------------------------------------------------------------
  File: CHR_spi.c
  Author: CH Robotics
  Version: 1.0
  
  Description: Functions for handling SPI interactions
------------------------------------------------------------------------------ */ 

#include "stm32f10x.h"
#include "CHR_spi.h"

// State variable for keeping track of SPI waiting/read/write/error
uint8_t gSPI_State;

// The address that is being read from or written to
uint8_t gSPI_Curr_Address;

// A counter for keeping track of the number of bytes sent/received
// during the current SPI transaction
uint8_t gSPI_Bytes_Shifted;

// A buffer for holding the values to be sent, or for piecing together 
// the 32 bits being received
uint32_t gSPI_Buffer;



/*******************************************************************************
* Function Name  : SPI_ResetFSM
* Input          : None
* Output         : None
* Return         : None
* Description    : Resets the finite state machine used by the UM6 SPI interface
*******************************************************************************/

inline void SPI_ResetFSM(){
	gSPI_State = SPI_STATE_WAITING;
	gSPI_Curr_Address = SPI_NEED_ADDRESS;
	gSPI_Bytes_Shifted = 0;
	gSPI_Buffer = 0;
}


/*******************************************************************************
* Function Name  : SPI1_IRQHandler
* Input          : None
* Output         : None
* Return         : None
* Description    : SPI1 Slave Interrupt handler.
*******************************************************************************/
void SPI1_IRQHandler(void)
{
	//Struct used for sticking a command packet into the RX buffer
	USARTPacket cmdPacket;

	if(SPI_I2S_GetFlagStatus(SPI_SLAVE, SPI_I2S_FLAG_RXNE) == SET){
		// Received data

		uint8_t valueIn = SPI_I2S_ReceiveData(SPI_SLAVE);
		uint8_t valueOut = 0;

		switch(gSPI_State){
			case SPI_STATE_WAITING:
				
				//check that a valid read/write command has been received			
				if (valueIn == SPI_READ_CMD){
					gSPI_State = SPI_STATE_READ;
				}else if (valueIn == SPI_WRITE_CMD){
					gSPI_State = SPI_STATE_WRITE;
				}else{
					// Bad request or master is out of sync
				}
			
				break;
			
			case SPI_STATE_READ:

				//read address if this is the first time here
				//or if the master is switching addresses
				if((gSPI_Curr_Address == SPI_NEED_ADDRESS)|| (valueIn != 0 && gSPI_Bytes_Shifted > 0)) {

					//track the address to be read from
					gSPI_Curr_Address = valueIn;
					
					//buffer the value to be sent
					if(gSPI_Curr_Address < DATA_REG_START_ADDRESS){
						//Config Register
						gSPI_Buffer = g_config.r[gSPI_Curr_Address - CONFIG_REG_START_ADDRESS];	
					}else if(gSPI_Curr_Address < COMMAND_START_ADDRESS){
						//Data Register
						gSPI_Buffer = g_data.r[gSPI_Curr_Address - DATA_REG_START_ADDRESS];
					}else{
						//Command Register or Invalid Register
						//TODO: how do we want to handle this?
						gSPI_Buffer = 0x0000; //shift out zeros for now
					}
				
					//zero the bytes shifted counter
					gSPI_Bytes_Shifted = 0;
				}
						
				// Get the correct value ready to be sent out via SPI
				switch(gSPI_Bytes_Shifted){

					case 0:
						//Send first most significant byte (B3)
						valueOut = (uint8_t)((gSPI_Buffer >> 24) & 0xff); 
						break;
					case 1:
						//Send second most significant byte (B2)
						valueOut = (uint8_t)((gSPI_Buffer >> 16) & 0xff); 
						break;
					case 2:
						//Send third most significant byte (B1)
						valueOut = (uint8_t)((gSPI_Buffer >> 8) & 0xff); 
						break;
					case 3:
						//Send least significant byte (B0)
						valueOut = (uint8_t)(gSPI_Buffer & 0xff); 
						break;
					default:
						break;
						//This should not be reachable
					
						//At this point there is nothing left to shift
						//and we are not allowing sequential reading at this time
						//SPI_ResetFSM();
				}
			
				gSPI_Bytes_Shifted += 1;
			

				if (gSPI_Bytes_Shifted > 4){
					SPI_ResetFSM();
				}
			
				break; //break SPI_STATE_READ case

			case SPI_STATE_WRITE:
				//first get an address, then buffer 4 bytes, then set the register
			
				valueOut = 0; 
			
				//read address if this is the first time here
				//or if the master is switching addresses
				if(gSPI_Curr_Address == SPI_NEED_ADDRESS) {

					//track the address to be read from
					gSPI_Curr_Address = valueIn;
					
					//TODO: handle writing to bad registers
				
					//zero the bytes shifted counter
					gSPI_Bytes_Shifted = 0;
				
					//zero the buffer
					gSPI_Buffer = 0;

					//If a command was received, add it to the rx buffer immediately
					if (gSPI_Curr_Address >= COMMAND_START_ADDRESS){ 
									
						//Just send the version ID if that is the address they want
						//The SPI_READ_STATE code can handle most of this 
						if (gSPI_Curr_Address == CHR_GET_FW_VERSION){
							//Send first 8 bits and then switch to READ state to handle the rest
							gSPI_Buffer = UM6_FIRMWARE_REVISION;
							//Send first most significant byte (B3)
							valueOut = (uint8_t)((gSPI_Buffer >> 24) & 0xff);
							gSPI_Bytes_Shifted = 1;
							gSPI_State = SPI_STATE_READ;
							SPI_I2S_SendData(SPI_SLAVE, valueOut);    
							return;
						}
					
						//Create a corresponding USART packet and put it in the RX queue 
						cmdPacket.PT = PACKET_NO_DATA; // Command packet, specified as 0 in the datasheet
						cmdPacket.address = gSPI_Curr_Address; 
						//Calling ComputeChecksum() is also an option:
						cmdPacket.checksum = 's' + 'n' + 'p' + cmdPacket.PT + cmdPacket.address; 
						cmdPacket.data_length = 0;
						cmdPacket.address_type = ADDRESS_TYPE_COMMAND;

						//Add the command packet to the RX packet buffer
						//AddRXPacket(&cmdPacket);
						//Copied the AddRXPacket code here
						gRXPacketBuffer[gRXPacketBufferEnd] = cmdPacket;
						gRXPacketBufferEnd++;
						if( gRXPacketBufferEnd >= RX_PACKET_BUFFER_SIZE )  gRXPacketBufferEnd = 0;						
					
						//Indicate that a new packet has been received
						gRXPacketReceived = 1;
					}				
				
					//otherwise don't send any bytes this time
					break;
				}
		
			
				// Read 4 bytes into the buffer, which will be written to the specified register
				switch(gSPI_Bytes_Shifted){

					case 0:
						//Buffer first most significant byte (B3)
						gSPI_Buffer |=  ((uint32_t)valueIn << 24);						
						break;
					case 1:
						//Buffer second most significant byte (B2)
						gSPI_Buffer |=  ((uint32_t)valueIn << 16);
						break;
					case 2:
						//Buffer third most significant byte (B1)
						gSPI_Buffer |=  ((uint32_t)valueIn << 8);
						break;
					case 3:
						//Buffer least significant byte (B0)
						gSPI_Buffer |=  (uint32_t)valueIn;

						if(gSPI_Curr_Address < DATA_REG_START_ADDRESS){
							//Config Register
							g_config.r[gSPI_Curr_Address - CONFIG_REG_START_ADDRESS] = gSPI_Buffer;	
							CopyConfigArrayToStates();
						}else if(gSPI_Curr_Address < COMMAND_START_ADDRESS){
							//Data Register
							//Don't do anything (writing to a data register is not useful)
							valueOut = 0;
						}else{
							//Command Register
							valueOut = 0;
						}										
					
						SPI_ResetFSM();
						break;
					default:
						//This should not be reachable
					
						//At this point there is nothing left to shift
						//and we are not allowing sequential writing at this time
						SPI_ResetFSM();
				}

				if(gSPI_Curr_Address != SPI_NEED_ADDRESS) gSPI_Bytes_Shifted += 1;
			
				break; //break SPI_STATE_WRITE case

			default:

				//This should not be reachable
				//If if has been reached the FSM is not
				//being used correctly.
				SPI_ResetFSM();
		}

		// Send the chosen byte
		SPI_I2S_SendData(SPI_SLAVE, valueOut);    
	}
}