/*
 * EventScheduler.c
 *
 * Created: 4/10/2014 12:45:03 PM
 *  Author: Caleb
 */ 

#include "stm32f10x.h"
#include "EventScheduler.h"

volatile uint8_t g_event_scheduler_running;
volatile uint32_t g_event_count;
volatile UTC_Timing g_UTC_timing;
volatile float g_sys_x0;
volatile uint8_t g_timer_rolled_over;

volatile Event g_event_array[MAX_EVENTS];

/*******************************************************************************
* Function Name  : TIM2_IRQHandler
* Input          : None
* Output         : None
* Return         : None
* Description    : On counter rollover, add the overflow time to the system time.
*******************************************************************************/
void TIM2_IRQHandler( void )
{
    if( TIM_GetITStatus(TIM2, TIM_IT_Update ) != RESET )
    {
        g_sys_x0 += COUNTER_OVERFLOW_TIME;
		g_timer_rolled_over = 1;
	
		if (g_sys_x0 > SECONDS_IN_EPOCH)
		{
			g_sys_x0 -= SECONDS_IN_EPOCH;
			g_UTC_timing.t_sys_k_d1 -= SECONDS_IN_EPOCH;
		}

        // Clear pending interrupt bit
        TIM_ClearITPendingBit(TIM2, TIM_IT_Update);
    }
}

/*******************************************************************************
* Function Name  : init_event_scheduler
* Input          : NA
* Output         : NA
* Return         : None
* Description    : Performs all required initialization to get the event scheduler
					ready to go.
*******************************************************************************/
void init_event_scheduler()
{
	// Preinitialize event scheduler control variables
	g_event_scheduler_running = 0;
    g_event_count = 0;
	g_sys_x0 = 0;
}

/*******************************************************************************
* Function Name  : run_event_handlers
* Input          : NA
* Output         : NA
* Return         : None
* Description    : Evaluates each scheduled event and runs whatever needs to run
					This function should be called regularly from the main loop
					to handle events.
*******************************************************************************/
void run_event_handlers()
{
	int index;
	float current_time = get_system_time_uncal();

	// Return if the event scheduler isn't running
	if( g_event_scheduler_running == 0 )
	{
		return;
	}

	// Iterate through all events we have.  Execute those whose time has expired.
	for( index = 0; index < g_event_count; index++ )
	{
		float event_period = g_event_array[index].properties.period;
		float delta_t = unwrap_delta_T(current_time - g_event_array[index].last_event_time);
		
		if( delta_t >= event_period )
		{
			// Execute callback function
			g_event_array[index].properties.callback_func(current_time, g_event_array[index].properties.user_args );
		
			// Remove this event if it isn't periodic
			if( g_event_array[index].properties.is_periodic == 0 )
			{
				remove_event(index);
				index--;
			}
			else
			{
				// Set current_time as the last event time
				g_event_array[index].last_event_time = current_time;
			}
		}
	}
}

/*******************************************************************************
* Function Name  : clear_all_events
* Input          : None
* Output         : None
* Return         : None
* Description    :

Stops the event scheduler and removes all events.

*******************************************************************************/
void clear_all_events( )
{
	g_event_count = 0;
}

/*******************************************************************************
* Function Name  : clear_all_COM_events
* Input          : None
* Output         : None
* Return         : None
* Description    :

Removes all COM events from the scheduler.  This allows us to easily reconfigure
COM packet transmissions when the configuration settings are changed.  We simply
remove all COM events and then re-add the new ones.

*******************************************************************************/
void clear_all_COM_events( )
{
	int index;

	// Remove any existing COM events in the scheduler
	for( index = 0; index < (int)g_event_count; index++ )
	{
		if( g_event_array[index].properties.is_COM_event == 1 )
		{
			remove_event(index);
			index--;
		}
	}

	return;
}

/*******************************************************************************
* Function Name  : stop_event_scheduler
* Input          : NA
* Output         : NA
* Return         : None
* Description    : Stops the event scheduler
*******************************************************************************/
void stop_event_scheduler()
{
	g_event_scheduler_running = 0;

	TIM_SetCounter(TIM2,0);
	TIM_Cmd(TIM2, DISABLE);
}

/*******************************************************************************
* Function Name  : start_event_scheduler
* Input          : NA
* Output         : NA
* Return         : None
* Description    : Starts the event scheduler
*******************************************************************************/
void start_event_scheduler()
{
	g_event_scheduler_running = 1;

	// Start timer2 (system clock, also used by the event scheduler)
	TIM_SetCounter(TIM2,0);
	TIM_Cmd(TIM2, ENABLE);
}

/*******************************************************************************
* Function Name  : remove_event
* Input          : NA
* Output         : NA
* Return         : None
* Description    : Removes the event with the given index from the event scheduler

*******************************************************************************/
uint32_t remove_event( uint32_t event_index )
{
	uint32_t index;
	
	for( index = event_index; event_index < (g_event_count-1); index++ )
	{
		g_event_array[index] = g_event_array[index+1];
	}
	
	g_event_count--;
	
	return 0;
}

/*******************************************************************************
* Function Name  : remove_event_by_id
* Input          : NA
* Output         : NA
* Return         : 0 if the id was removed, 1 if the id wasn't found
* Description    : Removes the event with the given id from the event scheduler

*******************************************************************************/
uint32_t remove_event_by_id( uint32_t id )
{
	int index;

	for( index = 0; index < g_event_count; index++ )
	{
		if( g_event_array[index].properties.event_id == id )
		{
			remove_event(index);
			return 0;
		}
	}

	return 1;
}

/*******************************************************************************
* Function Name  : add_event
* Input          : NA
* Output         : NA
* Return         : None
* Description    : Adds an event to the event scheduler
*******************************************************************************/
uint32_t add_event( EventProperties* event_properties )
{
	if( g_event_count == (MAX_EVENTS - 1) )
	{
		return 0;
	}
	
	g_event_array[g_event_count].properties = *event_properties;
	g_event_array[g_event_count].last_event_time = get_system_time_uncal();
	
	g_event_count++;
	
	return 1;
}

/*******************************************************************************
* Function Name  : get_system_time
* Input          : NA
* Output         : NA
* Return         : None
* Description    : Gets the current system time from the variable g_sys_x0 + RTC value
*******************************************************************************/
float get_system_time_uncal()
{
	float counter_value;
	float x0;

	// If the system time counter rolls over between when we extract the counter value and when we use g_sys_x0, then
	// the reported system time will be correct.  On rollover, the g_counter_rollover flag is set.  So, before we extract anything,
	// we set the flag low.  Then we check it again after we extract the time variables.  If the counter rolled over sometime during the
	// process, we do it again to make sure we have a good system time.
	// The variable g_timer_rolled_over MUST be declared volatile so that the optimizer doesn't get rid of the while loop.
	do
	{
		g_timer_rolled_over = 0;
	
		counter_value = (float)TIM_GetCounter(TIM2);
		x0 = g_sys_x0;
	} while( g_timer_rolled_over );
	
	return x0 + counter_value*COUNTER_PERIOD;
}

/*******************************************************************************
* Function Name  : get_UTC_time
* Input          : NA
* Output         : NA
* Return         : None
* Description    : Gets the current calibrated system time.  If there is not a good
                   GPS lock, this will just return the system time from the free-
                   running counter.  Wraps by the day.
*******************************************************************************/
float get_UTC_time()
{
    return ((g_UTC_timing.c1*get_system_time_uncal()) + g_UTC_timing.c0); 
}

/*******************************************************************************
* Function Name  : init_time_struct
* Input          :
* Output         :
* Return         :
* Description    :

Initializes the values in the time calibration.
Set to pass system time through by default.

*******************************************************************************/
void init_gps_time_struct()
{
    g_UTC_timing.t_sys_k = get_system_time_uncal();
    g_UTC_timing.t_sys_k_d1 = -100.0;

    g_UTC_timing.t_utc_k = g_UTC_timing.t_sys_k;
    g_UTC_timing.t_utc_k_d1 = -100.0;

    g_UTC_timing.c0 = 0.0;
    g_UTC_timing.c1 = 1.0;

    g_UTC_timing.time_of_PPS = 0.0;
}

/*******************************************************************************
* Function Name  : compute_time_calibration
* Input          :
* Output         :
* Return         :
* Description    :

Computes the calibration for UTC time
See the UTC_Timing struct declaration for details on the calculations

*******************************************************************************/
void compute_time_calibration()
{
    // New value for t_sys_k should be set in PPS interrupt handler.
    // New value for t_utc_k should be set by caller of this function (use set_UTC_time_at_PPS(float) function)
    
    // Find calibration PPS pulses have been successive
    if (abs(g_UTC_timing.t_utc_k - g_UTC_timing.t_utc_k_d1) < GPS_TIME_DELAY_INTERVAL)
    {
        double den = g_UTC_timing.t_sys_k - g_UTC_timing.t_sys_k_d1;
        double tmp1 = g_UTC_timing.c0 = g_UTC_timing.t_sys_k*g_UTC_timing.t_utc_k_d1 - g_UTC_timing.t_sys_k_d1*g_UTC_timing.t_utc_k;

        g_UTC_timing.c0 = (float) (tmp1/den);

        tmp1 = g_UTC_timing.t_utc_k - g_UTC_timing.t_utc_k_d1;

        g_UTC_timing.c1 = (float) (tmp1/den);
    }

    // Update delays
    g_UTC_timing.t_sys_k_d1 = g_UTC_timing.t_sys_k;
    g_UTC_timing.t_utc_k_d1 = g_UTC_timing.t_utc_k;
}

/*******************************************************************************
* Function Name  : set_UTC_time_at_PPS
* Input          :
* Output         :
* Return         :
* Description    :

Sets the UTC time in the timing calibration structure given the UTC time 
from the first GPS packet after the PPS pulse.

Handles wrapping around the epoch time and setting of the delayed value g_UTC_timing.t_utc_k_d1
if wrapping occurred.  This wrapping is a bit more complicated then the system time delay wrapping
because the GPS time coming in does not wrap, only the internal UTC time is wrapped.

*******************************************************************************/
void set_UTC_time_at_PPS(float time)
{
    int timeout = 0;    // Timeout counter for the wrapping.  1000 should be plenty if the SECONDS_IN_EPOCH is kept
                        // above the system timer wrap value (~171sec) since the UTC time coming from the GPS will 
                        // never be above 24 hrs in seconds (~86400sec)
                        // A good value for SECONDS_IN_EPOCH would be 86400 anyway.
    
    if (time > SECONDS_IN_EPOCH)
    {
        while (time > SECONDS_IN_EPOCH && timeout < 1000)
        {
            time -= SECONDS_IN_EPOCH;
            timeout++;
        }

        // Wrap the delayed time if needed
        // Since t_utc_k_d1 is only set from the wrapped utc time, it only needs to be reduced by SECONDS_IN_EPOCH
        if ((double)time - g_UTC_timing.t_utc_k_d1 < -(double)SECONDS_IN_EPOCHd2)
            g_UTC_timing.t_utc_k_d1 -= (double)SECONDS_IN_EPOCH; 
    }

    g_UTC_timing.t_utc_k = (double)time;     
}

/*******************************************************************************
* Function Name  : wrap_delta_T
* Input          : NA
* Output         : NA
* Return         : None
* Description    : Checks a given delta_T value to determine if the system time 
                   has wrapped to 0.  If so, subtract SECONDS_IN_EPOCH.
                   Wraps on delta_T > SECONDS_IN_EPOCH/2. 
*******************************************************************************/
float unwrap_delta_T(float delta_T)
{
    if (delta_T < -SECONDS_IN_EPOCHd2)
        return (delta_T + SECONDS_IN_EPOCH);
    else
        return delta_T;
}
