/* ------------------------------------------------------------------------------
  File: MPU9150.c
  Author: CH Robotics
  Version: 1.0
  
  Description: Functions for interacting with the MPU9150
------------------------------------------------------------------------------ */ 

#include "stm32f10x.h"
#include "MPU9150.h"
#include "CHR_i2c.h"

/*******************************************************************************
* Function Name  : MPU9150_init
* Input          : None
* Output         : uint8_t status
* Return         : 1 if success, 0 if fail
* Description    : Initializes the MPU9150
						 
*******************************************************************************/
uint8_t MPU9150_init( )
{
    uint8_t txBuf[5];
    uint8_t returnval;

	// Set up clock and turn off i2c master mode
    txBuf[0] = MPU9150_REG_USR_CNTRL;
	txBuf[1] = 0x00;			// Make sure i2c master mode and FIFO are disabled
    txBuf[2] = 0x01;            // 0b00000001 - Disable sleep mode and select x-axis gyro as reference clock

    returnval = i2cBatchWrite( MPU9150_SLAVE_ADDR, txBuf, 3 );

    if( returnval != I2C_SUCCESS )
    {
        return returnval;
    }

	// Enable I2C bypass to provide direct access to magnetometer
	txBuf[0] = MPU9150_REG_INT_PIN_CONFIG;
	txBuf[1] = 0x02;			// Make sure i2c master mode and FIFO are disabled

    returnval = i2cBatchWrite( MPU9150_SLAVE_ADDR, txBuf, 2 );

    if( returnval != I2C_SUCCESS )
    {
        return returnval;
    }

	// Configure rate gyro and accelerometer
    txBuf[0] = MPU9150_REG_CONFIG;
    txBuf[1] = 0x02;            // FSYNC input disabled, 94 Hz accel bandwidth, 98 Hz gyro bandwidth
    txBuf[2] = 0x18;            // 0b00011000, +/- 2000 degree gyro FS range, all self-test features turned off
    txBuf[3] = 0x10;            // 0b00010000, +/- 8g accel FS range, self-test turned off
    txBuf[4] = 0x00;            // All FIFO features turned off

    returnval = i2cBatchWrite( MPU9150_SLAVE_ADDR, txBuf, 5 );

    return returnval;
}

/*******************************************************************************
* Function Name  : MPU9150_get_accel_gyro_data
* Input          : None
* Output         : uint8_t status
* Return         : 1 if success, 0 if fail
* Description    : Retrieves gyro, accelerometer, and temperature data from the MPU9150
						 
*******************************************************************************/
uint8_t MPU9150_get_accel_gyro_data( )
{
    i2cBatchRead( MPU9150_SLAVE_ADDR, MPU9150_REG_INT_STATUS, g_i2cRxBuf, 15 );

    return 1;
}

/*******************************************************************************
* Function Name  : MPU9150_start_mag_conversion
* Input          : None
* Output         : uint8_t status
* Return         : 1 if success, 0 if fail
* Description    : 

Tells the magnetometer on the MPU9150 to start a conversion.
						 
*******************************************************************************/
uint8_t MPU9150_start_mag_conversion( )
{
	uint8_t returnval;
	uint8_t txBuf[2];

	txBuf[0] = MPU9150_REG_MAG_CNTL;
	txBuf[1] = 0x01;

	returnval = i2cBatchWrite( MPU9150_MAG_SLAVE_ADDR, txBuf, 2 );
	
	return returnval;
}

/*******************************************************************************
* Function Name  : MPU9150_get_mag_data
* Input          : None
* Output         : uint8_t status
* Return         : 1 if success, 0 if fail
* Description    : 

Reads the magnetometer data registers on the MPU9150.  Prior to calling this,
MPU9150_start_mag_conversion() should be called and given time to complete
before reading the data.
						 
*******************************************************************************/
uint8_t MPU9150_get_mag_data( )
{
	i2cBatchRead( MPU9150_MAG_SLAVE_ADDR, MPU9150_REG_MAG_XH, g_i2cRxBuf, 6 );

    return 1;
}
