/* ------------------------------------------------------------------------------
  File: UM6_config.c
  Author: CH Robotics
  Version: 1.0
  
  Description: Function definitions for AHRS configuration
------------------------------------------------------------------------------ */ 
#include <math.h>
#include "stm32f10x.h"
#include "UM6_config.h"
#include "UM6_usart.h"
#include "CHR_matrix.h"
#include "UM6_states.h"
#include "EventScheduler.h"
#include "util.h"

// Global structure storing IMU configuration parameters
CHR_config g_config;
CHR_data g_data;
CHR_hidden_settings g_hidden_settings;

uint16_t g_zeroing_gyros;
float g_gyro_sum[3];
uint16_t g_gyro_samples_collected;
uint16_t gGyrosCalibrated;

uint8_t g_save_flash_config;
uint8_t g_save_factory_config;

uint16_t gSelfTestEnabled;
uint16_t gSelfTestSamplesIgnored;

volatile uint8_t g_PPS_received;

/*******************************************************************************
* Function Name  : GetConfiguration
* Input          : None
* Output         : None
* Return         : None
* Description    : Fills the g_config structure with IMU configuration data.
						  If configuration data has been written to flash, then load it.
						  If not, then use factory defaults.
*******************************************************************************/
void GetConfiguration( )
{
	// Set up hidden settings pointers to point to the hidden settings array.
	// We need to do this every time on startup because, unlike our code on the Atmel chip,
	// these pointers are not saved to flash - only what they point to is saved.
	g_hidden_settings.p_gyro_variance = &g_hidden_settings.f[0];
	g_hidden_settings.p_accel_variance = &g_hidden_settings.f[1];
	g_hidden_settings.p_mag_variance = &g_hidden_settings.f[2];
	g_hidden_settings.p_GPS_course_variance = &g_hidden_settings.f[3];
	g_hidden_settings.p_GPS_position_variance = &g_hidden_settings.f[4];
	g_hidden_settings.p_GPS_velocity_variance = &g_hidden_settings.f[5];
	g_hidden_settings.p_static_press_variance = &g_hidden_settings.f[6];
	g_hidden_settings.p_diff_press_variance = &g_hidden_settings.f[7];

	g_hidden_settings.p_Q_uvw = &g_hidden_settings.f[9];
	g_hidden_settings.p_Q_quaternion = &g_hidden_settings.f[10];
	g_hidden_settings.p_Q_gps_position = &g_hidden_settings.f[11];
	g_hidden_settings.p_Q_bias = &g_hidden_settings.f[12];
	g_hidden_settings.p_Q_euler_angles = &g_hidden_settings.f[13];

	g_hidden_settings.p_low_vg_accel_noise_factor = &g_hidden_settings.f[14];

	//		g_hidden_settings.p_lpf_tau_airspeed = &g_hidden_settings.f[15];
	g_hidden_settings.p_lpf_tau_groundspeed = &g_hidden_settings.f[16];
	g_hidden_settings.p_lpf_tau_accel = &g_hidden_settings.f[17];
	g_hidden_settings.p_lpf_tau_gyro = &g_hidden_settings.f[18];
	g_hidden_settings.p_lpf_tau_mag = &g_hidden_settings.f[19];
	//		g_hidden_settings.p_lpf_tau_abs_pressure = &g_hidden_settings.f[20];

	//		g_hidden_settings.p_adc_raw_to_volts = &g_hidden_settings.f[21];
	//		g_hidden_settings.p_rho = &g_hidden_settings.f[23];

	g_hidden_settings.p_Cgbx0 = &g_hidden_settings.f[25];
	g_hidden_settings.p_Cgbx1 = &g_hidden_settings.f[26];
	g_hidden_settings.p_Cgbx2 = &g_hidden_settings.f[27];
	g_hidden_settings.p_Cgbx3 = &g_hidden_settings.f[28];

	g_hidden_settings.p_Cgby0 = &g_hidden_settings.f[29];
	g_hidden_settings.p_Cgby1 = &g_hidden_settings.f[30];
	g_hidden_settings.p_Cgby2 = &g_hidden_settings.f[31];
	g_hidden_settings.p_Cgby3 = &g_hidden_settings.f[32];

	g_hidden_settings.p_Cgbz0 = &g_hidden_settings.f[33];
	g_hidden_settings.p_Cgbz1 = &g_hidden_settings.f[34];
	g_hidden_settings.p_Cgbz2 = &g_hidden_settings.f[35];
	g_hidden_settings.p_Cgbz3 = &g_hidden_settings.f[36];

	g_hidden_settings.p_Cgsx0 = &g_hidden_settings.f[37];
	g_hidden_settings.p_Cgsx1 = &g_hidden_settings.f[38];
	g_hidden_settings.p_Cgsx2 = &g_hidden_settings.f[39];
	g_hidden_settings.p_Cgsx3 = &g_hidden_settings.f[40];

	g_hidden_settings.p_Cgsy0 = &g_hidden_settings.f[41];
	g_hidden_settings.p_Cgsy1 = &g_hidden_settings.f[42];
	g_hidden_settings.p_Cgsy2 = &g_hidden_settings.f[43];
	g_hidden_settings.p_Cgsy3 = &g_hidden_settings.f[44];

	g_hidden_settings.p_Cgsz0 = &g_hidden_settings.f[45];
	g_hidden_settings.p_Cgsz1 = &g_hidden_settings.f[46];
	g_hidden_settings.p_Cgsz2 = &g_hidden_settings.f[47];
	g_hidden_settings.p_Cgsz3 = &g_hidden_settings.f[48];
	
	g_hidden_settings.Mg.data = &g_hidden_settings.f[49];		// This is a 3x3 matrix and should occupy 9 array indices
	g_hidden_settings.Mg.rows = 3;
	g_hidden_settings.Mg.columns = 3;

	g_hidden_settings.p_Cabx0 = &g_hidden_settings.f[58];
	g_hidden_settings.p_Cabx1 = &g_hidden_settings.f[59];
	g_hidden_settings.p_Cabx2 = &g_hidden_settings.f[60];
	g_hidden_settings.p_Cabx3 = &g_hidden_settings.f[61];

	g_hidden_settings.p_Caby0 = &g_hidden_settings.f[62];
	g_hidden_settings.p_Caby1 = &g_hidden_settings.f[63];
	g_hidden_settings.p_Caby2 = &g_hidden_settings.f[64];
	g_hidden_settings.p_Caby3 = &g_hidden_settings.f[65];

	g_hidden_settings.p_Cabz0 = &g_hidden_settings.f[66];
	g_hidden_settings.p_Cabz1 = &g_hidden_settings.f[67];
	g_hidden_settings.p_Cabz2 = &g_hidden_settings.f[68];
	g_hidden_settings.p_Cabz3 = &g_hidden_settings.f[69];

	g_hidden_settings.p_Casx0 = &g_hidden_settings.f[70];
	g_hidden_settings.p_Casx1 = &g_hidden_settings.f[71];
	g_hidden_settings.p_Casx2 = &g_hidden_settings.f[72];
	g_hidden_settings.p_Casx3 = &g_hidden_settings.f[73];

	g_hidden_settings.p_Casy0 = &g_hidden_settings.f[74];
	g_hidden_settings.p_Casy1 = &g_hidden_settings.f[75];
	g_hidden_settings.p_Casy2 = &g_hidden_settings.f[76];
	g_hidden_settings.p_Casy3 = &g_hidden_settings.f[77];

	g_hidden_settings.p_Casz0 = &g_hidden_settings.f[78];
	g_hidden_settings.p_Casz1 = &g_hidden_settings.f[79];
	g_hidden_settings.p_Casz2 = &g_hidden_settings.f[80];
	g_hidden_settings.p_Casz3 = &g_hidden_settings.f[81];

	g_hidden_settings.Ma.data = &g_hidden_settings.f[82];	// This is a 3x3 matrix and should occupy 9 array indices
	g_hidden_settings.Ma.rows = 3;
	g_hidden_settings.Ma.columns = 3;

	g_hidden_settings.p_Cmbx0 = &g_hidden_settings.f[91];
	g_hidden_settings.p_Cmbx1 = &g_hidden_settings.f[92];
	g_hidden_settings.p_Cmbx2 = &g_hidden_settings.f[93];
	g_hidden_settings.p_Cmbx3 = &g_hidden_settings.f[94];

	g_hidden_settings.p_Cmby0 = &g_hidden_settings.f[95];
	g_hidden_settings.p_Cmby1 = &g_hidden_settings.f[96];
	g_hidden_settings.p_Cmby2 = &g_hidden_settings.f[97];
	g_hidden_settings.p_Cmby3 = &g_hidden_settings.f[98];

	g_hidden_settings.p_Cmbz0 = &g_hidden_settings.f[99];
	g_hidden_settings.p_Cmbz1 = &g_hidden_settings.f[100];
	g_hidden_settings.p_Cmbz2 = &g_hidden_settings.f[101];
	g_hidden_settings.p_Cmbz3 = &g_hidden_settings.f[102];

	g_hidden_settings.p_Cmsx0 = &g_hidden_settings.f[103];
	g_hidden_settings.p_Cmsx1 = &g_hidden_settings.f[104];
	g_hidden_settings.p_Cmsx2 = &g_hidden_settings.f[105];
	g_hidden_settings.p_Cmsx3 = &g_hidden_settings.f[106];

	g_hidden_settings.p_Cmsy0 = &g_hidden_settings.f[107];
	g_hidden_settings.p_Cmsy1 = &g_hidden_settings.f[108];
	g_hidden_settings.p_Cmsy2 = &g_hidden_settings.f[109];
	g_hidden_settings.p_Cmsy3 = &g_hidden_settings.f[110];

	g_hidden_settings.p_Cmsz0 = &g_hidden_settings.f[111];
	g_hidden_settings.p_Cmsz1 = &g_hidden_settings.f[112];
	g_hidden_settings.p_Cmsz2 = &g_hidden_settings.f[113];
	g_hidden_settings.p_Cmsz3 = &g_hidden_settings.f[114];

	g_hidden_settings.Mm.data = &g_hidden_settings.f[115];		// This is a 3x3 matrix and should occupy 9 array indices
	g_hidden_settings.Mm.rows = 3;
	g_hidden_settings.Mm.columns = 3;

	g_hidden_settings.p_mag_ref_x = &g_hidden_settings.f[124];
	g_hidden_settings.p_mag_ref_y = &g_hidden_settings.f[125];
	g_hidden_settings.p_mag_ref_z = &g_hidden_settings.f[126];
		
	 // If flash has not been programmed yet, then use default configuration.  Otherwise, load configuration from flash
	 if( FGET_FLASH_UNINITIALIZED() )
	 {
		  ResetToFactory();
	 }
	 else
	 {
		  LoadConfigFromFLASH( CHR_USE_CONFIG_ADDRESS );
	 }
	 
	 CopyConfigArrayToStates();
}

/*******************************************************************************
* Function Name  : ClearGlobalData
* Input          : None
* Output         : None
* Return         : None
* Description    : 

Zeroes out all data in the g_data.r[] array.
*******************************************************************************/
void ClearGlobalData()
{
	 int i;
	 
	 for( i = 0; i <  DATA_ARRAY_SIZE; i++ )
	 {
		  g_data.r[i] = 0;
	 }
}

/*******************************************************************************
* Function Name  : LoadConfigFromFLASH( )
* Input          : None
* Output         : None
* Return         : None
* Description    : Loads the current sensor configuration from FLASH
*******************************************************************************/
void LoadConfigFromFLASH( int flash_address_flag )
{
	int i;

	uint32_t flash_start_address;

	if( flash_address_flag == CHR_USE_CONFIG_ADDRESS )
	{
		flash_start_address = FLASH_START_ADDRESS;
	}
	else
	{
		flash_start_address = FACTORY_FLASH_ADDRESS;
	}	 

	// Load configuration settings
	for( i = 0; i < CONFIG_ARRAY_SIZE; i++ )
	{
		g_config.r[i] = (uint32_t)( *(__IO uint32_t*)(flash_start_address + 4*i) );
	}
 
	// Load hidden settings
	for( i = 0; i < HIDDEN_ARRAY_SIZE; i++ )
	{
		g_hidden_settings.r[i] = (uint32_t)( *(__IO uint32_t*)(flash_start_address + 4*CONFIG_ARRAY_SIZE + 4*i) );
	}
}

/*******************************************************************************
* Function Name  : WriteConfigurationToFlash
* Input          : None
* Output         : None
* Return         : None
* Description    : Writes the current IMU configuration to flash.
*******************************************************************************/
int32_t WriteConfigurationToFlash( int write_location_flag )
{
	 FLASH_Status FLASHStatus;
	 uint32_t i;
	 
	 uint32_t flash_start_address;
	 
	 if( write_location_flag == CHR_USE_FACTORY_ADDRESS )
	 {
		  flash_start_address = FACTORY_FLASH_ADDRESS;
	 }
	 else
	 {
		  flash_start_address = FLASH_START_ADDRESS;
	 }

	 FLASH_Unlock();

	 // Clear all pending flags
	 FLASH_ClearFlag(FLASH_FLAG_BSY | FLASH_FLAG_EOP | FLASH_FLAG_PGERR | FLASH_FLAG_WRPRTERR);
	
	// Erase the flash page that we'll be writing to.  This assumes that flash_start_address is the start of a page, and that everything will fit in that page.
	FLASHStatus = FLASH_ErasePage(flash_start_address);
			  
	if( FLASHStatus != FLASH_COMPLETE )
	{
		FLASH_Lock();
		return FLASHStatus;
	}
 
	 // Write configuration data
	 for (i = 0; i < CONFIG_ARRAY_SIZE; i++ )
	 {
		  // Write FLASH data
		  FLASHStatus = FLASH_ProgramWord(flash_start_address + 4*i, g_config.r[i]);
		  
		  if( FLASHStatus != FLASH_COMPLETE )
		  {
				FLASH_Lock();
				return FLASHStatus;
		  }

		  // Make sure new flash memory contents match
		  if( g_config.r[i] != (uint32_t)( *(__IO uint32_t*)(flash_start_address + 4*i) ) )
		  {
				FLASH_Lock();
				return FLASH_TIMEOUT;
		  }		  
	 }

	// Write hidden data
	for (i = 0; i < HIDDEN_ARRAY_SIZE; i++ )
	{
		// Write FLASH data
		FLASHStatus = FLASH_ProgramWord(flash_start_address + 4*CONFIG_ARRAY_SIZE + 4*i, g_hidden_settings.r[i]);

		if( FLASHStatus != FLASH_COMPLETE )
		{
			FLASH_Lock();
			return FLASHStatus;
		}

		// Make sure new flash memory contents match
		if( g_hidden_settings.r[i] != (uint32_t)( *(__IO uint32_t*)(flash_start_address + 4*CONFIG_ARRAY_SIZE + 4*i) ) )
		{
			FLASH_Lock();
			return FLASH_TIMEOUT;
		}		  
	}

	FLASH_Lock();

	return FLASH_COMPLETE;
	 
}

/*******************************************************************************
* Function Name  : StartSelfTest
* Input          : None
* Output         : None
* Return         : None
* Description    : Sets the gSelfTestEnable flag, which triggers the self-test
				       sequence on the IMU.
*******************************************************************************/
void StartSelfTest( void )
{
	 gSelfTestEnabled = 1;
	 gSelfTestSamplesIgnored = 0;

	 // TODO: Activate magnetometer bias test
	 
	 // Assert self-test pin
	 // TODO: Figure out which self-test pin to actually use on the UM6
//	 GPIO_WriteBit( GPIOA, GPIO_Pin_13, Bit_SET );
}

/*******************************************************************************
* Function Name  : StopSelfTest
* Input          : None
* Output         : None
* Return         : None
* Description    : 
*******************************************************************************/
void StopSelfTest( uint16_t result )
{
	 // Clear self-test pin
	 // TODO: Figure out which self-test pin to actually use on the UM6
//	 GPIO_WriteBit( GPIOA, GPIO_Pin_13, Bit_RESET );
	 
	 // TODO: Stop magnetometer bias test
	 
	 gSelfTestEnabled = 0;
	 
	 // TODO: write self-test results to the UM6_STATUS register
	 
	 // TODO: send a packet with the test results
}

/*******************************************************************************
* Function Name  : ResetToFactory
* Input          : None
* Output         : None
* Return         : None
* Description    : Resets all AHRS settings to factory default.
*******************************************************************************/
void ResetToFactory( void )
{
	fConvert f2int;

	// If flash has not been programmed yet, then use default configuration.  Otherwise, load configuration from flash
	if( FGET_FACTORY_UNINITIALIZED() )
	{
		g_config.r[CREG_COM_SETTINGS-CONFIG_REG_START_ADDRESS] = (BAUD_115200 << COM_BAUD_START);		                                                                 
		g_config.r[CREG_COM_RATES1-CONFIG_REG_START_ADDRESS] = 0;
		g_config.r[CREG_COM_RATES2-CONFIG_REG_START_ADDRESS] = (20 << RATE2_ALL_RAW_START);
		g_config.r[CREG_COM_RATES3-CONFIG_REG_START_ADDRESS] = 0;
		g_config.r[CREG_COM_RATES4-CONFIG_REG_START_ADDRESS] = (20 << RATE4_ALL_PROC_START);
		g_config.r[CREG_COM_RATES5-CONFIG_REG_START_ADDRESS] = (0 << RATE5_POSITION_START) | (0 << RATE5_VELOCITY_START) | (20 << RATE5_EULER_START) | (0 << RATE5_QUAT_START);
		g_config.r[CREG_COM_RATES6-CONFIG_REG_START_ADDRESS] = (5 << RATE6_HEALTH_START);
		g_config.r[CREG_COM_RATES7-CONFIG_REG_START_ADDRESS] = ( 0 << RATE7_NMEA_HEALTH_START);
		g_config.r[CREG_MISC_SETTINGS-CONFIG_REG_START_ADDRESS] = MAG_UPDATES_ENABLED;

		g_config.f[CREG_HOME_NORTH-CONFIG_REG_START_ADDRESS] = 0.0f;
		g_config.f[CREG_HOME_EAST-CONFIG_REG_START_ADDRESS] = 0.0f;
		g_config.f[CREG_HOME_UP-CONFIG_REG_START_ADDRESS] = 0.0f;
		g_config.f[CREG_GYRO_TRIM_X-CONFIG_REG_START_ADDRESS] = 0.0f;
		g_config.f[CREG_GYRO_TRIM_Y-CONFIG_REG_START_ADDRESS] = 0.0f;
		g_config.f[CREG_GYRO_TRIM_Z-CONFIG_REG_START_ADDRESS] = 0.0f;

		g_config.f[CREG_MAG_CAL1_1-CONFIG_REG_START_ADDRESS] = 0.008f;
		g_config.f[CREG_MAG_CAL1_2-CONFIG_REG_START_ADDRESS] = 0.0f;
		g_config.f[CREG_MAG_CAL1_3-CONFIG_REG_START_ADDRESS] = 0.0f;
		g_config.f[CREG_MAG_CAL2_1-CONFIG_REG_START_ADDRESS] = 0.0f;
		g_config.f[CREG_MAG_CAL2_2-CONFIG_REG_START_ADDRESS] = 0.008f;
		g_config.f[CREG_MAG_CAL2_3-CONFIG_REG_START_ADDRESS] = 0.0f;
		g_config.f[CREG_MAG_CAL3_1-CONFIG_REG_START_ADDRESS] = 0.0f;
		g_config.f[CREG_MAG_CAL3_2-CONFIG_REG_START_ADDRESS] = 0.0f;
		g_config.f[CREG_MAG_CAL3_3-CONFIG_REG_START_ADDRESS] = 0.008f;

		g_config.f[CREG_MAG_BIAS_X-CONFIG_REG_START_ADDRESS] = 0.0f;
		g_config.f[CREG_MAG_BIAS_Y-CONFIG_REG_START_ADDRESS] = 0.0f;
		g_config.f[CREG_MAG_BIAS_Z-CONFIG_REG_START_ADDRESS] = 0.0f;
		
		// Set default hidden settings
		*g_hidden_settings.p_gyro_variance = 0.01f;
		*g_hidden_settings.p_accel_variance = 0.1f;
		*g_hidden_settings.p_mag_variance = 0.01f;
		*g_hidden_settings.p_GPS_course_variance = 10.0f; // ** Now used for yaw noise scale factor
		*g_hidden_settings.p_GPS_position_variance = 5.0f;
		*g_hidden_settings.p_GPS_velocity_variance = 1.0f;
		*g_hidden_settings.p_static_press_variance = 2.25f;
		*g_hidden_settings.p_diff_press_variance = 0.1f;

		*g_hidden_settings.p_Q_uvw = 0.1f;
		*g_hidden_settings.p_Q_quaternion = 5e-8f;
		*g_hidden_settings.p_Q_gps_position = 0.1f;
		*g_hidden_settings.p_Q_bias = 1e-7f;
		*g_hidden_settings.p_Q_euler_angles = 0.001f;
		*g_hidden_settings.p_low_vg_accel_noise_factor = 100.0f; // larger number gives less effect from accels (in Low Vg / No GPS mode update) -> smoother output, slower response
		//		g_hidden_settings.yaw_noise_factor = 1.0f; // Now using course variance for the scale factor

		//		*g_hidden_settings.p_lpf_tau_airspeed = 0.55f;
		*g_hidden_settings.p_lpf_tau_accel = 0.05f;
		*g_hidden_settings.p_lpf_tau_gyro = 0.001f;
		*g_hidden_settings.p_lpf_tau_mag = 0.05;
		//		*g_hidden_settings.p_lpf_tau_abs_pressure = 0.25;
		*g_hidden_settings.p_lpf_tau_groundspeed = 0.06;

		//		*g_hidden_settings.p_adc_raw_to_volts = 0.002539f;

		// Pressure terms
		//        *g_hidden_settings.p_rho = 1.05;

		// Gyro bias correction terms
		*g_hidden_settings.p_Cgbx0 = 0.00f;
		*g_hidden_settings.p_Cgbx1 = 0.00f;
		*g_hidden_settings.p_Cgbx2 = 0.00f;
		*g_hidden_settings.p_Cgbx3 = 0.00f;

		*g_hidden_settings.p_Cgby0 = 0.00f;
		*g_hidden_settings.p_Cgby1 = 0.00f;
		*g_hidden_settings.p_Cgby2 = 0.00f;
		*g_hidden_settings.p_Cgby3 = 0.00f;

		*g_hidden_settings.p_Cgbz0 = 0.00f;
		*g_hidden_settings.p_Cgbz1 = 0.00f;
		*g_hidden_settings.p_Cgbz2 = 0.00f;
		*g_hidden_settings.p_Cgbz3 = 0.00f;

		// Gyro scale factor correction terms
		*g_hidden_settings.p_Cgsx0 = 0.0609756f;
		*g_hidden_settings.p_Cgsx1 = 0.00f;
		*g_hidden_settings.p_Cgsx2 = 0.00f;
		*g_hidden_settings.p_Cgsx3 = 0.00f;

		*g_hidden_settings.p_Cgsy0 = 0.0609756f;
		*g_hidden_settings.p_Cgsy1 = 0.00f;
		*g_hidden_settings.p_Cgsy2 = 0.00f;
		*g_hidden_settings.p_Cgsy3 = 0.00f;

		*g_hidden_settings.p_Cgsz0 = 0.0609756f;
		*g_hidden_settings.p_Cgsz1 = 0.00f;
		*g_hidden_settings.p_Cgsz2 = 0.00f;
		*g_hidden_settings.p_Cgsz3 = 0.00f;

		// Gyro alignment matrix
		mat_create_identity(&g_hidden_settings.Mg, 3, 3);

		// Accel bias correction terms
		*g_hidden_settings.p_Cabx0 = 0.0f;
		*g_hidden_settings.p_Cabx1 = 0.00f;
		*g_hidden_settings.p_Cabx2 = 0.00f;
		*g_hidden_settings.p_Cabx3 = 0.00f;

		*g_hidden_settings.p_Caby0 = 0.0f;
		*g_hidden_settings.p_Caby1 = 0.00f;
		*g_hidden_settings.p_Caby2 = 0.00f;
		*g_hidden_settings.p_Caby3 = 0.00f;

		*g_hidden_settings.p_Cabz0 = 0.0f;
		*g_hidden_settings.p_Cabz1 = 0.00f;
		*g_hidden_settings.p_Cabz2 = 0.00f;
		*g_hidden_settings.p_Cabz3 = 0.00f;

		// Accel scale factor correction terms
		*g_hidden_settings.p_Casx0 = 0.00024414f;
		*g_hidden_settings.p_Casx1 = 0.00f;
		*g_hidden_settings.p_Casx2 = 0.00f;
		*g_hidden_settings.p_Casx3 = 0.00f;

		*g_hidden_settings.p_Casy0 = 0.00024414f;
		*g_hidden_settings.p_Casy1 = 0.00f;
		*g_hidden_settings.p_Casy2 = 0.00f;
		*g_hidden_settings.p_Casy3 = 0.00f;

		*g_hidden_settings.p_Casz0 = 0.00024414f;
		*g_hidden_settings.p_Casz1 = 0.00f;
		*g_hidden_settings.p_Casz2 = 0.00f;
		*g_hidden_settings.p_Casz3 = 0.00f;

		// Accel alignment matrix
		mat_create_identity(&g_hidden_settings.Ma, 3, 3);

		// Mag bias correction terms
		*g_hidden_settings.p_Cmbx0 = 0.00f;
		*g_hidden_settings.p_Cmbx1 = 0.00f;
		*g_hidden_settings.p_Cmbx2 = 0.00f;
		*g_hidden_settings.p_Cmbx3 = 0.00f;

		*g_hidden_settings.p_Cmby0 = 0.00f;
		*g_hidden_settings.p_Cmby1 = 0.00f;
		*g_hidden_settings.p_Cmby2 = 0.00f;
		*g_hidden_settings.p_Cmby3 = 0.00f;

		*g_hidden_settings.p_Cmbz0 = 0.00f;
		*g_hidden_settings.p_Cmbz1 = 0.00f;
		*g_hidden_settings.p_Cmbz2 = 0.00f;
		*g_hidden_settings.p_Cmbz3 = 0.00f;

		// Mag scale factor correction terms
		*g_hidden_settings.p_Cmsx0 = 1.00f;
		*g_hidden_settings.p_Cmsx1 = 0.00f;
		*g_hidden_settings.p_Cmsx2 = 0.00f;
		*g_hidden_settings.p_Cmsx3 = 0.00f;

		*g_hidden_settings.p_Cmsy0 = 1.0f;
		*g_hidden_settings.p_Cmsy1 = 0.00f;
		*g_hidden_settings.p_Cmsy2 = 0.00f;
		*g_hidden_settings.p_Cmsy3 = 0.00f;

		*g_hidden_settings.p_Cmsz0 = 1.0f;
		*g_hidden_settings.p_Cmsz1 = 0.00f;
		*g_hidden_settings.p_Cmsz2 = 0.00f;
		*g_hidden_settings.p_Cmsz3 = 0.00f;

		// Mag alignment matrix
		mat_create_identity(&g_hidden_settings.Mm, 3, 3);

		*g_hidden_settings.p_mag_ref_x = 0.136f;
		*g_hidden_settings.p_mag_ref_y = 0.4065f;
		*g_hidden_settings.p_mag_ref_z = 0.894f;
	}
	else
	{
		LoadConfigFromFLASH( CHR_USE_FACTORY_ADDRESS );
	}

	CopyConfigArrayToStates();
}

/*******************************************************************************
* Function Name  : copy_data_to_output
* Input          : None
* Output         : None
* Return         : None
* Description    : 

Copies the data stored in the g_states structure (defined in UM6_states.c)
and the g_sensor_data structure (also defined in UM6_states.c)into the g_data 
structure (defined in UM6_config.c).  This is done so that
the communication code can easily retrieve and transmit all data from the
sensor.

*******************************************************************************/
void copy_data_to_output( void )
{
	int16_t converted_data1;
	int16_t converted_data2;
	
	// Copy raw rate gyro data
	g_data.r[DREG_GYRO_RAW_XY-DATA_REG_START_ADDRESS] = (uint32_t)((g_sensor_raw.gyro_x << 16) | (g_sensor_raw.gyro_y & 0x0FFFF));
	g_data.r[DREG_GYRO_RAW_Z-DATA_REG_START_ADDRESS] = (uint32_t)(g_sensor_raw.gyro_z << 16);
	g_data.f[DREG_GYRO_TIME-DATA_REG_START_ADDRESS] = g_sensor.gyro_time;
	
	// Copy raw accel data
	g_data.r[DREG_ACCEL_RAW_XY-DATA_REG_START_ADDRESS] = (uint32_t)((g_sensor_raw.accel_x << 16) | (g_sensor_raw.accel_y & 0x0FFFF));
	g_data.r[DREG_ACCEL_RAW_Z-DATA_REG_START_ADDRESS] = (uint32_t)((g_sensor_raw.accel_z << 16));
	g_data.f[DREG_ACCEL_TIME-DATA_REG_START_ADDRESS] = g_sensor.accel_time;
	
	// Copy raw mag data
	g_data.r[DREG_MAG_RAW_XY-DATA_REG_START_ADDRESS] = (uint32_t)((g_sensor_raw.mag_x << 16) | (g_sensor_raw.mag_y & 0x0FFFF));
	g_data.r[DREG_MAG_RAW_Z-DATA_REG_START_ADDRESS] = (uint32_t)((g_sensor_raw.mag_z << 16));
	g_data.f[DREG_MAG_RAW_TIME-DATA_REG_START_ADDRESS] = g_sensor.mag_time;
	
	// Copy raw board temperature
	g_data.f[DREG_TEMPERATURE-DATA_REG_START_ADDRESS] = (g_sensor_raw.temperature);
	g_data.f[DREG_TEMPERATURE_TIME-DATA_REG_START_ADDRESS] = g_sensor_raw.temperature_time;
	
	// Copy processed rate gyro data
	g_data.f[DREG_GYRO_PROC_X-DATA_REG_START_ADDRESS] = g_sensor.gyro_x;
	g_data.f[DREG_GYRO_PROC_Y-DATA_REG_START_ADDRESS] = g_sensor.gyro_y;
	g_data.f[DREG_GYRO_PROC_Z-DATA_REG_START_ADDRESS] = g_sensor.gyro_z;
	g_data.f[DREG_GYRO_PROC_TIME-DATA_REG_START_ADDRESS] = g_sensor.gyro_time;
	
	// Copy processed accel data
	g_data.f[DREG_ACCEL_PROC_X-DATA_REG_START_ADDRESS] = g_sensor.accel_x;
	g_data.f[DREG_ACCEL_PROC_Y-DATA_REG_START_ADDRESS] = g_sensor.accel_y;
	g_data.f[DREG_ACCEL_PROC_Z-DATA_REG_START_ADDRESS] = g_sensor.accel_z;
	g_data.f[DREG_ACCEL_PROC_TIME-DATA_REG_START_ADDRESS] = g_sensor.accel_time;
	
	// Copy processed mag data 
	g_data.f[DREG_MAG_PROC_X-DATA_REG_START_ADDRESS] = g_sensor.mag_x;
	g_data.f[DREG_MAG_PROC_Y-DATA_REG_START_ADDRESS] =  g_sensor.mag_y;
	g_data.f[DREG_MAG_PROC_Z-DATA_REG_START_ADDRESS] = g_sensor.mag_z;
	g_data.f[DREG_MAG_PROC_TIME-DATA_REG_START_ADDRESS] = g_sensor.mag_time;
		
	// Quaternions
	// Quaternions are normalized, so the maximum value of any single element is 1.0.  However, let the value go to +/- 1.1 to 
	// to avoid overflow problems at 1.0
	// 2^16/2.2 = 29789.09091
	converted_data1 = (int16_t)(g_states.qib.a*29789.09091f);
	converted_data2 = (int16_t)(g_states.qib.b*29789.09091f);
	g_data.r[DREG_QUAT_AB-DATA_REG_START_ADDRESS] = (uint32_t)((converted_data1 << 16) | (converted_data2 & 0x0FFFF));
	converted_data1 = (int16_t)(g_states.qib.c*29789.09091f);
	converted_data2 = (int16_t)(g_states.qib.d*29789.09091f);
	g_data.r[DREG_QUAT_CD-DATA_REG_START_ADDRESS] = (uint32_t)((converted_data1 << 16) | (converted_data2 & 0x0FFFF));
	g_data.f[DREG_QUAT_TIME-DATA_REG_START_ADDRESS] = g_states.quat_time; //1.0f/g_ekf_time_delta;
	
	// Euler angles
	// Maximum euler angle values are +/-360 degrees (+/-2*pi).
	// 2^16/(360*2) = 91.0222222222 => 91.022222222*180/pi = 5215.189175235
	converted_data1 = (int16_t)(g_states.phi*91.0222222222f);
	converted_data2 = (int16_t)(g_states.theta*91.0222222222f);
	g_data.r[DREG_EULER_PHI_THETA-DATA_REG_START_ADDRESS] = (uint32_t)((converted_data1 << 16) | (converted_data2 & 0x0FFFF));
	converted_data1 = (int16_t)(g_states.psi*91.0222222222f);
	g_data.r[DREG_EULER_PSI-DATA_REG_START_ADDRESS] = (uint32_t)(converted_data1 << 16);
	
	// Euler rates
	// Maximum measured angular rate +/- 2000 deg/s
	// 2^16/(2000*2) = 16.384.  Round down to 16.0
	converted_data1 = (int16_t)(g_states.phi_dot*16.0f);
	converted_data2 = (int16_t)(g_states.theta_dot*16.0f);
	g_data.r[DREG_EULER_PHI_THETA_DOT-DATA_REG_START_ADDRESS] = (uint32_t)((converted_data1 << 16) | (converted_data2 & 0x0FFFF));
	converted_data1 = (int16_t)(g_states.psi_dot*16.0f);
	g_data.r[DREG_EULER_PSI_DOT-DATA_REG_START_ADDRESS] = (uint32_t)(converted_data1 << 16);
	g_data.f[DREG_EULER_TIME-DATA_REG_START_ADDRESS] = g_states.euler_time;
	
	// GPS data
	g_data.f[DREG_GPS_LATITUDE-DATA_REG_START_ADDRESS] = g_sensor_raw.GPS_latitude;
	g_data.f[DREG_GPS_LONGITUDE-DATA_REG_START_ADDRESS] = g_sensor_raw.GPS_longitude;
	g_data.f[DREG_GPS_ALTITUDE-DATA_REG_START_ADDRESS] = g_sensor_raw.GPS_altitude;
	 
	g_data.f[DREG_GPS_COURSE-DATA_REG_START_ADDRESS] = g_sensor.GPS_course*180.0f/PI;
	g_data.f[DREG_GPS_SPEED-DATA_REG_START_ADDRESS] = g_sensor.GPS_speed;
	g_data.f[DREG_GPS_TIME-DATA_REG_START_ADDRESS] = g_sensor.GPS_course_time;
    //g_data.f[DREG_GPS_DATE-DATA_REG_START_ADDRESS] = g_sensor_raw.GPS_date;
	
	// Relative Position
	g_data.f[DREG_POSITION_NORTH-DATA_REG_START_ADDRESS] = g_states.pn;
	g_data.f[DREG_POSITION_EAST-DATA_REG_START_ADDRESS] = g_states.pe;
	g_data.f[DREG_POSITION_UP-DATA_REG_START_ADDRESS] = -g_states.pd;
	g_data.f[DREG_POSITION_TIME-DATA_REG_START_ADDRESS] = g_states.position_time;
	
	// Velocity
    // Rotate from body to inertial frame now to keep output fresh:
//    compute_inertial_frame_velocities(&g_states);
	g_data.f[DREG_VELOCITY_NORTH-DATA_REG_START_ADDRESS] = g_states.vn;
	g_data.f[DREG_VELOCITY_EAST-DATA_REG_START_ADDRESS] = g_states.ve;
	g_data.f[DREG_VELOCITY_UP-DATA_REG_START_ADDRESS] = -g_states.vd;

	// Velocity Time
	g_data.f[DREG_VELOCITY_TIME-DATA_REG_START_ADDRESS] = g_states.velocity_time;
	
	// Some health information
	// Clear satellite count and Hdop values
	g_data.r[DREG_HEALTH-DATA_REG_START_ADDRESS] &= ~(HEALTH_SATS_USED_MASK << HEALTH_SATS_USED_START);
	g_data.r[DREG_HEALTH-DATA_REG_START_ADDRESS] &= ~(HEALTH_SATS_IN_VIEW_MASK << HEALTH_SATS_IN_VIEW_START);
	g_data.r[DREG_HEALTH-DATA_REG_START_ADDRESS] &= ~(HEALTH_HDOP_MASK << HEALTH_HDOP_START);
	// Copy in new satellite count and Hdop values
	g_data.r[DREG_HEALTH-DATA_REG_START_ADDRESS] |= ((g_sensor_raw.GPS_satellites_used & HEALTH_SATS_USED_MASK) << HEALTH_SATS_USED_START);
	g_data.r[DREG_HEALTH-DATA_REG_START_ADDRESS] |= ((g_sensor_raw.GPS_satellites_in_view & HEALTH_SATS_IN_VIEW_MASK) << HEALTH_SATS_IN_VIEW_START);
	g_data.r[DREG_HEALTH-DATA_REG_START_ADDRESS] |= (((uint16_t)(10*g_sensor_raw.GPS_Hdop) & HEALTH_HDOP_MASK) << HEALTH_HDOP_START);
	
	// TODO: Make sure that GPS Satellite details are transmitted
}

/*******************************************************************************
* Function Name  : UpdateSerialBaud
* Input          : None
* Output         : None
* Return         : None
* Description    : 

Updates the serial baudrate to the baudrate selected in the global configuration
array.

*******************************************************************************/
void UpdateSerialBaud()
{
	int baud_rate = 115200;
	 
	uint8_t baud_selection = (g_config.r[CREG_COM_SETTINGS] >> COM_BAUD_START) & COM_BAUD_MASK;
	 
	switch(baud_selection)
	{
		case BAUD_9600:
			baud_rate = 9600;
		break;

		case BAUD_14400:
			baud_rate = 14400;
		break;

		case BAUD_19200:
			baud_rate = 19200;
		break;

		case BAUD_38400:
			baud_rate = 38400;
		break;

		case BAUD_57600:
			baud_rate = 57600;
		break;

		case BAUD_115200:
			baud_rate = 115200;
		break;
		
		case BAUD_128000:
			baud_rate = 128000;
		break;
		
		case BAUD_153600:
			baud_rate = 153600;
		break;
		
		case BAUD_230400:
			baud_rate = 230400;
		break;
		
		case BAUD_256000:
			baud_rate = 256000;
		break;
		
		case BAUD_460800:
			baud_rate = 460800;
		break;
		
		case BAUD_921600:
			baud_rate = 921600;
		break;

		default:
			baud_rate = 115200;
		break;
	}
	
	USART1_Configuration( baud_rate );
	
	// Now configure the GPS baud rate
	baud_selection = (g_config.r[CREG_COM_SETTINGS] >> COM_GPS_BAUD_START) & COM_GPS_BAUD_MASK;

	switch(baud_selection)
	{
		case BAUD_9600:
			baud_rate = 9600;
		break;

		case BAUD_14400:
			baud_rate = 14400;
		break;

		case BAUD_19200:
			baud_rate = 19200;
		break;

		case BAUD_38400:
			baud_rate = 38400;
		break;

		case BAUD_57600:
			baud_rate = 57600;
		break;

		case BAUD_115200:
			baud_rate = 115200;
		break;

		default:
			baud_rate = 115200;
		break;
	}

	USART2_Configuration( baud_rate );
}

/*******************************************************************************
* Function Name  : CopyConfigArrayToStates
* Input          : None
* Output         : None
* Return         : None
* Description    : 

Copies certain portions of the global configuration array (g_config, defined in 
UM6_config.c) to the g_states structure (defined in UM6_states.c).  This is
done for convenience, since many of the configuration options such as gyro and
accel scale factors are actually floating point values.  The g_config array stores
them as 32-bit unsigned integers, which makes it difficult to access and use them.

This function makes it easier to access floating-point configuration values that
need to be used often by the software.

*******************************************************************************/
void CopyConfigArrayToStates( void )
{
	// Copy GPS home position info
	g_states.GPS_lat_home = g_config.f[CREG_HOME_NORTH-CONFIG_REG_START_ADDRESS];
	g_states.GPS_lon_home = g_config.f[CREG_HOME_EAST-CONFIG_REG_START_ADDRESS];
	g_states.GPS_alt_home = g_config.f[CREG_HOME_UP-CONFIG_REG_START_ADDRESS];
	
	// Copy out gyro trim values
	g_states.gyro_x_trim = g_config.f[CREG_GYRO_TRIM_X-CONFIG_REG_START_ADDRESS];
	g_states.gyro_y_trim = g_config.f[CREG_GYRO_TRIM_Y-CONFIG_REG_START_ADDRESS];
	g_states.gyro_z_trim = g_config.f[CREG_GYRO_TRIM_Z-CONFIG_REG_START_ADDRESS];

	if( (g_config.r[CREG_MISC_SETTINGS] & QUATERNION_MODE_ENABLED) == 0 )
	{
		gEKF_mode = EKF_MODE_EULER;
	}
	else
	{
		gEKF_mode = EKF_MODE_QUAT;
	}

	// Copy mag calibration matrix

    // Update the Q matrices in case any of the process noise settings were changed
//    set_Q_matrices();
}

/*******************************************************************************
* Function Name  : setup_com_events
* Input          : None
* Output         : None
* Return         : None
* Description    : 

Uses the data in the configuration registers to set up periodic broadcasts
from the sensor.

*******************************************************************************/
void setup_com_events( )
{
	int id_count = 0;
	EventProperties COM_event;
	float freq;
	
	COM_event.is_periodic = 1;
	COM_event.is_COM_event = 1;

	// Remove any existing COM events
	clear_all_events();

	// Create and add new events
	// Health packet
	if( ((g_config.r[CREG_COM_RATES6] >> RATE6_HEALTH_START) & RATE6_HEALTH_MASK ) != 0 )
	{
		COM_event.callback_func = (void*)&send_data_packet;
		COM_event.event_id = id_count++;
		
		COM_event.user_args.val1 = DREG_HEALTH;			// Address of first data register
		COM_event.user_args.val2 = ADDRESS_TYPE_DATA;	// Type of register being accessed
		COM_event.user_args.val3 = 0;					// Packet batch indicator
		COM_event.user_args.val4 = 0;					// Number of registers in the batch
		
		switch( (g_config.r[CREG_COM_RATES6] >> RATE6_HEALTH_START) & RATE6_HEALTH_MASK )
		{
			case 0:
			COM_event.period = 0.0f;
			break;
			case 1:
			COM_event.period = 8.0f;
			break;
			case 2:
			COM_event.period = 4.0f;
			break;
			case 3:
			COM_event.period = 2.0f;
			break;
			case 4:
			COM_event.period = 1.0f;
			break;
			case 5:
			COM_event.period = 0.5f;
			break;
			case 6:
			COM_event.period = 0.25f;
			break;
			default:
			COM_event.period = 1.0f;
			break;
		}
		
		add_event(&COM_event);
	}

	// Raw sensor data
	freq = (float)((g_config.r[CREG_COM_RATES2] >> RATE2_ALL_RAW_START) & RATE2_ALL_RAW_MASK);
	
	if( freq > 0 )
	{
		COM_event.callback_func = (void*)&send_data_packet;
		COM_event.event_id = id_count++;
				
		COM_event.period = 1.0f/freq;
		
		COM_event.user_args.val1 = DREG_GYRO_RAW_XY;		// Address of first data register
		COM_event.user_args.val2 = ADDRESS_TYPE_DATA;	// Type of register being accessed
		COM_event.user_args.val3 = PACKET_IS_BATCH;		// Packet batch indicator
		COM_event.user_args.val4 = 11;					// Number of registers in the batch
		
		add_event(&COM_event);
	}			
	// Only transmit individual packets with raw sensor data if all of them aren't being transmitted in a single batch
	else
	{
		// Raw gyros
		freq = (float)((g_config.r[CREG_COM_RATES1] >> RATE1_RAW_GYRO_START) & RATE1_RAW_GYRO_MASK);
		
		if( freq > 0 )
		{
			
			COM_event.callback_func = (void*)&send_data_packet;
			COM_event.event_id = id_count++;
			COM_event.period = 1.0f/freq;
		
			COM_event.user_args.val1 = DREG_GYRO_RAW_XY;		// Address of first data register
			COM_event.user_args.val2 = ADDRESS_TYPE_DATA;	// Type of register being accessed
			COM_event.user_args.val3 = PACKET_IS_BATCH;		// Packet batch indicator
			COM_event.user_args.val4 = 3;					// Number of registers in the batch
			
			add_event(&COM_event);
		}
		
		freq = (float)((g_config.r[CREG_COM_RATES1] >> RATE1_RAW_ACCEL_START) & RATE1_RAW_ACCEL_MASK);
			
		// Accelerometer Raw
		if( freq > 0 )
		{
			COM_event.callback_func = (void*)&send_data_packet;
			COM_event.event_id = id_count++;
			COM_event.period = 1.0f/freq;
		
			COM_event.user_args.val1 = DREG_ACCEL_RAW_XY;		// Address of first data register
			COM_event.user_args.val2 = ADDRESS_TYPE_DATA;	// Type of register being accessed
			COM_event.user_args.val3 = PACKET_IS_BATCH;		// Packet batch indicator
			COM_event.user_args.val4 = 3;					// Number of registers in the batch
			
			add_event(&COM_event);
		}
		
		// Magnetometer Raw
		freq = (float)((g_config.r[CREG_COM_RATES1] >> RATE1_RAW_MAG_START) & RATE1_RAW_MAG_MASK);
		
		if( freq > 0 )
		{
			COM_event.callback_func = (void*)&send_data_packet;
			COM_event.event_id = id_count++;
			COM_event.period = 1.0f/freq;
		
			COM_event.user_args.val1 = DREG_MAG_RAW_XY;		// Address of first data register
			COM_event.user_args.val2 = ADDRESS_TYPE_DATA;	// Type of register being accessed
			COM_event.user_args.val3 = PACKET_IS_BATCH;		// Packet batch indicator
			COM_event.user_args.val4 = 3;					// Number of registers in the batch
			
			add_event(&COM_event);
		}
		
		// Temperature
		freq = (float)((g_config.r[CREG_COM_RATES2] >> RATE2_TEMPERATURE_START) & RATE2_TEMPERATURE_MASK);
			
		if( freq > 0 )
		{
			COM_event.callback_func = (void*)&send_data_packet;
			COM_event.event_id = id_count++;
			COM_event.period = 1.0f/freq;
		
			COM_event.user_args.val1 = DREG_TEMPERATURE;		// Address of first data register
			COM_event.user_args.val2 = ADDRESS_TYPE_DATA;		// Type of register being accessed
			COM_event.user_args.val3 = PACKET_IS_BATCH;			// Packet batch indicator
			COM_event.user_args.val4 = 2;						// Number of registers in the batch
			
			add_event(&COM_event);
		}
	}
	
	// Processed sensor data
	freq = (float)((g_config.r[CREG_COM_RATES4] >> RATE4_ALL_PROC_START) & RATE4_ALL_PROC_MASK);
		
	if( freq > 0 )
	{
		// Have to break this into multiple packets because of the amount of data that must be sent
		COM_event.callback_func = (void*)&send_data_packet;
		COM_event.event_id = id_count++;
		COM_event.period = 1.0f/freq;
		
		COM_event.user_args.val1 = DREG_GYRO_PROC_X;	// Address of first data register
		COM_event.user_args.val2 = ADDRESS_TYPE_DATA;	// Type of register being accessed
		COM_event.user_args.val3 = PACKET_IS_BATCH;		// Packet batch indicator
		COM_event.user_args.val4 = 12;					// Number of registers in the batch
		
		add_event(&COM_event);
	}
	// Only transmit individual processed sensor data packets if all aren't being transmitted at once
	else
	{
		// Gyros
		freq = (float)((g_config.r[CREG_COM_RATES3] >> RATE3_PROC_GYRO_START) & RATE3_PROC_GYRO_MASK);
			
		if( freq > 0 )
		{
			COM_event.callback_func = (void*)&send_data_packet;
			COM_event.event_id = id_count++;
			COM_event.period = 1.0f/freq;
		
			COM_event.user_args.val1 = DREG_GYRO_PROC_X;		// Address of first data register
			COM_event.user_args.val2 = ADDRESS_TYPE_DATA;	// Type of register being accessed
			COM_event.user_args.val3 = PACKET_IS_BATCH;		// Packet batch indicator
			COM_event.user_args.val4 = 4;					// Number of registers in the batch
		
			add_event(&COM_event);
		}
		
		// Accels
		freq = (float)((g_config.r[CREG_COM_RATES3] >> RATE3_PROC_ACCEL_START) & RATE3_PROC_ACCEL_MASK);
			
		if( freq > 0 )
		{
			COM_event.callback_func = (void*)&send_data_packet;
			COM_event.event_id = id_count++;
			COM_event.period = 1.0f/freq;
		
			COM_event.user_args.val1 = DREG_ACCEL_PROC_X;		// Address of first data register
			COM_event.user_args.val2 = ADDRESS_TYPE_DATA;	// Type of register being accessed
			COM_event.user_args.val3 = PACKET_IS_BATCH;		// Packet batch indicator
			COM_event.user_args.val4 = 4;					// Number of registers in the batch
		
			add_event(&COM_event);
		}
		
		// Mag
		freq = (float)((g_config.r[CREG_COM_RATES3] >> RATE3_PROC_MAG_START) & RATE3_PROC_MAG_MASK);
			
		if( freq > 0 )
		{
			COM_event.callback_func = (void*)&send_data_packet;
			COM_event.event_id = id_count++;
			COM_event.period = 1.0f/freq;
		
			COM_event.user_args.val1 = DREG_MAG_PROC_X;		// Address of first data register
			COM_event.user_args.val2 = ADDRESS_TYPE_DATA;	// Type of register being accessed
			COM_event.user_args.val3 = PACKET_IS_BATCH;		// Packet batch indicator
			COM_event.user_args.val4 = 4;					// Number of registers in the batch
		
			add_event(&COM_event);
		}
	}
	
	// Quaternions
	freq = (float)((g_config.r[CREG_COM_RATES5] >> RATE5_QUAT_START) & RATE5_QUAT_MASK);
		
	if( freq > 0 )
	{
		COM_event.callback_func = (void*)&send_data_packet;
		COM_event.event_id = id_count++;
		COM_event.period = 1.0f/freq;
		
		COM_event.user_args.val1 = DREG_QUAT_AB;		// Address of first data register
		COM_event.user_args.val2 = ADDRESS_TYPE_DATA;	// Type of register being accessed
		COM_event.user_args.val3 = PACKET_IS_BATCH;		// Packet batch indicator
		COM_event.user_args.val4 = 3;					// Number of registers in the batch
		
		add_event(&COM_event);
	}
	
	// GPS data (NO EVENT ADDED BECAUSE NEW DATA WILL BE TRANSMITTED WHENEVER IT ARRIVES)
	
	// Pose (Euler attitude and position)
	freq = (float)((g_config.r[CREG_COM_RATES6] >> RATE6_POSE_START) & RATE6_POSE_MASK);
		
	if( freq > 0 )
	{
		COM_event.callback_func = (void*)&send_data_packet;
		COM_event.event_id = id_count++;
		COM_event.period = 1.0f/freq;
		
		COM_event.user_args.val1 = DREG_EULER_PHI_THETA;		// Address of first data register
		COM_event.user_args.val2 = ADDRESS_TYPE_DATA;	// Type of register being accessed
		COM_event.user_args.val3 = PACKET_IS_BATCH;		// Packet batch indicator
		COM_event.user_args.val4 = 9;					// Number of registers in the batch
		
		add_event(&COM_event);
	}
	else
	{
		// Euler Angles
		freq = (float)((g_config.r[CREG_COM_RATES5] >> RATE5_EULER_START) & RATE5_EULER_MASK);
		
		if( freq > 0 )
		{
			COM_event.callback_func = (void*)&send_data_packet;
			COM_event.event_id = id_count++;
			COM_event.period = 1.0f/freq;
		
			COM_event.user_args.val1 = DREG_EULER_PHI_THETA;		// Address of first data register
			COM_event.user_args.val2 = ADDRESS_TYPE_DATA;	// Type of register being accessed
			COM_event.user_args.val3 = PACKET_IS_BATCH;		// Packet batch indicator
			COM_event.user_args.val4 = 5;					// Number of registers in the batch
		
			add_event(&COM_event);
		}
	
		// Position
		freq = (float)((g_config.r[CREG_COM_RATES5] >> RATE5_POSITION_START) & RATE5_POSITION_MASK);
			
		if( freq > 0 )
		{
			COM_event.callback_func = (void*)&send_data_packet;
			COM_event.event_id = id_count++;
			COM_event.period = 1.0f/freq;
		
			COM_event.user_args.val1 = DREG_POSITION_NORTH;		// Address of first data register
			COM_event.user_args.val2 = ADDRESS_TYPE_DATA;	// Type of register being accessed
			COM_event.user_args.val3 = PACKET_IS_BATCH;		// Packet batch indicator
			COM_event.user_args.val4 = 4;					// Number of registers in the batch
		
			add_event(&COM_event);
		}
	}
	
	// Velocity
	freq = (float)((g_config.r[CREG_COM_RATES5] >> RATE5_VELOCITY_START) & RATE5_VELOCITY_MASK);
		
	if( freq > 0 )
	{
		COM_event.callback_func = (void*)&send_data_packet;
		COM_event.event_id = id_count++;
		COM_event.period = 1.0f/freq;
		
		COM_event.user_args.val1 = DREG_VELOCITY_NORTH;		// Address of first data register
		COM_event.user_args.val2 = ADDRESS_TYPE_DATA;	// Type of register being accessed
		COM_event.user_args.val3 = PACKET_IS_BATCH;		// Packet batch indicator
		COM_event.user_args.val4 = 4;					// Number of registers in the batch
		
		add_event(&COM_event);
	}		
	
	// NMEA health
	if( get_NMEA_period((g_config.r[CREG_COM_RATES7] >> RATE7_NMEA_HEALTH_START) & RATE7_NMEA_HEALTH_MASK) != 0 )
	{
		COM_event.callback_func = (void*)&send_NMEA_packet;
		COM_event.event_id = id_count++;
		COM_event.user_args.val1 = NMEA_HEALTH_PACKET;		// Type of NMEA packet
		
		COM_event.period = get_NMEA_period((g_config.r[CREG_COM_RATES7] >> RATE7_NMEA_HEALTH_START) & RATE7_NMEA_HEALTH_MASK);
		
		add_event(&COM_event);
	}
	
	// NMEA pose
	if( get_NMEA_period((g_config.r[CREG_COM_RATES7] >> RATE7_NMEA_POSE_START) & RATE7_NMEA_POSE_MASK) != 0 )
	{
		COM_event.callback_func = (void*)&send_NMEA_packet;
		COM_event.event_id = id_count++;
		COM_event.user_args.val1 = NMEA_POSE_PACKET;		// Type of NMEA packet
		
		COM_event.period = get_NMEA_period((g_config.r[CREG_COM_RATES7] >> RATE7_NMEA_POSE_START) & RATE7_NMEA_POSE_MASK);
		
		add_event(&COM_event);
	}
	
	// NMEA attitude
	if( get_NMEA_period((g_config.r[CREG_COM_RATES7] >> RATE7_NMEA_ATTITUDE_START) & RATE7_NMEA_ATTITUDE_MASK) )
	{
		COM_event.callback_func = (void*)&send_NMEA_packet;
		COM_event.event_id = id_count++;
		COM_event.user_args.val1 = NMEA_ATTITUDE_PACKET;		// Type of NMEA packet
		
		COM_event.period = get_NMEA_period((g_config.r[CREG_COM_RATES7] >> RATE7_NMEA_ATTITUDE_START) & RATE7_NMEA_ATTITUDE_MASK);
		
		add_event(&COM_event);
	}
	
	// NMEA sensor data
	if( get_NMEA_period((g_config.r[CREG_COM_RATES7] >> RATE7_NMEA_SENSOR_START) & RATE7_NMEA_SENSOR_MASK) )
	{
		COM_event.callback_func = (void*)&send_NMEA_packet;
		COM_event.event_id = id_count++;
		COM_event.user_args.val1 = NMEA_SENSOR_PACKET;		// Type of NMEA packet
		
		COM_event.period = get_NMEA_period((g_config.r[CREG_COM_RATES7] >> RATE7_NMEA_SENSOR_START) & RATE7_NMEA_SENSOR_MASK);
		
		add_event(&COM_event);
	}

	// NMEA rate data
	if( get_NMEA_period((g_config.r[CREG_COM_RATES7] >> RATE7_NMEA_RATE_START) & RATE7_NMEA_RATE_MASK) )
	{
		COM_event.callback_func = (void*)&send_NMEA_packet;
		COM_event.event_id = id_count++;
		COM_event.user_args.val1 = NMEA_RATE_PACKET;		// Type of NMEA packet
		
		COM_event.period = get_NMEA_period((g_config.r[CREG_COM_RATES7] >> RATE7_NMEA_RATE_START) & RATE7_NMEA_RATE_MASK);
		
		add_event(&COM_event);
	}

	// NMEA rate data
	if( get_NMEA_period((g_config.r[CREG_COM_RATES7] >> RATE7_NMEA_GPS_POSE_START) & RATE7_NMEA_GPS_POSE_MASK) )
	{
		COM_event.callback_func = (void*)&send_NMEA_packet;
		COM_event.event_id = id_count++;
		COM_event.user_args.val1 = NMEA_GPS_POSE_PACKET;		// Type of NMEA packet
		
		COM_event.period = get_NMEA_period((g_config.r[CREG_COM_RATES7] >> RATE7_NMEA_GPS_POSE_START) & RATE7_NMEA_GPS_POSE_MASK);
		
		add_event(&COM_event);
	}

	// NMEA quaternion data
	if( get_NMEA_period((g_config.r[CREG_COM_RATES7] >> RATE7_NMEA_QUATERNION_START) & RATE7_NMEA_QUATERNION_MASK) )
	{
		COM_event.callback_func = (void*)&send_NMEA_packet;
		COM_event.event_id = id_count++;
		COM_event.user_args.val1 = NMEA_QUATERNION_PACKET;		// Type of NMEA packet
		
		COM_event.period = get_NMEA_period((g_config.r[CREG_COM_RATES7] >> RATE7_NMEA_QUATERNION_START) & RATE7_NMEA_QUATERNION_MASK);
		
		add_event(&COM_event);
	}
}

/*******************************************************************************
* Function Name  : send_data_packet
* Input          : None
* Output         : None
* Return         : None
* Description    : 

Sends a data packet containing data specified in args.
						 
*******************************************************************************/
void send_data_packet( uint32_t system_time, UserArgs args )
{
	SendGlobalData((uint8_t)args.val1, (uint8_t)args.val2, (uint8_t)args.val3, (uint8_t)args.val4);
}

/*******************************************************************************
* Function Name  : send_NMEA_packet
* Input          : None
* Output         : None
* Return         : None
* Description    : 

Send a NMEA human-readable ascii packet.
						 
*******************************************************************************/
void send_NMEA_packet( uint32_t system_time, UserArgs args )
{
	if( args.val1 == NMEA_HEALTH_PACKET )
	{
		send_NMEA_health_packet();
	}
	else if( args.val1 == NMEA_POSE_PACKET )
	{
		send_NMEA_pose_packet();
	}
	else if( args.val1 == NMEA_ATTITUDE_PACKET )
	{
		send_NMEA_attitude_packet();
	}
	else if( args.val1 == NMEA_SENSOR_PACKET )
	{
		send_NMEA_sensor_packet();
	}
	else if( args.val1 == NMEA_RATE_PACKET )
	{
		send_NMEA_rate_packet();
	}
	else if( args.val1 == NMEA_GPS_POSE_PACKET )
	{
		send_NMEA_GPS_pose_packet();
	}
	else if( args.val1 == NMEA_QUATERNION_PACKET )
	{
		send_NMEA_quaternion_packet();
	}
}

/*******************************************************************************
* Function Name  : send_NMEA_sensor_packet
* Input          : None
* Output         : None
* Return         : None
* Description    : 

Send a NMEA human-readable ascii packet.
						 
*******************************************************************************/
void send_NMEA_sensor_packet()
{
	float current_time;
	USARTPacket rate;

	rate.is_ASCII = 1;

	current_time = get_UTC_time();

	// Make local pointer to packet buffer for convenience
	char* buf = (char*)rate.packet_data;
	uint8_t index = 0;

	// Write the packet header
	buf[index++] = '$';
	buf[index++] = 'P';
	buf[index++] = 'C';
	buf[index++] = 'H';
	buf[index++] = 'R';
	buf[index++] = 'S';

	buf[index++] = ',';
	
	// We'll send a group of 3 total packets to get all the sensor data out
	// Start with rate gyros
	// Packet ID
	buf[index++] = '0';
	buf[index++] = ',';
	
	// Add the system time
	uint32_t bytes_written = print_float( current_time, 3, &buf[index], 13 );
	index += bytes_written;
	buf[index++] = ',';

	// Gyro data
	bytes_written = print_float( g_sensor.gyro_x, 2, &buf[index], 9 );
	index += bytes_written;
	buf[index++] = ',';
	
	bytes_written = print_float( g_sensor.gyro_y, 2, &buf[index], 9 );
	index += bytes_written;
	buf[index++] = ',';
	
	bytes_written = print_float( g_sensor.gyro_z, 2, &buf[index], 9 );
	index += bytes_written;
	buf[index++] = ',';
	
	// Packet finished.  Write the checksum to it and set the data length
	close_NMEA_packet( buf, index, &rate );
	
	SendTXPacketSafe( &rate );
	
	// ------------------------------------------------------------
	// Go back and do this again for accel data
	index = 7;
	
	// Packet ID
	buf[index++] = '1';
	buf[index++] = ',';
	
	// Add the system time
	bytes_written = print_float( current_time, 3, &buf[index], 13 );
	index += bytes_written;
	buf[index++] = ',';

	// Accel data
	bytes_written = print_float( g_sensor.accel_x, 4, &buf[index], 11 );
	index += bytes_written;
	buf[index++] = ',';
	
	bytes_written = print_float( g_sensor.accel_y, 4, &buf[index], 11 );
	index += bytes_written;
	buf[index++] = ',';
	
	bytes_written = print_float( g_sensor.accel_z, 4, &buf[index], 11 );
	index += bytes_written;
	buf[index++] = ',';
	
	// Packet finished.  Write the checksum to it and set the data length
	close_NMEA_packet( buf, index, &rate );
	
	SendTXPacketSafe( &rate );
	
	// ------------------------------------------------------------
	// One last time for magnetometer data
	index = 7;
	
	// Packet ID
	buf[index++] = '2';
	buf[index++] = ',';
	
	// Add the system time
	bytes_written = print_float( current_time, 3, &buf[index], 13 );
	index += bytes_written;
	buf[index++] = ',';
	
	// Mag data
	bytes_written = print_float( g_sensor.mag_x, 4, &buf[index], 11 );
	index += bytes_written;
	buf[index++] = ',';
	
	bytes_written = print_float( g_sensor.mag_y, 4, &buf[index], 11 );
	index += bytes_written;
	buf[index++] = ',';
	
	bytes_written = print_float( g_sensor.mag_z, 4, &buf[index], 11 );
	index += bytes_written;
	buf[index++] = ',';
	
	// Packet finished.  Write the checksum to it and set the data length
	close_NMEA_packet( buf, index, &rate );
	
	SendTXPacketSafe( &rate );
}

/*******************************************************************************
* Function Name  : send_NMEA_rate_packet
* Input          : None
* Output         : None
* Return         : None
* Description    : 

Send a NMEA human-readable ascii packet.
						 
*******************************************************************************/
void send_NMEA_rate_packet()
{
	USARTPacket rate;

	rate.is_ASCII = 1;

	// Make local pointer to packet buffer for convenience
	char* buf = (char*)rate.packet_data;
	uint8_t index = 0;

	// Write the packet header
	buf[index++] = '$';
	buf[index++] = 'P';
	buf[index++] = 'C';
	buf[index++] = 'H';
	buf[index++] = 'R';
	buf[index++] = 'R';

	buf[index++] = ',';

	// Add the system time
	uint32_t bytes_written = print_float( get_UTC_time(), 3, &buf[index], 13 );
	index += bytes_written;
	buf[index++] = ',';

	// North velocity
	bytes_written = print_float( g_states.vn, 2, &buf[index], 7 );
	index += bytes_written;
	buf[index++] = ',';
	
	// East velocity
	bytes_written = print_float( g_states.ve, 2, &buf[index], 7 );
	index += bytes_written;
	buf[index++] = ',';
	
	// Altitude velocity
	bytes_written = print_float( -g_states.vd, 2, &buf[index], 7 );
	index += bytes_written;
	buf[index++] = ',';
	
	// Phi dot
	bytes_written = print_float( g_states.phi_dot, 2, &buf[index], 8 );
	index += bytes_written;
	buf[index++] = ',';
	
	// Theta dot
	bytes_written = print_float( g_states.theta_dot, 2, &buf[index], 8 );
	index += bytes_written;
	buf[index++] = ',';
	
	// Psi dot
	bytes_written = print_float( g_states.psi_dot, 2, &buf[index], 8 );
	index += bytes_written;
	buf[index++] = ',';
	
	// Packet finished.  Write the checksum to it and set the data length
	close_NMEA_packet( buf, index, &rate );
	
	SendTXPacketSafe( &rate );
}

/*******************************************************************************
* Function Name  : send_NMEA_GPS_pose_packet
* Input          : None
* Output         : None
* Return         : None
* Description    : 

Send a NMEA human-readable ascii packet.
						 
*******************************************************************************/
void send_NMEA_GPS_pose_packet()
{
	USARTPacket pose;

	pose.is_ASCII = 1;

	// Make local pointer to packet buffer for convenience
	char* buf = (char*)pose.packet_data;
	uint8_t index = 0;

	// Write the packet header
	buf[index++] = '$';
	buf[index++] = 'P';
	buf[index++] = 'C';
	buf[index++] = 'H';
	buf[index++] = 'R';
	buf[index++] = 'G';

	buf[index++] = ',';

	// Add the system time
	uint32_t bytes_written = print_float( get_UTC_time(), 3, &buf[index], 13 );
	index += bytes_written;
	buf[index++] = ',';

	// Add latitude
	bytes_written = print_float( g_sensor_raw.GPS_latitude, 6, &buf[index], 11 );
	index += bytes_written;
	buf[index++] = ',';
	
	// Add longitude
	bytes_written = print_float( g_sensor_raw.GPS_longitude, 6, &buf[index], 11 );
	index += bytes_written;
	buf[index++] = ',';
	
	// Add altitude
	bytes_written = print_float( g_sensor_raw.GPS_altitude, 3, &buf[index], 11 );
	index += bytes_written;
	buf[index++] = ',';
	
	// Add roll angle
	bytes_written = print_float( g_states.phi, 2, &buf[index], 7 );
	index += bytes_written;
	buf[index++] = ',';
	
	// Add pitch angle
	bytes_written = print_float( g_states.theta, 2, &buf[index], 7 );
	index += bytes_written;
	buf[index++] = ',';
	
	// Add yaw angle
	bytes_written = print_float( g_states.psi, 2, &buf[index], 7 );
	index += bytes_written;
	buf[index++] = ',';
	
	// TODO: Add GPS heading angle
	
	// Packet finished.  Write the checksum to it and set the data length
	close_NMEA_packet( buf, index, &pose );
	
	SendTXPacketSafe( &pose );
}

/*******************************************************************************
* Function Name  : send_NMEA_quaternion_packet
* Input          : None
* Output         : None
* Return         : None
* Description    : 

Send a NMEA human-readable ascii packet.
						 
*******************************************************************************/
void send_NMEA_quaternion_packet()
{
	USARTPacket attitude;

	attitude.is_ASCII = 1;

	// Make local pointer to packet buffer for convenience
	char* buf = (char*)attitude.packet_data;
	uint8_t index = 0;

	// Write the packet header
	buf[index++] = '$';
	buf[index++] = 'P';
	buf[index++] = 'C';
	buf[index++] = 'H';
	buf[index++] = 'R';
	buf[index++] = 'Q';

	buf[index++] = ',';

	// Add the system time
	uint32_t bytes_written = print_float( get_UTC_time(), 3, &buf[index], 13 );
	index += bytes_written;
	buf[index++] = ',';
	
	// Add quaternion a
	bytes_written = print_float( g_states.qib.a, 5, &buf[index], 8 );
	index += bytes_written;
	buf[index++] = ',';
	
	// Add quaternion b
	bytes_written = print_float( g_states.qib.b, 5, &buf[index], 8 );
	index += bytes_written;
	buf[index++] = ',';
	
	// Add quaternion c
	bytes_written = print_float( g_states.qib.c, 5, &buf[index], 8 );
	index += bytes_written;
	buf[index++] = ',';
	
	// Add quaternion d
	bytes_written = print_float( g_states.qib.d, 5, &buf[index], 8 );
	index += bytes_written;
	buf[index++] = ',';
	
	// Packet finished.  Write the checksum to it and set the data length
	close_NMEA_packet( buf, index, &attitude );
	
	SendTXPacketSafe( &attitude );
}

/*******************************************************************************
* Function Name  : send_NMEA_attitude_packet
* Input          : None
* Output         : None
* Return         : None
* Description    : 

Send a NMEA human-readable ascii packet.
						 
*******************************************************************************/
void send_NMEA_attitude_packet()
{
	USARTPacket attitude;

	attitude.is_ASCII = 1;

	// Make local pointer to packet buffer for convenience
	char* buf = (char*)attitude.packet_data;
	uint8_t index = 0;

	// Write the packet header
	buf[index++] = '$';
	buf[index++] = 'P';
	buf[index++] = 'C';
	buf[index++] = 'H';
	buf[index++] = 'R';
	buf[index++] = 'A';

	buf[index++] = ',';

	// Add the system time
	uint32_t bytes_written = print_float( get_UTC_time(), 3, &buf[index], 13 );
	index += bytes_written;
	buf[index++] = ',';
	
	// Add roll angle
	bytes_written = print_float( g_states.phi, 2, &buf[index], 7 );
	index += bytes_written;
	buf[index++] = ',';
	
	// Add pitch angle
	bytes_written = print_float( g_states.theta, 2, &buf[index], 7 );
	index += bytes_written;
	buf[index++] = ',';
	
	// Add yaw angle
	bytes_written = print_float( g_states.psi, 2, &buf[index], 7 );
	index += bytes_written;
	buf[index++] = ',';
	
	// TODO: Add heading angle
	
	// Packet finished.  Write the checksum to it and set the data length
	close_NMEA_packet( buf, index, &attitude );
	
	SendTXPacketSafe( &attitude );
}

/*******************************************************************************
* Function Name  : send_NMEA_pose_packet
* Input          : None
* Output         : None
* Return         : None
* Description    : 

Send a NMEA human-readable ascii packet.
						 
*******************************************************************************/
void send_NMEA_pose_packet()
{
	USARTPacket pose;

	pose.is_ASCII = 1;

	// Make local pointer to packet buffer for convenience
	char* buf = (char*)pose.packet_data;
	uint8_t index = 0;

	// Write the packet header
	buf[index++] = '$';
	buf[index++] = 'P';
	buf[index++] = 'C';
	buf[index++] = 'H';
	buf[index++] = 'R';
	buf[index++] = 'P';

	buf[index++] = ',';

	// Add the system time
	uint32_t bytes_written = print_float( get_UTC_time(), 3, &buf[index], 13 );
	index += bytes_written;
	buf[index++] = ',';

	// Add north position
	bytes_written = print_float( g_states.pn, 3, &buf[index], 10 );
	index += bytes_written;
	buf[index++] = ',';
	
	// Add east position
	bytes_written = print_float( g_states.pe, 3, &buf[index], 10 );
	index += bytes_written;
	buf[index++] = ',';
	
	// Add up position
	bytes_written = print_float( -g_states.pd, 3, &buf[index], 10 );
	index += bytes_written;
	buf[index++] = ',';
	
	// Add roll angle
	bytes_written = print_float( g_states.phi, 2, &buf[index], 7 );
	index += bytes_written;
	buf[index++] = ',';
	
	// Add pitch angle
	bytes_written = print_float( g_states.theta, 2, &buf[index], 7 );
	index += bytes_written;
	buf[index++] = ',';
	
	// Add yaw angle
	bytes_written = print_float( g_states.psi, 2, &buf[index], 7 );
	index += bytes_written;
	buf[index++] = ',';
	
	// TODO: Add GPS heading
	
	// Compute the checksum.  Skip the $ character per NMEA spec.
	int chk_index;
	uint8_t checksum = 0;
	for( chk_index = 1; chk_index < index; chk_index++ )
	{
		checksum ^= buf[chk_index];
	}

	// Packet finished.  Write the checksum to it and set the data length
	close_NMEA_packet( buf, index, &pose );

	SendTXPacketSafe( &pose );
}

/*******************************************************************************
* Function Name  : send_NMEA_health_packet
* Input          : None
* Output         : None
* Return         : None
* Description    : 

Send a NMEA human-readable ascii packet.
						 
*******************************************************************************/
void send_NMEA_health_packet()
{
	USARTPacket health;

	health.is_ASCII = 1;

	// Make local pointer to packet buffer for convenience
	char* buf = (char*)health.packet_data;
	uint8_t index = 0;

	uint8_t COM_overflow = 0;
	uint8_t accel_fault = 0;
	uint8_t gyro_fault = 0;
	uint8_t mag_fault = 0;
	uint8_t GPS_fault = 0;

	if( g_data.r[DREG_HEALTH-DATA_REG_START_ADDRESS] & HEALTH_COM_OVERFLOW )
	{
		COM_overflow = 1;
	}

	if( g_data.r[DREG_HEALTH-DATA_REG_START_ADDRESS] & HEALTH_ACCEL )
	{
		accel_fault = 1;
	}

	if( g_data.r[DREG_HEALTH-DATA_REG_START_ADDRESS] & HEALTH_GYRO )
	{
		gyro_fault = 1;
	}

	if( g_data.r[DREG_HEALTH-DATA_REG_START_ADDRESS] & HEALTH_MAG )
	{
		mag_fault = 1;
	}

	if( g_data.r[DREG_HEALTH-DATA_REG_START_ADDRESS] & HEALTH_GPS )
	{
		GPS_fault = 1;
	}
	
	// Write the packet header
	buf[index++] = '$';
	buf[index++] = 'P';
	buf[index++] = 'C';
	buf[index++] = 'H';
	buf[index++] = 'R';
	buf[index++] = 'H';

	buf[index++] = ',';

	// Add the system time
	uint32_t bytes_written = print_float( get_UTC_time(), 3, &buf[index], 13 );
	index += bytes_written;
	buf[index++] = ',';

	// Add satellites used
	bytes_written = print_unsigned( g_sensor_raw.GPS_satellites_used, 2, &buf[index], 2 );
	index += bytes_written;
	buf[index++] = ',';

	// Add satellites in view
	bytes_written = print_unsigned( g_sensor_raw.GPS_satellites_in_view, 2, &buf[index], 2 );
	index += bytes_written;
	buf[index++] = ',';
	
	// Add HDOP
	bytes_written = print_float( g_sensor_raw.GPS_Hdop, 1, &buf[index], 3 );
	index += bytes_written;
	buf[index++] = ',';
	
	// Add mode (0 for Euler Angle mode, 1 for Quaternion mode)
	if( (g_config.r[CREG_MISC_SETTINGS] & QUATERNION_MODE_ENABLED) == 0 )
	{
		buf[index++] = '0';
	}
	else
	{
		buf[index++] = '1';
	}
	buf[index++] = ',';
	
	// Add COM Overflow flag
	bytes_written = print_unsigned( COM_overflow, 0, &buf[index], 1 );
	index += bytes_written;
	buf[index++] = ',';
	
	// Add accel fault
	bytes_written = print_unsigned( accel_fault, 0, &buf[index], 1 );
	index += bytes_written;
	buf[index++] = ',';
	
	// Add gyro fault
	bytes_written = print_unsigned( gyro_fault, 0, &buf[index], 1 );
	index += bytes_written;
	buf[index++] = ',';
	
	// Add mag fault
	bytes_written = print_unsigned( mag_fault, 0, &buf[index], 1 );
	index += bytes_written;
	buf[index++] = ',';
	
	// Add GPS fault
	bytes_written = print_unsigned( GPS_fault, 0, &buf[index], 1 );
	index += bytes_written;
	buf[index++] = ',';
	
	// Add placeholder for abs pressure fault
	buf[index++] = '0';
	buf[index++] = ',';
	
	// Add placeholder for diff pressure fault
	buf[index++] = '0';
	buf[index++] = ',';
	
	// Add placeholder for future fault
	buf[index++] = '0';
	buf[index++] = ',';
	
	// Packet finished.  Write the checksum to it and set the data length
	close_NMEA_packet( buf, index, &health );
	
	SendTXPacketSafe( &health );
}

/*******************************************************************************
* Function Name  : close_NMEA_packet
* Input          : None
* Output         : None
* Return         : None
* Description    : 

Adds the final components of the NMEA packet.  We do this a lot, so this function
call was added to stop duplicating code.
						 
*******************************************************************************/
void close_NMEA_packet( char* buf, uint8_t index, USARTPacket* packet )
{
	// Compute the checksum.  Skip the $ character per NMEA spec.
	int chk_index;
	uint8_t checksum = 0;
	for( chk_index = 1; chk_index < index; chk_index++ )
	{
		checksum ^= buf[chk_index];
	}

	// Add the checksum
	buf[index++] = '*';
	uint32_t bytes_written = print_hex( (uint32_t)checksum, 2, &buf[index], 2 );
	index += bytes_written;

	// Add terminatin characters and transmit the packet
	buf[index++] = '\r';
	buf[index++] = '\n';

	packet->data_length = index;
}

/*******************************************************************************
* Function Name  : get_NMEA_period
* Input          : Output
* None         : None
* Return         : None
* Description    : 

For the given input setting, extracts the NMEA period governing length of time
between packet transmissions.  The setting corresponds to the data stored in
the COM rates register for NMEA packets.
						 
*******************************************************************************/
float get_NMEA_period( uint8_t period_setting )
{
	float period;
	
	if( period_setting == 0 )		// Off
	{
		period = 0;
	}
	else if( period_setting == 1 )	// 1 Hz
	{
		period = 1.0f;
	}
	else if( period_setting == 2 )	// 2 Hz
	{
		period = 0.5f;
	}
	else if( period_setting == 3 )	// 4 Hz
	{
		period = 0.25f;
	}
	else if( period_setting == 4 )	// 5 Hz
	{
		period = 0.2f;
	}
	else if( period_setting == 5 )	// 10 Hz
	{
		period = 0.1f;
	}
	else if( period_setting == 6 )	// 15 Hz
	{
		period = 0.06667f;
	}
	else if( period_setting == 7 )	// 20 Hz
	{
		period = 0.05f;
	}
	else if( period_setting == 8 )	// 30 Hz
	{
		period = 0.03333f;
	}
	else if( period_setting == 9 ) 	// 40 Hz
	{
		period = 0.025f;
	}
	else if( period_setting == 10 ) // 50 Hz
	{
		period = 0.02f;
	}
	else if( period_setting == 11 ) // 60 Hz
	{
		period = 0.016667f;
	}
	else if( period_setting == 12 ) // 70 Hz
	{
		period = 0.0142857f;
	}
	else if( period_setting == 13 ) // 80 Hz
	{
		period = 0.0125f;
	}
	else if( period_setting == 14 ) // 90 Hz
	{
		period = 0.0111f;
	}
	else if( period_setting == 15 ) // 100 Hz
	{
		period = 0.01f;
	}
	else
	{
		period = 1.0f;
	}
	
	return period;
}

/*******************************************************************************
* Function Name  : start_gyro_zeroing
* Input          : 
* Output		:
* None         : None
* Return         : None
* Description    : 

Sets flag to enable gyro zeroing

*******************************************************************************/
void start_gyro_zeroing( void )
{
	g_gyro_samples_collected = 0;
	g_gyro_sum[0] = 0.0f;
	g_gyro_sum[1] = 0.0f;
	g_gyro_sum[2] = 0.0f;
	
	g_zeroing_gyros = 1;
}

/*******************************************************************************
* Function Name  : stop_gyro_zeroing
* Input          : 
* Output		:
* None         : None
* Return         : None
* Description    : 

Stop the gyro zeroing operation and copies new trim values to the output
configuration
						 
*******************************************************************************/
void stop_gyro_zeroing( void )
{
	fConvert i2f;
	float gyro_x_avg, gyro_y_avg, gyro_z_avg;
	
	g_zeroing_gyros = 0;
	
	gyro_x_avg = g_gyro_sum[GYRO_X_INDEX]/(float)GYRO_ZERO_SAMPLES;
	gyro_y_avg = g_gyro_sum[GYRO_Y_INDEX]/(float)GYRO_ZERO_SAMPLES;
	gyro_z_avg = g_gyro_sum[GYRO_Z_INDEX]/(float)GYRO_ZERO_SAMPLES;
	
	// Copy to sensor array for future use
	g_states.gyro_x_trim = gyro_x_avg;
	g_states.gyro_y_trim = gyro_y_avg;
	g_states.gyro_z_trim = gyro_z_avg;
	
	// Copy to global configuration array
	i2f.float_val = g_states.gyro_x_trim;
	g_config.r[CREG_GYRO_TRIM_X-CONFIG_REG_START_ADDRESS] = i2f.uint32_val;
	
	i2f.float_val = g_states.gyro_y_trim;
	g_config.r[CREG_GYRO_TRIM_Y-CONFIG_REG_START_ADDRESS] = i2f.uint32_val;
	
	i2f.float_val = g_states.gyro_z_trim;
	g_config.r[CREG_GYRO_TRIM_Z-CONFIG_REG_START_ADDRESS] = i2f.uint32_val;

	// Send a packet with the new gyro trim values
	SendGlobalData(CREG_GYRO_TRIM_X,ADDRESS_TYPE_CONFIG,PACKET_IS_BATCH,3);
}

/*******************************************************************************
* Function Name  : configure_tx2_pin
* Input          : 
* Output		:
* None         : None
* Return         : None
* Description    : 

Uses global settings to configure the TX2 pin as either a push-pull output or
an input pin with internal pullup.  This is set up so that we can connect an
external GPS module to the UM7 and use the TX2 pin as the PPS input for time
synchronization.
						 
*******************************************************************************/
void configure_tx2_pin( void )
{
	NVIC_InitTypeDef NVIC_InitStructure;
	EXTI_InitTypeDef EXTI_InitStruct;
	GPIO_InitTypeDef GPIO_InitStructure;

	// Set up interrupts
	EXTI_InitStruct.EXTI_Line = EXTI_Line2;
	EXTI_InitStruct.EXTI_Mode = EXTI_Mode_Interrupt;
	EXTI_InitStruct.EXTI_Trigger = EXTI_Trigger_Rising;

	NVIC_InitStructure.NVIC_IRQChannel = EXTI2_IRQn;
    NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 1;
    NVIC_InitStructure.NVIC_IRQChannelSubPriority = 1;

	// Check to see whether we are setting up the PIN as PPS input or not
	if( g_config.r[CREG_MISC_SETTINGS] & USE_TX2_AS_PPS_INPUT )
	{
		// Configure USART2 Tx as input with internal pull-down
		GPIO_InitStructure.GPIO_Pin = GPIO_Pin_2;
		GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
		GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IPD;
		GPIO_Init(GPIOA, &GPIO_InitStructure);
		
		EXTI_InitStruct.EXTI_LineCmd = ENABLE;
		NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
	}
	else
	{
		// Configure USART2 Tx as alternate function push-pull
		GPIO_InitStructure.GPIO_Pin = GPIO_Pin_2;
		GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
		GPIO_InitStructure.GPIO_Mode = GPIO_Mode_AF_PP;
		GPIO_Init(GPIOA, &GPIO_InitStructure);
	
		EXTI_InitStruct.EXTI_LineCmd = DISABLE;
		NVIC_InitStructure.NVIC_IRQChannelCmd = DISABLE;
	}
	
	NVIC_Init(&NVIC_InitStructure);
	EXTI_Init(&EXTI_InitStruct);
}

/*******************************************************************************
* Function Name  : EXTI2_IRQHandler
* Input          : 
* Output		:
* None         : None
* Return         : None
* Description    : 

When the USART TX2 pin is configured as a PPS input, this interrupt handler
is called whenever the PPS pin rises.  This is used to calibrate the system clock
so that it matches UTC time of day.
						 
*******************************************************************************/
void EXTI2_IRQHandler( void )
{
	static LED_state = 1;

	// For now, toggle the LED so that we know that the interrupt handler is
	// being called.
	if( LED_state )
	{
		GPIO_WriteBit(GPIOA, GPIO_Pin_8, Bit_SET);
		LED_state = 0;
	}
	else
	{
		GPIO_WriteBit(GPIOA, GPIO_Pin_8, Bit_RESET);
		LED_state = 1;
	}

	// Get the current uncalibrated system time
	// Store the current system time
    float time = get_system_time_uncal();

    // Store the system time in the calibration structure
    // Wrapping g_UTC_timing.t_sys_k_d1 is handled when the system time is wrapped
    g_UTC_timing.t_sys_k = (double)time;
    g_PPS_received = 1;

	EXTI_ClearITPendingBit(EXTI_Line2);
}
