/* ------------------------------------------------------------------------------
  File: UM6_states.c
  Author: CH Robotics
  Version: 1.0
  
  Description: Function definitions for CHR-6dm state estimation.
------------------------------------------------------------------------------ */ 

#include <math.h>
#include "UM6_states.h"
#include "UM6_config.h"
#include "CHR_matrix.h"
#include "CHR_quat.h"
#include "EventScheduler.h"

// Data structures for holding sensor data and estimated states.
RawSensorData g_sensor_raw;
ProcessedSensorData g_sensor;
StateData g_states;

uint8_t gEKF_mode;

/*******************************************************************************
* Function Name  : EKF_Init
* Input          : None
* Output         : Pre-initialized state estimate structure
* Return         : None
* Description    : Fills an AHRS_states structure with zero initial values.
*******************************************************************************/
void EKF_Init( StateData* states )
{
	states->initialized = 1;

	states->phi = 0;
	states->theta = 0;
	states->psi = 0;

	states->phi_dot = 0;
	states->theta_dot = 0;
	states->psi_dot = 0;

	states->qib.a = 1;
	states->qib.b = 0;
	states->qib.c = 0;
	states->qib.d = 0;
	
	// Assign pointers used to store matrix data
	states->P.data = states->P_data;
	states->Q.data = states->Q_data;
	states->Q_input.data = states->Q_input_data;

	// Also for mag data
	states->Msoft.data = &g_config.f[CREG_MAG_CAL1_1];
	states->Msoft.rows = 3;
	states->Msoft.columns = 3;
	states->p_mag_x_bias = &g_config.f[CREG_MAG_BIAS_X];
	states->p_mag_y_bias = &g_config.f[CREG_MAG_BIAS_Y];
	states->p_mag_z_bias = &g_config.f[CREG_MAG_BIAS_Z];

	if( gEKF_mode == EKF_MODE_EULER )
	{
		mat_zero( &states->Q, 3, 3 );
		mat_zero( &states->P, 3, 3 );

		states->P.data[0*3 + 0] = *g_hidden_settings.p_Q_euler_angles;
		states->P.data[1*3 + 1] = *g_hidden_settings.p_Q_euler_angles;
		states->P.data[2*3 + 2] = *g_hidden_settings.p_Q_euler_angles;

		states->Q.data[0*3 + 0] = *g_hidden_settings.p_Q_euler_angles;
		states->Q.data[1*3 + 1] = *g_hidden_settings.p_Q_euler_angles;
		states->Q.data[2*3 + 2] = *g_hidden_settings.p_Q_euler_angles;
	}
	else
	{
		  mat_zero( &states->Q, 4, 4 );
		  mat_zero( &states->P, 4, 4 );
		  
		  states->P.data[0*4 + 0] = *g_hidden_settings.p_Q_quaternion;
		  states->P.data[1*4 + 1] = *g_hidden_settings.p_Q_quaternion;
		  states->P.data[2*4 + 2] = *g_hidden_settings.p_Q_quaternion;
		  states->P.data[3*4 + 3] = *g_hidden_settings.p_Q_quaternion;
		
		  states->Q.data[0*4 + 0] = *g_hidden_settings.p_Q_quaternion;
		  states->Q.data[1*4 + 1] = *g_hidden_settings.p_Q_quaternion;
		  states->Q.data[2*4 + 2] = *g_hidden_settings.p_Q_quaternion;
		  states->Q.data[3*4 + 3] = *g_hidden_settings.p_Q_quaternion;
	 }
	 
	 CopyConfigArrayToStates();
}

/*******************************************************************************
* Function Name  : EKF_InitFromSensors
* Input          : AHRS_states* states, RawSensorData* sensor_data
* Output         : None
* Return         : None
* Description    : Uses sensor data to fill initial state estimate structure.
						 Specifically, accels are used to calculate an initial pitch and
						 roll guess, while the magnetomer is used to compute an initial
						 heading guess.
						 Gyro outputs are copied directly into estimated angular rate states.
*******************************************************************************/
void EKF_InitFromSensors( StateData* states, ProcessedSensorData* sensor_data )
{
	 int i, j;
	 
	 // TODO: Write this function.  For now, initialization just calls EKF_Init(.)
	 EKF_Init( states );
}

/*******************************************************************************
* Function Name  : process_sensor_data
* Input          : RawSensorData* raw_data, ProcessedSensorData* processed_data
* Output         : None
* Return         : None
* Description    : 

Converts all raw sensor data to processed data for state estimation

*******************************************************************************/
void process_sensor_data( RawSensorData* raw_data, ProcessedSensorData* processed_data )
{
	// Handle new rate gyro data if it is available
	if( raw_data->new_gyro_data == 1 )
	{		
		process_gyro_data(raw_data,processed_data);
	}
	
	// Handle new accelerometer data if it is available
	if( raw_data->new_accel_data == 1 )
	{		
		process_accel_data(raw_data,processed_data);
	
		float accel_norm = sqrt(processed_data->accel_x*processed_data->accel_x + processed_data->accel_y*processed_data->accel_y + processed_data->accel_z*processed_data->accel_z);
				
		// Is the accelerometer measurement close to 1G?  If so, clear the accel norm flag in the health register to indicate that we are using the accel
		if( fabs(accel_norm - 1.0f) < ACCEL_NORM_THRESHOLD )
		{
			g_data.r[DREG_HEALTH-DATA_REG_START_ADDRESS] &= ~HEALTH_ACCEL_NORM;
		}
		// If the measurement is too far away from 1G, then clear the new_accel_data flag so that the estimator won't use it during update.
		else
		{
			g_data.r[DREG_HEALTH-DATA_REG_START_ADDRESS] |= HEALTH_ACCEL_NORM;
			processed_data->new_accel_data = 0;
		}
	}
	
	// Magnetometer data
	if (raw_data->new_mag_data == 1)
	{
		process_mag_data(raw_data,processed_data);
	
		float mag_norm = sqrt(processed_data->mag_x*processed_data->mag_x + processed_data->mag_y*processed_data->mag_y + processed_data->mag_z*processed_data->mag_z);
				
		// Is the magnetometer measurement close to 1.0?  If so, clear the mag norm flag in the health register to indicate that we are using the mag
		if( fabs(mag_norm - 1.0f) < MAG_NORM_THRESHOLD )
		{
			g_data.r[DREG_HEALTH-DATA_REG_START_ADDRESS] &= ~HEALTH_MAG_NORM;
		}
		// If the measurement is too far away from 1.0, then clear the new_mag_data flag so that the estimator won't use it during update.
		else
		{
			g_data.r[DREG_HEALTH-DATA_REG_START_ADDRESS] |= HEALTH_MAG_NORM;
			processed_data->new_mag_data = 0;
		}
	}
}

void process_accel_data( RawSensorData* raw_data, ProcessedSensorData* processed_data )
{
	float accel_x, accel_y, accel_z;
	
	static float accel_x_old = 0.0f;
	static float accel_y_old = 0.0f;
	static float accel_z_old = 0.0f;
	
	static float T_old = 0.0f;
	
	float tau = *g_hidden_settings.p_lpf_tau_accel;

	float current_time = get_system_time_uncal();
	float T = unwrap_delta_T(current_time - T_old);
	
	fMatrix vector;
	float vector_data[3];
	
	// Retrieve scale factors
	float Casx0 = *g_hidden_settings.p_Casx0;
	float Casx1 = *g_hidden_settings.p_Casx1;
	float Casx2 = *g_hidden_settings.p_Casx2;
	float Casx3 = *g_hidden_settings.p_Casx3;
	
	float Casy0 = *g_hidden_settings.p_Casy0;
	float Casy1 = *g_hidden_settings.p_Casy1;
	float Casy2 = *g_hidden_settings.p_Casy2;
	float Casy3 = *g_hidden_settings.p_Casy3;
	
	float Casz0 = *g_hidden_settings.p_Casz0;
	float Casz1 = *g_hidden_settings.p_Casz1;
	float Casz2 = *g_hidden_settings.p_Casz2;
	float Casz3 = *g_hidden_settings.p_Casz3;
	
	// Retrieve biases
	float Cabx0 = *g_hidden_settings.p_Cabx0;
	float Cabx1 = *g_hidden_settings.p_Cabx1;
	float Cabx2 = *g_hidden_settings.p_Cabx2;
	float Cabx3 = *g_hidden_settings.p_Cabx3;
	
	float Caby0 = *g_hidden_settings.p_Caby0;
	float Caby1 = *g_hidden_settings.p_Caby1;
	float Caby2 = *g_hidden_settings.p_Caby2;
	float Caby3 = *g_hidden_settings.p_Caby3;
	
	float Cabz0 = *g_hidden_settings.p_Cabz0;
	float Cabz1 = *g_hidden_settings.p_Cabz1;
	float Cabz2 = *g_hidden_settings.p_Cabz2;
	float Cabz3 = *g_hidden_settings.p_Cabz3;
	
	vector.rows = 3;
	vector.columns = 1;
	vector.data = vector_data;
	
	accel_x = (float)raw_data->accel_x;
	accel_y = (float)raw_data->accel_y;
	accel_z = (float)raw_data->accel_z;
	
	// Do axis misalignment correction
	vector.data[0] = accel_x;
	vector.data[1] = accel_y;
	vector.data[2] = accel_z;
	mat_mult(&g_hidden_settings.Ma,&vector,&vector);
		
	accel_x = vector.data[0];
	accel_y = vector.data[1];
	accel_z = vector.data[2];

	// Apply temperature-compensated scale factors and biases
	float accel_temp = (float)(raw_data->temperature);
	float accel_temp2 = accel_temp*accel_temp;
	float accel_temp3 = accel_temp2*accel_temp;
	
	accel_x = (accel_x) * (Casx0 + Casx1*accel_temp + Casx2*accel_temp2 + Casx3*accel_temp3) + (Cabx0 + Cabx1*accel_temp + Cabx2*accel_temp2 + Cabx3*accel_temp3);
	accel_y = (accel_y) * (Casy0 + Casy1*accel_temp + Casy2*accel_temp2 + Casy3*accel_temp3) + (Caby0 + Caby1*accel_temp + Caby2*accel_temp2 + Caby3*accel_temp3);
	accel_z = (accel_z) * (Casz0 + Casz1*accel_temp + Casz2*accel_temp2 + Casz3*accel_temp3) + (Cabz0 + Cabz1*accel_temp + Cabz2*accel_temp2 + Cabz3*accel_temp3);

	// Low-pass filter
	processed_data->accel_x = lpf_sp_bt(tau,accel_x,accel_x_old,processed_data->accel_x,T);
	processed_data->accel_y = lpf_sp_bt(tau,accel_y,accel_y_old,processed_data->accel_y,T);
	processed_data->accel_z = lpf_sp_bt(tau,accel_z,accel_z_old,processed_data->accel_z,T);
	
	processed_data->accel_time = raw_data->accel_time;
	
	accel_x_old = accel_x;
	accel_y_old = accel_y;
	accel_z_old = accel_z;
	
	T_old = current_time;
	
	processed_data->new_accel_data = 1;
	raw_data->new_accel_data = 0;
}

void process_gyro_data( RawSensorData* raw_data, ProcessedSensorData* processed_data )
{
	float gyro_x, gyro_y, gyro_z;
	
	static float gyro_x_old = 0.0f;
	static float gyro_y_old = 0.0f;
	static float gyro_z_old = 0.0f;
	
	static float T_old = 0.0f;

	float current_time = get_system_time_uncal();
	float T = unwrap_delta_T(current_time - T_old);
	
	float tau = *g_hidden_settings.p_lpf_tau_gyro;
	
	fMatrix vector;
	float vector_data[3];
	
	// Retrieve scale factors
	float Cgsx0 = *g_hidden_settings.p_Cgsx0;
	float Cgsx1 = *g_hidden_settings.p_Cgsx1;
	float Cgsx2 = *g_hidden_settings.p_Cgsx2;
	float Cgsx3 = *g_hidden_settings.p_Cgsx3;
		
	float Cgsy0 = *g_hidden_settings.p_Cgsy0;
	float Cgsy1 = *g_hidden_settings.p_Cgsy1;
	float Cgsy2 = *g_hidden_settings.p_Cgsy2;
	float Cgsy3 = *g_hidden_settings.p_Cgsy3;
		
	float Cgsz0 = *g_hidden_settings.p_Cgsz0;
	float Cgsz1 = *g_hidden_settings.p_Cgsz1;
	float Cgsz2 = *g_hidden_settings.p_Cgsz2;
	float Cgsz3 = *g_hidden_settings.p_Cgsz3;
		
	// Retrieve biases
	float Cgbx0 = *g_hidden_settings.p_Cgbx0;
	float Cgbx1 = *g_hidden_settings.p_Cgbx1;
	float Cgbx2 = *g_hidden_settings.p_Cgbx2;
	float Cgbx3 = *g_hidden_settings.p_Cgbx3;
		
	float Cgby0 = *g_hidden_settings.p_Cgby0;
	float Cgby1 = *g_hidden_settings.p_Cgby1;
	float Cgby2 = *g_hidden_settings.p_Cgby2;
	float Cgby3 = *g_hidden_settings.p_Cgby3;
		
	float Cgbz0 = *g_hidden_settings.p_Cgbz0;
	float Cgbz1 = *g_hidden_settings.p_Cgbz1;
	float Cgbz2 = *g_hidden_settings.p_Cgbz2;
	float Cgbz3 = *g_hidden_settings.p_Cgbz3;
		
	vector.rows = 3;
	vector.columns = 1;
	vector.data = vector_data;

	// Do axis misalignment correction
	vector.data[0] = (float)raw_data->gyro_x;
	vector.data[1] = (float)raw_data->gyro_y;
	vector.data[2] = (float)raw_data->gyro_z;
	mat_mult(&g_hidden_settings.Mg,&vector,&vector);
	
	gyro_x = vector.data[0];
	gyro_y = vector.data[1];
	gyro_z = vector.data[2];	

	// Apply temperature-compensated scale factors and biases		
	float gyro_temp = (float)processed_data->temperature;
	float gyro_temp2 = gyro_temp*gyro_temp;
	float gyro_temp3 = gyro_temp2*gyro_temp;
	
	gyro_x =  (gyro_x - (Cgbx0 + Cgbx1*gyro_temp + Cgbx2*gyro_temp2 + Cgbx3*gyro_temp3)) * (Cgsx0 + Cgsx1*gyro_temp + Cgsx2*gyro_temp2 + Cgsx3*gyro_temp3) - g_states.gyro_x_trim;
	gyro_y =  (gyro_y - (Cgby0 + Cgby1*gyro_temp + Cgby2*gyro_temp2 + Cgby3*gyro_temp3)) * (Cgsy0 + Cgsy1*gyro_temp + Cgsy2*gyro_temp2 + Cgsy3*gyro_temp3) - g_states.gyro_y_trim;
	gyro_z =  (gyro_z - (Cgbz0 + Cgbz1*gyro_temp + Cgbz2*gyro_temp2 + Cgbz3*gyro_temp3)) * (Cgsz0 + Cgsz1*gyro_temp + Cgsz2*gyro_temp2 + Cgsz3*gyro_temp3) - g_states.gyro_z_trim;

	// Now low-pass filter the data
	processed_data->gyro_x = lpf_sp_bt(tau, gyro_x, gyro_x_old, processed_data->gyro_x, T);
	processed_data->gyro_y = lpf_sp_bt(tau, gyro_y, gyro_y_old, processed_data->gyro_y, T);
	processed_data->gyro_z = lpf_sp_bt(tau, gyro_z, gyro_z_old, processed_data->gyro_z, T);
	
	processed_data->gyro_time_dt = raw_data->gyro_time - processed_data->gyro_time;
	processed_data->gyro_time = raw_data->gyro_time;
	
	// Handle gyro zeroing operation if zeroing is enabled
	if( g_zeroing_gyros )
	{
		g_gyro_sum[GYRO_X_INDEX] += (gyro_x +  g_states.gyro_x_trim);
		g_gyro_sum[GYRO_Y_INDEX] += (gyro_y + g_states.gyro_y_trim);
		g_gyro_sum[GYRO_Z_INDEX] += (gyro_z + g_states.gyro_z_trim);
			
		g_gyro_samples_collected++;
					
		if( g_gyro_samples_collected == GYRO_ZERO_SAMPLES )
		{
			stop_gyro_zeroing();
		}
	}

	gyro_x_old = gyro_x;
	gyro_y_old = gyro_y;
	gyro_z_old = gyro_z;
	
	T_old = current_time;
	
	processed_data->new_gyro_data = 1;
	raw_data->new_gyro_data = 0;
}

void process_mag_data( RawSensorData* raw_data, ProcessedSensorData* processed_data )
{
	float mag_x, mag_y, mag_z;
	
	static float mag_x_old = 0.0f;
	static float mag_y_old = 0.0f;
	static float mag_z_old = 0.0f;
	
	static float T_old = 0.0f;

	float current_time = get_system_time_uncal();
	float T = unwrap_delta_T(current_time - T_old);
	
	float tau = *g_hidden_settings.p_lpf_tau_mag;
	
	fMatrix vector;
	float vector_data[3];
	
	// Retrieve scale factors
	float Cmsx0 = *g_hidden_settings.p_Cmsx0;
	float Cmsx1 = *g_hidden_settings.p_Cmsx1;
	float Cmsx2 = *g_hidden_settings.p_Cmsx2;
	float Cmsx3 = *g_hidden_settings.p_Cmsx3;
	
	float Cmsy0 = *g_hidden_settings.p_Cmsy0;
	float Cmsy1 = *g_hidden_settings.p_Cmsy1;
	float Cmsy2 = *g_hidden_settings.p_Cmsy2;
	float Cmsy3 = *g_hidden_settings.p_Cmsy3;
	
	float Cmsz0 = *g_hidden_settings.p_Cmsz0;
	float Cmsz1 = *g_hidden_settings.p_Cmsz1;
	float Cmsz2 = *g_hidden_settings.p_Cmsz2;
	float Cmsz3 = *g_hidden_settings.p_Cmsz3;
	
	// Retrieve biases
	float Cmbx0 = *g_hidden_settings.p_Cmbx0;
	float Cmbx1 = *g_hidden_settings.p_Cmbx1;
	float Cmbx2 = *g_hidden_settings.p_Cmbx2;
	float Cmbx3 = *g_hidden_settings.p_Cmbx3;
	
	float Cmby0 = *g_hidden_settings.p_Cmby0;
	float Cmby1 = *g_hidden_settings.p_Cmby1;
	float Cmby2 = *g_hidden_settings.p_Cmby2;
	float Cmby3 = *g_hidden_settings.p_Cmby3;
	
	float Cmbz0 = *g_hidden_settings.p_Cmbz0;
	float Cmbz1 = *g_hidden_settings.p_Cmbz1;
	float Cmbz2 = *g_hidden_settings.p_Cmbz2;
	float Cmbz3 = *g_hidden_settings.p_Cmbz3;
		
	vector.rows = 3;
	vector.columns = 1;
	vector.data = vector_data;

	// Do axis misalignment correction
	vector.data[0] = (float)raw_data->mag_x;
	vector.data[1] = (float)raw_data->mag_y;
	vector.data[2] = (float)raw_data->mag_z;
	mat_mult(&g_hidden_settings.Mm,&vector,&vector);
	
	mag_x = vector.data[0];
	mag_y = vector.data[1];
	mag_z = vector.data[2];

	// Apply temperature-compensated scale factors and biases		
	float mag_temp = processed_data->temperature;
	float mag_temp2 = mag_temp*mag_temp;
	float mag_temp3 = mag_temp2*mag_temp;
	
	mag_x =  (mag_x - (Cmbx0 + Cmbx1*mag_temp + Cmbx2*mag_temp2 + Cmbx3*mag_temp3)) * (Cmsx0 + Cmsx1*mag_temp + Cmsx2*mag_temp2 + Cmsx3*mag_temp3);
	mag_y =  (mag_y - (Cmby0 + Cmby1*mag_temp + Cmby2*mag_temp2 + Cmby3*mag_temp3)) * (Cmsy0 + Cmsy1*mag_temp + Cmsy2*mag_temp2 + Cmsy3*mag_temp3);
	mag_z =  (mag_z - (Cmbz0 + Cmbz1*mag_temp + Cmbz2*mag_temp2 + Cmbz3*mag_temp3)) * (Cmsz0 + Cmsz1*mag_temp + Cmsz2*mag_temp2 + Cmsz3*mag_temp3);
	
	// Apply user soft and hard iron calibration
	vector.data[0] = mag_x - *g_states.p_mag_x_bias;
	vector.data[1] = mag_y - *g_states.p_mag_y_bias;
	vector.data[2] = mag_z - *g_states.p_mag_z_bias;
	mat_mult(&g_states.Msoft,&vector,&vector);
	
	mag_x = vector.data[0];
	mag_y = vector.data[1];
	mag_z = vector.data[2];
	
	// Now low-pass filter the data
	processed_data->mag_x = lpf_sp_bt(tau, mag_x, mag_x_old, processed_data->mag_x, T);
	processed_data->mag_y = lpf_sp_bt(tau, mag_y, mag_y_old, processed_data->mag_y, T);
	processed_data->mag_z = lpf_sp_bt(tau, mag_z, mag_z_old, processed_data->mag_z, T);
	
	processed_data->mag_time_dt = raw_data->mag_time - processed_data->mag_time;
	processed_data->mag_time = raw_data->mag_time;
		
	mag_x_old = mag_x;
	mag_y_old = mag_y;
	mag_z_old = mag_z;
	
	T_old = current_time;
	
	// Finally, normalize if needed
	/*
	float mag_norm = sqrtf(square_f(processed_data->mag_x) + square_f(processed_data->mag_y) + square_f(processed_data->mag_z));
	if (mag_norm > 1.001f || mag_norm < 0.999f)
	{
		processed_data->mag_x /= mag_norm;
		processed_data->mag_y /= mag_norm;
		processed_data->mag_z /= mag_norm;
	}
	*/
	
	// Set flags
	processed_data->new_mag_data = 1;
	raw_data->new_mag_data = 0; 
}

/*******************************************************************************
* Function Name  : EKF_EstimateStates
* Input          : AHRS_states* states, RawSensorData* sensor_data
* Output         : None
* Return         : None
* Description    : 
*******************************************************************************/
void EKF_EstimateStates( StateData* state_data, ProcessedSensorData* sensor_data )
{
	 static int initialized = 0;
	 fConvert f2int;
	 fConvert scale;
	 
	 if( !initialized )
	 {
		  EKF_Init( state_data );
		  initialized = 1;
	 }
 
	// Check to see if the EKF has crashed (Euler Angle mode only)
	if( gEKF_mode == EKF_MODE_EULER )
	{
		if( isinf(state_data->phi) || isinf(state_data->theta) || isinf(state_data->psi) || isnan(state_data->phi) || isnan(state_data->theta) || isnan(state_data->psi) )
		{
			EKF_Init( state_data );
		}
 	}

	 // Run EKF prediction step
	 EKF_Predict( state_data, sensor_data );
	 	 
	 // Run EKF update step
	 EKF_Update( state_data, sensor_data );
 
	 // Copy the new states into the communication interface structures
	 copy_data_to_output();
}

/*******************************************************************************
* Function Name  : EKF_Predict
* Input          : AHRS_states* states, RawSensorData* sensor_data
* Output         : None
* Return         : None
* Description    : EKF prediction step.  Uses rate gyros to make new orientation
						 estimate.
*******************************************************************************/
void EKF_Predict( StateData* states, ProcessedSensorData* sensor_data )
{
	float cos_phi,sin_phi,cos_theta,tan_theta,sin_theta;
    static float last_system_time = 0;
    float T,p,q,r;
    
	fMatrix A,Atranspose,temp1,temp2;
	float A_data[4*4];
	float Atranspose_data[4*4];
	float temp1_data[4*4];
	float temp2_data[4*4];

    fConvert i2f;

	// Assign pointer locations for matrices
	A.data = A_data;
	Atranspose.data = Atranspose_data;
	temp1.data = temp1_data;
	temp2.data = temp2_data;

    float current_time = get_system_time_uncal();

    // Compute sample period
    T = unwrap_delta_T(current_time - last_system_time);
    last_system_time = current_time;

    // Copy body frame angular rates to local variables for convenience
    p = sensor_data->gyro_x;
    q = sensor_data->gyro_y;
    r = sensor_data->gyro_z;
	
	// Precompute trigonometric functions - these will be used more than once
	cos_phi = cos(states->phi*.01745329f);
	sin_phi = sin(states->phi*.01745329f);
	cos_theta = cos(states->theta*.01745329f);
	sin_theta = sin(states->theta*.01745329f);	 
	tan_theta = tan(states->theta*.01745329f);
		
	// Compute rotation rates based on body frame angular rates measured by the rate gyros
 //       phi_dot = p + r*cos(phi)*tan(theta) + q*sin(phi)*tan(theta)
 //       theta_dot = q*cos(phi) - r*sin(phi)
 //       psi_dot = (r*cos(phi))/cos(theta) + (q*sin(phi))/cos(theta)
	states->phi_dot = p + r*cos_phi*tan_theta + q*sin_phi*tan_theta;
	states->theta_dot = q*cos_phi - r*sin_phi;
	states->psi_dot = (r*cos_phi)/cos_theta + (q*sin_phi)/cos_theta;
	
	states->euler_time = sensor_data->gyro_time;
	states->quat_time = sensor_data->gyro_time;
		  
    // Euler Angle Estimation
    if( (g_config.r[CREG_MISC_SETTINGS] & QUATERNION_MODE_ENABLED) == 0 )
    {
        A.rows = 3;
        A.columns = 3;
	
        Atranspose.rows = 3;
        Atranspose.columns = 3;
	
        temp1.rows = 3;
        temp1.columns = 3;
	
        temp2.rows = 3;
        temp2.columns = 3;

        // Use measured rotation rates in the body frame to compute new angle estimates
        states->phi += T*states->phi_dot;
        states->theta += T*states->theta_dot;
        states->psi += T*states->psi_dot;

        // DISCRETE STATE TRANSITION
        //	 [ T*q*cos(phi)*tan(theta) - T*r*sin(phi)*tan(theta) + 1,               T*r*cos(phi)*(tan(theta)^2 + 1) + T*q*sin(phi)*(tan(theta)^2 + 1), 0]
        //	 [                         - T*q*sin(phi) - T*r*cos(phi),                                                                               1, 0]
        //	 [ (T*q*cos(phi))/cos(theta) - (T*r*sin(phi))/cos(theta), (T*r*cos(phi)*sin(theta))/cos(theta)^2 + (T*q*sin(phi)*sin(theta))/cos(theta)^2, 1]
        A.data[0*3 + 0] = T*(q*cos_phi*tan_theta - r*sin_phi*tan_theta) + 1;
        A.data[0*3 + 1] = T*(r*cos_phi*(tan_theta*tan_theta + 1) + q*sin_phi*(tan_theta*tan_theta + 1));
        A.data[0*3 + 2] = 0;
        A.data[1*3 + 0] = T*(-r*cos_phi - q*sin_phi);
        A.data[1*3 + 1] = 1;
        A.data[1*3 + 2] = 0;
        A.data[2*3 + 0] = T*((q*cos_phi)/cos_theta - (r*sin_phi)/cos_theta);
        A.data[2*3 + 1] = T*((r*cos_phi*sin_theta)/(cos_theta*cos_theta) + (q*sin_phi*sin_theta)/(cos_theta*cos_theta)); 
        A.data[2*3 + 2] = 1;

        // Compute the new covariance estimate (discrete update: P = A*P*Atranspose + Q
        mat_transpose( &A, &Atranspose );
        mat_mult( &A, &states->P, &temp1 );
        mat_mult( &temp1, &Atranspose, &temp2 );
        mat_add( &temp2, &states->Q, &states->P );

        // Finally, "unroll" states so that they range from -360 to 360 degrees
        unroll_states( states );
    }
    // Quaternion estimation
    else
    {
        A.rows = 4;
        A.columns = 4;
	
        Atranspose.rows = 4;
        Atranspose.columns = 4;
	
        temp1.rows = 4;
        temp1.columns = 4;
	
        temp2.rows = 4;
        temp2.columns = 4;

        float a, b, c, d;

        // Make local copies of the current attitude quaternion for convenience
        a = states->qib.a;
        b = states->qib.b;
        c = states->qib.c;
        d = states->qib.d;

        // Convert p, q, and r to rad/s
        p = p*3.14159/180;
        q = q*3.14159/180;
        r = r*3.14159/180;

        // Create a quaternion to represent rotation rate
        quat pqr_quat;
        pqr_quat.a = 0;
        pqr_quat.b = p;
        pqr_quat.c = q;
        pqr_quat.d = r;

        // Predict new quaternion state based on gyro data
        quat temp_quat;
        quat_mult( &states->qib, &pqr_quat, &temp_quat );
        quat_scalar_mult( &temp_quat, 0.5*T, &temp_quat );
        quat_add( &states->qib, &temp_quat, &states->qib );

        // Normalize new predicted state
        quat_norm( &states->qib );

        // PROPAGATE COVARIANCE
        // Compute linearized state transition matrix
//            [       1, -(T*p)/2, -(T*q)/2, -(T*r)/2]
//            [ (T*p)/2,        1,  (T*r)/2, -(T*q)/2]
//            [ (T*q)/2, -(T*r)/2,        1,  (T*p)/2]
//            [ (T*r)/2,  (T*q)/2, -(T*p)/2,        1]
        A.data[0*4 + 0] = 1;
        A.data[0*4 + 1] = -(T*p)/2;
        A.data[0*4 + 2] = -(T*q)/2;
        A.data[0*4 + 3] = -(T*r)/2;

        A.data[1*4 + 0] = (T*p)/2;
        A.data[1*4 + 1] = 1;
        A.data[1*4 + 2] = (T*r)/2;
        A.data[1*4 + 3] = -(T*q)/2;

        A.data[2*4 + 0] = (T*q)/2;
        A.data[2*4 + 1] = -(T*r)/2;
        A.data[2*4 + 2] = 1;
        A.data[2*4 + 3] = (T*p)/2;

        A.data[3*4 + 0] = (T*r)/2;
        A.data[3*4 + 1] = (T*q)/2;
        A.data[3*4 + 2] = -(T*p)/2;
        A.data[3*4 + 3] = 1;

        // Compute the new covariance estimate (discrete update: P = A*P*Atranspose + Q
        mat_transpose( &A, &Atranspose );
        mat_mult( &A, &states->P, &temp1 );
        mat_mult( &temp1, &Atranspose, &temp2 );
        mat_add( &temp2, &states->Q, &states->P );

        // Now use the new quaternion to compute Euler Angles
        compute_euler_angles( states );
    }
}

/*******************************************************************************
* Function Name  : EKF_Update
* Input          : AHRS_states* states, RawSensorData* sensor_data
* Output         : None
* Return         : None
* Description    : EKF update step.  Uses accels to correct pitch and roll errors,
						 and magnetic sensors to correct yaw errors.  Compensation is
						 only applied when new data is available, as specified by the
						 new_mag_data and new_accel_data flags in the sensor_data structure.
*******************************************************************************/
void EKF_Update( StateData* states, ProcessedSensorData* sensor_data )
{
	if( (g_config.r[CREG_MISC_SETTINGS] & QUATERNION_MODE_ENABLED) == 0 )
	{
		float cos_phi, cos_theta, cos_psi, sin_phi, sin_theta, sin_psi;
		fMatrix C;
		float C_data[3];

		C.data = C_data;
		C.rows = 1;
		C.columns = 3;

		// If there is new accelerometer data available and accelerometer state updates are enabled,
		// run an update using the new accelerometer sensor data
		if( sensor_data->new_accel_data )
		{
			sensor_data->new_accel_data = 0;
					
			float ax_hat, ay_hat, az_hat;
			float ax_ref, ay_ref, az_ref;
		
			// Precompute trigonometric functions - these will be used more than once
			cos_phi = cos(states->phi*.01745329);
			cos_theta = cos(states->theta*.01745329);
			cos_psi = cos(states->psi*.01745329);

			sin_phi = sin(states->phi*.01745329);
			sin_theta = sin(states->theta*.01745329);
			sin_psi = sin(states->psi*.01745329);

			// Copy data into local variables for convenience		  
			ax_ref = 0.0f;
			ay_ref = 0.0f;
			az_ref = -1.0f;

			// Compute expected accelerometer output based on current state estimates (the expected output is the accelerometer reference vector
			// rotated into the body-frame of the sensor
			ax_hat = ax_ref*cos_psi*cos_theta - az_ref*sin_theta + ay_ref*cos_theta*sin_psi;
			ay_hat = ay_ref*(cos_phi*cos_psi + sin_phi*sin_psi*sin_theta) - ax_ref*(cos_phi*sin_psi - cos_psi*sin_phi*sin_theta) + az_ref*cos_theta*sin_phi;
			az_hat = ax_ref*(sin_phi*sin_psi + cos_phi*cos_psi*sin_theta) - ay_ref*(cos_psi*sin_phi - cos_phi*sin_psi*sin_theta) + az_ref*cos_phi*cos_theta;

			// Compute linearized state transition matrix for each axis independently
			//		  [                                                                                                                                                  0,                          - az_ref*cos(theta) - ax_ref*cos(psi)*sin(theta) - ay_ref*sin(psi)*sin(theta),                                                                 ay_ref*cos(psi)*cos(theta) - ax_ref*cos(theta)*sin(psi)]
			//		  [ ax_ref*(sin(phi)*sin(psi) + cos(phi)*cos(psi)*sin(theta)) - ay_ref*(cos(psi)*sin(phi) - cos(phi)*sin(psi)*sin(theta)) + az_ref*cos(phi)*cos(theta), ax_ref*cos(psi)*cos(theta)*sin(phi) - az_ref*sin(phi)*sin(theta) + ay_ref*cos(theta)*sin(phi)*sin(psi), - ax_ref*(cos(phi)*cos(psi) + sin(phi)*sin(psi)*sin(theta)) - ay_ref*(cos(phi)*sin(psi) - cos(psi)*sin(phi)*sin(theta))]
			//		  [ ax_ref*(cos(phi)*sin(psi) - cos(psi)*sin(phi)*sin(theta)) - ay_ref*(cos(phi)*cos(psi) + sin(phi)*sin(psi)*sin(theta)) - az_ref*cos(theta)*sin(phi), ax_ref*cos(phi)*cos(psi)*cos(theta) - az_ref*cos(phi)*sin(theta) + ay_ref*cos(phi)*cos(theta)*sin(psi),   ax_ref*(cos(psi)*sin(phi) - cos(phi)*sin(psi)*sin(theta)) + ay_ref*(sin(phi)*sin(psi) + cos(phi)*cos(psi)*sin(theta))]

			// x-axis
			C.data[0] = 0;
			C.data[1] = -az_ref*cos_theta - ax_ref*cos_psi*sin_theta - ay_ref*sin_psi*sin_theta;
			C.data[2] = ay_ref*cos_psi*cos_theta - ax_ref*cos_theta*sin_psi;

			// Do correction
			EKF_Correction( &C, sensor_data->accel_x, ax_hat, *g_hidden_settings.p_accel_variance, states, ACCEL_UPDATE );

			// y-axis
			C.data[0] = ax_ref*(sin_phi*sin_psi + cos_phi*cos_psi*sin_theta) - ay_ref*(cos_psi*sin_phi - cos_phi*sin_psi*sin_theta) + az_ref*cos_phi*cos_theta;
			C.data[1] = ax_ref*cos_psi*cos_theta*sin_phi - az_ref*sin_phi*sin_theta + ay_ref*cos_theta*sin_phi*sin_psi;
			C.data[2] = -ax_ref*(cos_phi*cos_psi + sin_phi*sin_psi*sin_theta) - ay_ref*(cos_phi*sin_psi - cos_psi*sin_phi*sin_theta);

			// Do correction
			EKF_Correction( &C, sensor_data->accel_y, ay_hat, *g_hidden_settings.p_accel_variance, states, ACCEL_UPDATE );

			// z-axis
			C.data[0] = ax_ref*(cos_phi*sin_psi - cos_psi*sin_phi*sin_theta) - ay_ref*(cos_phi*cos_psi + sin_phi*sin_psi*sin_theta) - az_ref*cos_theta*sin_phi;
			C.data[1] = ax_ref*cos_phi*cos_psi*cos_theta - az_ref*cos_phi*sin_theta + ay_ref*cos_phi*cos_theta*sin_psi;
			C.data[2] = ax_ref*(cos_psi*sin_phi - cos_phi*sin_psi*sin_theta) + ay_ref*(sin_phi*sin_psi + cos_phi*cos_psi*sin_theta);

			// Do correction
			EKF_Correction( &C, sensor_data->accel_z, az_hat, *g_hidden_settings.p_accel_variance, states, ACCEL_UPDATE );

			// Finally, "unroll" states so that they range from -360 to 360 degrees
			unroll_states( states );
		}

		// Do magnetometer update if enabled and if there is new magnetic sensor data available
		if( sensor_data->new_mag_data && (g_config.r[CREG_MISC_SETTINGS] & MAG_UPDATES_ENABLED) )
		{
			sensor_data->new_mag_data = 0;
			
			float mx_hat, my_hat, mz_hat;
			float mx_ref, my_ref, mz_ref;
			float mx_v1, my_v1, mz_v1;

			// Precompute trigonometric functions - these will be used more than once
			cos_phi = cos(states->phi*.01745329);
			cos_theta = cos(states->theta*.01745329);
			cos_psi = cos(states->psi*.01745329);

			sin_phi = sin(states->phi*.01745329);
			sin_theta = sin(states->theta*.01745329);
			sin_psi = sin(states->psi*.01745329);

			// Copy data into local variables for convenience		  
			mx_ref = *g_hidden_settings.p_mag_ref_x;
			my_ref = *g_hidden_settings.p_mag_ref_y;
			mz_ref = *g_hidden_settings.p_mag_ref_z;

			// Rotate the sensor measurement into the vehicle-1 frame (undo pitch and roll)
			mx_v1 = sensor_data->mag_x*cos_theta + sensor_data->mag_z*cos_phi*sin_theta + sensor_data->mag_y*sin_phi*sin_theta;
			my_v1 = sensor_data->mag_y*cos_phi - sensor_data->mag_z*sin_phi;
			mz_v1 =  sensor_data->mag_z*cos_phi*cos_theta - sensor_data->mag_x*sin_theta + sensor_data->mag_y*cos_theta*sin_phi;

			// Compute expected magnetometer output based on current state estimates
			// (rotates the reference vector into the vehicle-1 frame)
			mx_hat = mx_ref*cos_psi + my_ref*sin_psi;
			my_hat = my_ref*cos_psi - mx_ref*sin_psi;
			mz_hat = mz_ref;

			// Compute linearized state transition matrix for each axis independently	  
			// x-axis
			C.data[0] = 0;
			C.data[1] = 0;
			C.data[2] = my_ref*cos_psi - mx_ref*sin_psi;

			// Do correction
			EKF_Correction( &C, mx_v1, mx_hat, *g_hidden_settings.p_mag_variance, states, MAG_UPDATE );

			// y-axis
			C.data[0] = 0;
			C.data[1] = 0;
			C.data[2] = -mx_ref*cos_psi - my_ref*sin_psi;

			// Do correction
			EKF_Correction( &C, my_v1, my_hat, *g_hidden_settings.p_mag_variance, states, MAG_UPDATE );

			//				z-axis doesn't do anything in the vehicle-1 frame
			// z-axis
			//				C.data[0*3 + 0] = 0;
			//				C.data[0*3 + 1] = 0;
			//				C.data[0*3 + 2] = 0;

			// Do correction
			//				EKF_Correction( &C, mz_v1, mz_hat, g_states.mag_var, states, MAG_UPDATE );


			// Finally, "unroll" states so that they range from -360 to 360 degrees
			unroll_states( states );
		}
	}
	// QUATERNION CODE
	else
	{
		float a,b,c,d;
		fMatrix C;
		float C_data[4];

		C.data = C_data;
		C.rows = 1;
		C.columns = 4;

		// Do accelerometer update if enabled and if there is new accelerometer sensor data available
		if( sensor_data->new_accel_data )
		{
			float ax_hat, ay_hat, az_hat;
			float ax_ref, ay_ref, az_ref;

			// Make local copies of the current quaternion state estimate for convenience
			a = states->qib.a;
			b = states->qib.b;
			c = states->qib.c;
			d = states->qib.d;

			// Make local copy of accelerometer reference vector for convenience
			ax_ref = 0.0f;
			ay_ref = 0.0f;
			az_ref = -1.0f;

			// Compute expected accelerometer measurements based on the current attitude quaternion and the accel reference vector
			ax_hat = ay_ref*(2*a*d + 2*b*c) - az_ref*(2*a*c - 2*b*d) + ax_ref*(a*a + b*b - c*c - d*d);
							 
			// Create linearized update matrix for x-axis accelerometer
			// For all axes, C is given by:
			//					 [ 2*a*ax - 2*az*c + 2*ay*d, 2*ax*b + 2*ay*c + 2*az*d, 2*ay*b - 2*a*az - 2*ax*c, 2*a*ay + 2*az*b - 2*ax*d]
			//					 [ 2*a*ay + 2*az*b - 2*ax*d, 2*a*az - 2*ay*b + 2*ax*c, 2*ax*b + 2*ay*c + 2*az*d, 2*az*c - 2*a*ax - 2*ay*d]
			//					 [ 2*a*az - 2*ay*b + 2*ax*c, 2*ax*d - 2*az*b - 2*a*ay, 2*a*ax - 2*az*c + 2*ay*d, 2*ax*b + 2*ay*c + 2*az*d]					 

			C.data[0] = 2*a*ax_ref - 2*az_ref*c + 2*ay_ref*d;
			C.data[1] = 2*ax_ref*b + 2*ay_ref*c + 2*az_ref*d;
			C.data[2] = 2*ay_ref*b - 2*a*az_ref - 2*ax_ref*c;
			C.data[3] = 2*a*ay_ref + 2*az_ref*b - 2*ax_ref*d;

			// Do correction
			EKF_Correction( &C, sensor_data->accel_x, ax_hat, *g_hidden_settings.p_accel_variance, states, ACCEL_UPDATE );

			// REPEAT FOR Y-AXIS
			// Make local copies of the current quaternion state estimate for convenience
			a = states->qib.a;
			b = states->qib.b;
			c = states->qib.c;
			d = states->qib.d;

			// Compute expected accelerometer measurements based on the current attitude quaternion and the accel reference vector
			ay_hat = az_ref*(2*a*b + 2*c*d) - ax_ref*(2*a*d - 2*b*c) + ay_ref*(a*a - b*b + c*c - d*d);

			C.data[0] = 2*a*ay_ref + 2*az_ref*b - 2*ax_ref*d;
			C.data[1] = 2*a*az_ref - 2*ay_ref*b + 2*ax_ref*c;
			C.data[2] = 2*ax_ref*b + 2*ay_ref*c + 2*az_ref*d;
			C.data[3] = 2*az_ref*c - 2*a*ax_ref - 2*ay_ref*d;

			// Do correction
			EKF_Correction( &C, sensor_data->accel_y, ay_hat, *g_hidden_settings.p_accel_variance, states, ACCEL_UPDATE );

			// REPEAT FOR Z-AXIS
			// Make local copies of the current quaternion state estimate for convenience
			a = states->qib.a;
			b = states->qib.b;
			c = states->qib.c;
			d = states->qib.d;

			// Compute expected accelerometer measurements based on the current attitude quaternion and the accel reference vector
			az_hat = ax_ref*(2*a*c + 2*b*d) - ay_ref*(2*a*b - 2*c*d) + az_ref*(a*a - b*b - c*c + d*d);

			C.data[0] = 2*a*az_ref - 2*ay_ref*b + 2*ax_ref*c;
			C.data[1] = 2*ax_ref*d - 2*az_ref*b - 2*a*ay_ref;
			C.data[2] = 2*a*ax_ref - 2*az_ref*c + 2*ay_ref*d;
			C.data[3] = 2*ax_ref*b + 2*ay_ref*c + 2*az_ref*d;

			// Do correction
			EKF_Correction( &C, sensor_data->accel_z, az_hat, *g_hidden_settings.p_accel_variance, states, ACCEL_UPDATE );
			
			quat_norm( &states->qib );
			compute_euler_angles( states );
		}
		  
		// Do magnetometer update if enabled and if there is new magnetic sensor data available
		if( sensor_data->new_mag_data && (g_config.r[CREG_MISC_SETTINGS] & MAG_UPDATES_ENABLED) )
		{
			sensor_data->new_mag_data = 0;

			float mx_hat, my_hat, mz_hat;
			float mx_ref, my_ref, mz_ref;

			// Make local copies of the current quaternion state estimate for convenience
			a = states->qib.a;
			b = states->qib.b;
			c = states->qib.c;
			d = states->qib.d;

			// Make local copy of accelerometer reference vector for convenience
			mx_ref = *g_hidden_settings.p_mag_ref_x;
			my_ref = *g_hidden_settings.p_mag_ref_y;
			mz_ref = *g_hidden_settings.p_mag_ref_z;

			// Compute expected accelerometer measurements based on the current attitude quaternion and the accel reference vector
			mx_hat = my_ref*(2*a*d + 2*b*c) - mz_ref*(2*a*c - 2*b*d) + mx_ref*(a*a + b*b - c*c - d*d);
								
			// Create linearized update matrix for x-axis accelerometer
			// For all axes, C is given by:
			//				[ 2*a*ax - 2*az*c + 2*ay*d, 2*ax*b + 2*ay*c + 2*az*d, 2*ay*b - 2*a*az - 2*ax*c, 2*a*ay + 2*az*b - 2*ax*d]
			//				[ 2*a*ay + 2*az*b - 2*ax*d, 2*a*az - 2*ay*b + 2*ax*c, 2*ax*b + 2*ay*c + 2*az*d, 2*az*c - 2*a*ax - 2*ay*d]
			//				[ 2*a*az - 2*ay*b + 2*ax*c, 2*ax*d - 2*az*b - 2*a*ay, 2*a*ax - 2*az*c + 2*ay*d, 2*ax*b + 2*ay*c + 2*az*d]					 

			C.data[0] = 2*a*mx_ref - 2*mz_ref*c + 2*my_ref*d;
			C.data[1] = 2*mx_ref*b + 2*my_ref*c + 2*mz_ref*d;
			C.data[2] = 2*my_ref*b - 2*a*mz_ref - 2*mx_ref*c;
			C.data[3] = 2*a*my_ref + 2*mz_ref*b - 2*mx_ref*d;

			// Do correction
			EKF_Correction( &C, sensor_data->mag_x, mx_hat, *g_hidden_settings.p_mag_variance, states, MAG_UPDATE );

			// REPEAT FOR Y-AXIS
			// Make local copies of the current quaternion state estimate for convenience
			a = states->qib.a;
			b = states->qib.b;
			c = states->qib.c;
			d = states->qib.d;

			// Compute expected accelerometer measurements based on the current attitude quaternion and the accel reference vector
			my_hat = mz_ref*(2*a*b + 2*c*d) - mx_ref*(2*a*d - 2*b*c) + my_ref*(a*a - b*b + c*c - d*d);
			mz_hat = mx_ref*(2*a*c + 2*b*d) - my_ref*(2*a*b - 2*c*d) + mz_ref*(a*a - b*b - c*c + d*d);

			C.data[0] = 2*a*my_ref + 2*mz_ref*b - 2*mx_ref*d;
			C.data[1] = 2*a*mz_ref - 2*my_ref*b + 2*mx_ref*c;
			C.data[2] = 2*mx_ref*b + 2*my_ref*c + 2*mz_ref*d;
			C.data[3] = 2*mz_ref*c - 2*a*mx_ref - 2*my_ref*d;

			// Do correction
			EKF_Correction( &C, sensor_data->mag_y, my_hat, *g_hidden_settings.p_mag_variance, states, MAG_UPDATE );

			// REPEAT FOR Z-AXIS
			// Make local copies of the current quaternion state estimate for convenience
			a = states->qib.a;
			b = states->qib.b;
			c = states->qib.c;
			d = states->qib.d;

			// Compute expected accelerometer measurements based on the current attitude quaternion and the accel reference vector
			mz_hat = mx_ref*(2*a*c + 2*b*d) - my_ref*(2*a*b - 2*c*d) + mz_ref*(a*a - b*b - c*c + d*d);

			C.data[0] = 2*a*mz_ref - 2*my_ref*b + 2*mx_ref*c;
			C.data[1] = 2*mx_ref*d - 2*mz_ref*b - 2*a*my_ref;
			C.data[2] = 2*a*mx_ref - 2*mz_ref*c + 2*my_ref*d;
			C.data[3] = 2*mx_ref*b + 2*my_ref*c + 2*mz_ref*d;

			// Do correction
			EKF_Correction( &C, sensor_data->mag_z, mz_hat, *g_hidden_settings.p_mag_variance, states, MAG_UPDATE );
			
			quat_norm( &states->qib );
			compute_euler_angles( states );
		}
	}
}

void EKF_Correction( fMatrix* C, float sensor_data, float sensor_hat, float sensor_covariance, StateData* states, int sensor_type )
{
	fMatrix Ctranspose,temp1,temp2,L;
	float Ctranspose_data[4*4];
	float temp1_data[4*4];
	float temp2_data[4*4];
	float L_data[4*4];

	float gain_scale, error;

	Ctranspose.data = Ctranspose_data;
	temp1.data = temp1_data;
	temp2.data = temp2_data;
	L.data = L_data;
	

	if( (g_config.r[CREG_MISC_SETTINGS] & QUATERNION_MODE_ENABLED) == 0 )
	{
		mat_zero( &temp1, 3, 1 );
		mat_zero( &temp2, 3, 3 );
	}
	else
	{
		mat_zero( &temp1, 4, 1 );
		mat_zero( &temp2, 4, 4 );
	}
	
	mat_transpose( C, &Ctranspose );

	// Compute Kalman Gain (L = P*C'*(C*P*C' + Q)^-1 )
	mat_mult(&states->P,&Ctranspose,&temp1);
	mat_mult(C,&states->P,&temp2);
	mat_mult(&temp2,&Ctranspose,&temp2);
	gain_scale = 1/(temp2.data[0] + sensor_covariance);
	mat_scalar_mult(gain_scale,&temp1,&L);

	// Update state estimates
	error = sensor_data - sensor_hat;

	if( (g_config.r[CREG_MISC_SETTINGS] & QUATERNION_MODE_ENABLED) == 0 )
	{
		states->phi += L.data[0]*error;
		states->theta += L.data[1]*error;
		states->psi += L.data[2]*error;

		mat_create_identity( &temp1, 3, 3 );
	}
	else
	{
		states->qib.a += L.data[0]*error;
		states->qib.b += L.data[1]*error;
		states->qib.c += L.data[2]*error;
		states->qib.d += L.data[3]*error;

		mat_create_identity( &temp1, 4, 4 );
	}

	// Now update the covariance estimate (P = (I - L*C)*P
	mat_mult(&L,C,&temp2);
	mat_scalar_mult(-1,&temp2,&temp2);
	mat_add(&temp1,&temp2,&temp1);
	mat_mult(&temp1,&states->P, &states->P);
}

/*******************************************************************************
* Function Name  : compute_euler_angles
* Input          : AHRS_states* states
* Output         : None
* Return         : None
* Description    : Converts quaternion attitude estimate to euler angles (yaw, pitch, roll)
*******************************************************************************/
void compute_euler_angles( StateData* states )
{
	 float q0, q1, q2, q3;
	 
	 q0 = states->qib.a;
	 q1 = states->qib.b;
	 q2 = states->qib.c;
	 q3 = states->qib.d;	 

	 states->phi = atan2(2*(q0*q1 + q2*q3),q3*q3 - q2*q2 - q1*q1 + q0*q0)*180/3.14159;
	 states->theta = -asin(2*(q1*q3 - q0*q2))*180/3.14159;
	 states->psi = atan2(2*(q0*q3+q1*q2),q1*q1 + q0*q0 - q3*q3 - q2*q2)*180/3.14159;
}

/*******************************************************************************
* Function Name  : unroll_states
* Input          : AHRS_states* states
* Output         : None
* Return         : None
* Description    : Keeps all angle estimates in the range of -360 to 360 degrees
*******************************************************************************/
void unroll_states( StateData* states )
{
	uint8_t timeout;

	// Check for a crashed filter
	if( isinf(states->phi) || isinf(states->theta) || isinf(states->psi) )
	{
		return;
	}

	// Check for a crashed filter
	if( isnan(states->phi) || isnan(states->theta) || isnan(states->psi) )
	{
		return;
	}

	timeout = 10;
	while( states->phi > 360 && timeout-- )
	{
		states->phi -= 360;
	}
	if( timeout == 0 && gEKF_mode == EKF_MODE_EULER )
	{
		EKF_Init(states);
		
		return;
	}

	timeout = 10;
	while( states->phi < -360 && timeout-- )
	{
		states->phi += 360;
	}
	if( timeout == 0 && gEKF_mode == EKF_MODE_EULER )
	{
		EKF_Init(states);
		
		return;
	}

	timeout = 10;
	while( states->theta > 360 && timeout-- )
	{
		states->theta -= 360;
	}
	if( timeout == 0 && gEKF_mode == EKF_MODE_EULER )
	{
		EKF_Init(states);
		
		return;
	}

	timeout = 10;
	while( states->theta < -360 && timeout-- )
	{
		states->theta += 360;
	}
	if( timeout == 0 && gEKF_mode == EKF_MODE_EULER )
	{
		EKF_Init(states);
		
		return;
	}

	timeout = 10;
	while( states->psi > 360 && timeout-- )
	{
		states->psi -= 360;
	}
	if( timeout == 0 && gEKF_mode == EKF_MODE_EULER )
	{
		EKF_Init(states);
		
		return;
	}

	timeout = 10;
	while( states->psi < -360 && timeout-- )
	{
		states->psi += 360;
	}
	if( timeout == 0 && gEKF_mode == EKF_MODE_EULER )
	{
		EKF_Init(states);
		
		return;
	}
	 
}

/*
	Single-pole low-pass filter with transfer function
	
	Z = 1/(tau*s+1)
	
	Converted to the discrete domain using the bilinear transform:
	
	y[k] = 1/(2*tau/T + 1)*(x[k] + x[k-1]) + (2*tau-T)/(2*tau+T)*(y[k-1])
*/
float lpf_sp_bt( float tau, float measured_value, float last_measured_value, float last_output, float deltaT )
{
	// y[k] = 1/(2*tau/T + 1)*(x[k] + x[k-1]) + (2*tau-T)/(2*tau+T)*(y[k-1])
	return 1.0f/(2.0f*tau/deltaT+1.0f)*(measured_value + last_measured_value) + (2.0f*tau-deltaT)/(2.0f*tau+deltaT)*last_output;
}
