/* ------------------------------------------------------------------------------
  File: chr6dm_states.c
  Author: CH Robotics
  Version: 1.0
  
  Description: Function declarations for CHR-6dm state estimation.
------------------------------------------------------------------------------ */ 

#ifndef _CHR6DM_STATES__
#define _CHR6DM_STATES__

#include "stm32f10x.h"
#include "CHR_quat.h"
#include "CHR_matrix.h"

#define NUMBER_OF_STATES    4
#define NUMBER_OF_INPUTS    1

#define ACCEL_NORM_THRESHOLD	0.3
#define MAG_NORM_THRESHOLD		0.4

typedef struct _raw_sensor_data
{
	volatile int16_t gyro_x;
	volatile int16_t gyro_y;
	volatile int16_t gyro_z;
	volatile uint8_t new_gyro_data;

	volatile float gyro_time;
	volatile float temperature;
	
	volatile int16_t accel_x;
	volatile int16_t accel_y;
	volatile int16_t accel_z;
	volatile float accel_time;
	volatile uint8_t new_accel_data;
	
	volatile int16_t mag_x;
	volatile int16_t mag_y;
	volatile int16_t mag_z;
	volatile float mag_time;
	volatile uint8_t new_mag_data;
	
	volatile int32_t absolute_pressure;
	volatile int16_t abs_pressure_temperature_raw;
	volatile uint8_t new_pressure_data;
	volatile float abs_pressure_time;
	
	int16_t diff_pressure;
	float diff_pressure_time;
	uint8_t new_diff_pressure;
	
	float temperature_time;
	
	float GPS_latitude;
	float GPS_longitude;
	float GPS_altitude;
	float GPS_course;
	float GPS_speed;
	uint8_t GPS_mode;
	uint8_t GPS_satellites_used;
	uint8_t GPS_satellites_in_view;
	float GPS_Hdop;
	float GPS_Vdop;
	float GPS_time;
	float GPS_dt;
	
	float GPS_last_received_time;
	
    uint32_t GPS_date;
	
	uint8_t GPS_SatID_1;
	uint8_t GPS_SatID_2;
	uint8_t GPS_SatID_3;
	uint8_t GPS_SatID_4;
	uint8_t GPS_SatID_5;
	uint8_t GPS_SatID_6;
	uint8_t GPS_SatID_7;
	uint8_t GPS_SatID_8;
	uint8_t GPS_SatID_9;
	uint8_t GPS_SatID_10;
	uint8_t GPS_SatID_11;
	uint8_t GPS_SatID_12;
	
	uint8_t GPS_SatSNR_1;
	uint8_t GPS_SatSNR_2;
	uint8_t GPS_SatSNR_3;
	uint8_t GPS_SatSNR_4;
	uint8_t GPS_SatSNR_5;
	uint8_t GPS_SatSNR_6;
	uint8_t GPS_SatSNR_7;
	uint8_t GPS_SatSNR_8;
	uint8_t GPS_SatSNR_9;
	uint8_t GPS_SatSNR_10;
	uint8_t GPS_SatSNR_11;
	uint8_t GPS_SatSNR_12;
		
	uint8_t new_GPS_position;
	uint8_t new_GPS_course_speed;
	uint8_t new_GPS_satellite_data;
	uint8_t new_GPS_satellite_summary;
	uint8_t new_RC_data;
	
} RawSensorData;

typedef struct _processed_sensor_data
{
	float gyro_x;
	float gyro_y;
	float gyro_z;
	float temperature;
	float gyro_time;
	float gyro_time_dt;
	uint8_t new_gyro_data;
	
	float accel_x;
	float accel_y;
	float accel_z;
	float accel_time;
	uint8_t new_accel_data;
	
	float mag_x;
	float mag_y;
	float mag_z;
	float mag_time;
	float mag_time_dt;
	uint8_t new_mag_data;
	
	float absolute_pressure_time;
	float absolute_pressure;
	float position_down;
	float pressure_temperature;
	uint8_t new_pressure_data;
	
	float airspeed;
	float airspeed_time;
	uint8_t new_airspeed_data;
	
	float board_temperature;
	float temperature_time;
	float temperature_dt;
	
	float GPS_position_north;
	float GPS_position_east;
	float GPS_position_up;
	float GPS_course;
	float GPS_course_delayed;
	float GPS_speed;
	float GPS_velocity_north;
	float GPS_velocity_east;
	float GPS_position_time;
	float GPS_course_time;
	uint8_t new_GPS_position_data;
	uint8_t new_GPS_course_data;

} ProcessedSensorData;

// Contains data for both filters
typedef struct _state_data
{
	uint8_t initialized;
	 
	quat qib;
	float quat_time;
	 
	float phi;
	float theta;
	float psi;
	float euler_time;
	
	float phi_dot;
	float theta_dot;
	float psi_dot;

	float pn,pe,pd;             // Position north, east, down
	float position_time;

	float u,v,w;				// Body frame components of ground speed vector in 10-state filter.
    float vn,ve,vd;				// Inertial frame velocities (or vehicle frame): velocity north, east, down.                        
							 
    float velocity_time;
	
	float gbx,gby,gbz;			// Gyro biases. 

	float wn, we, wd;			// Wind N, E, down
	uint8_t new_wind_data;
	 
	float alpha;				// AoA
	float beta;					// side slip angle 
	
	// Trim values - not actual states, but placed here for convenience
	float gyro_x_trim;
	float gyro_y_trim;
	float gyro_z_trim;
	float pitch_trim_rad;
	float roll_trim_rad;
	float absolute_pressure_trim;
	float differential_pressure_trim;
	 
	// GPS home position - not actual states, but placed here for convenience
	float GPS_lat_home;
	float GPS_lon_home;
	float GPS_alt_home;
	
	// Magnetometer soft and hard iron calibration - not actual states, but placed here for convenience
	fMatrix Msoft;
	// float mag_soft_data[3*3];	Instead of this, we'll actually use the configuration array directly to access the matrix data.  Just make sure they are both row-major order...
	float* p_mag_x_bias;
	float* p_mag_y_bias;
	float* p_mag_z_bias;
	 
	// Filter mode
	uint8_t filter_mode;
	uint8_t mode_switched;	// goes high for one cycle when the filter has just changed modes
	 
    // Propagation covariance
	fMatrix P;
	float P_data[NUMBER_OF_STATES*NUMBER_OF_STATES];
	// System process noise
	fMatrix Q;
	float Q_data[NUMBER_OF_STATES*NUMBER_OF_STATES];	
	// Input process noise
	fMatrix Q_input;
	float Q_input_data[NUMBER_OF_INPUTS*NUMBER_OF_INPUTS];
} StateData;

extern RawSensorData g_sensor_raw;
extern ProcessedSensorData g_sensor;
extern StateData g_states;

extern uint8_t gEKF_mode;

// Function declarations
void EKF_Init( StateData* estimated_states );
void EKF_InitFromSensors( StateData* estimated_states, ProcessedSensorData* sensor_data );
void EKF_EstimateStates( StateData* estimated_states, ProcessedSensorData* sensor_data );
void EKF_Predict( StateData* estimated_states, ProcessedSensorData* sensor_data );
void EKF_Update( StateData* estimated_states, ProcessedSensorData* sensor_data );
void EKF_Correction( fMatrix* C, float sensor_data, float sensor_hat, float sensor_covariance, StateData* estimated_states, int sensor_type );

void process_sensor_data( RawSensorData* raw_data, ProcessedSensorData* processed_data );
void process_accel_data( RawSensorData* raw_data, ProcessedSensorData* processed_data );
void process_gyro_data( RawSensorData* raw_data, ProcessedSensorData* processed_data );
void process_mag_data( RawSensorData* raw_data, ProcessedSensorData* processed_data );

float lpf_sp_bt( float tau, float measured_value, float last_measured_value, float last_output, float deltaT );
static inline float square_f(float v) { return (v*v); }
static inline float cube_f(float v) { return (v*v*v); }

#define		MAG_UPDATE			0
#define		ACCEL_UPDATE		1

// EKF "mode" stored in the global variable gEKF_mode
#define		EKF_MODE_QUAT		0
#define		EKF_MODE_EULER		1


void compute_euler_angles( StateData* estimated_states );

void unroll_states( StateData* states );

#endif
