/* ------------------------------------------------------------------------------
  File: main.c
  Author: CH Robotics
  Version: 1.0
  
  Description: Entry point for UM6 firmware
------------------------------------------------------------------------------ */ 

// Includes 
#include <math.h>
#include <string.h>
#include "stm32f10x.h"
#include "util.h"
#include "UM6_startup.h"
#include "UM6_config.h"
#include "UM6_states.h"
#include "UM6_usart.h"
#include "CHR_packet_handler.h"
#include "CHR_i2c.h"
#include "CHR_matrix.h"
#include "CHR_delays.h"
#include "CHR_spi.h"
#include "CHR_GPS.h"
#include "EventScheduler.h"
#include "MPU9150.h"

extern uint8_t gClockInUse;

// Flags to indicate when specific sensors have new data ready
__IO uint8_t gGyroReady;
__IO uint8_t gAccelReady;
__IO uint8_t gMagStartMeasurement;
__IO uint8_t gMagReady;


__IO uint8_t gSendStateData;

__IO uint8_t gMagStatesReceived;
__IO uint8_t gAccelStatesReceived;
__IO uint8_t gGyroStatesReceived;

void updateGPS_Position( GPS_Struct *GPS_data );
void copy_GPS_packet_to_raw( GPS_Struct *GPS_data );

/*******************************************************************************
* Function Name  : main
* Input          : None
* Output         : None
* Return         : None
* Description    : Entry point for CHR-6dm firmware
*******************************************************************************/
int main(void)
{
	uint32_t i;
	uint32_t j;
	int32_t new_data;
	int32_t temp = 0;
	int32_t returnval;

	uint8_t L3G_status;
	uint8_t HMC_status;
	uint8_t LSM_status;

	GPS_Struct GPS_data;

	int32_t states_initialized = 0;

	g_PPS_received = 0;

	GPS_data.GGA_packet_received = 0;
	GPS_data.GLL_packet_received = 0;
	GPS_data.GSA_packet_received = 0;
	GPS_data.GSV_packet_received = 0;
	GPS_data.RMC_packet_received = 0;
	GPS_data.VTG_packet_received = 0;

	g_sensor_raw.new_accel_data = 0;
	g_sensor_raw.new_mag_data = 0;
	g_sensor_raw.new_GPS_position = 0;
	g_sensor_raw.new_GPS_course_speed = 0;
	g_sensor_raw.new_GPS_satellite_data = 0;
	g_sensor_raw.new_GPS_satellite_summary = 0;
	g_sensor_raw.GPS_last_received_time = 0.0f;
	
	g_save_flash_config = 0;
	g_save_factory_config = 0;

	gGyrosCalibrated = 0;
	g_zeroing_gyros = 0;
	gGPSPacketReceived = 0;

	// Initialize the IMU clocks, ADC, DMA controller, GPIO pins, etc.
	Initialize_UM6();

	// Clear global data array
	ClearGlobalData();

	// Fill g_config structure from flash, or use default values if flash has not been initialized.
	GetConfiguration();
	
	// Configure TX2 pin depending on whether it will be used as a PPS input or an output
	configure_tx2_pin();
	
	// Initialize structure used for calibrating system time using GPS PPS pulses
	init_gps_time_struct();

	// Based on configuration settings, set serial baud rate
	UpdateSerialBaud();

	g_data.r[DREG_HEALTH-DATA_REG_START_ADDRESS] = 0;

	// Set i2c SDA and SCl lines
	i2cReadSDA();
	i2cReadSCL();

	// Wait a while for supply voltages to stabilize
	DelayMs( 50 );

	// Initialize the MPU9150
	MPU9150_init();
	
	// Start "zero gyros" command.  When finished, the global flag gGyrosCalibrated will be set, and
	// the main execution code will know to start estimating states and transmitting data.
	if( g_config.r[CREG_MISC_SETTINGS] & ZERO_GYROS_ON_STARTUP )
	{
		start_gyro_zeroing();
	}
	else
	{
		gGyrosCalibrated = 1;		  
	}
	
	// Initialize the event scheduler
	init_event_scheduler();
	
	// Add events for regular packet transmission
	setup_com_events( );

	// Start the event scheduler
	start_event_scheduler();

	// Start timer4 for gyro sample rate control
	// (As of 5/26/14, this sampled the rate gyros at 500 Hz)
	TIM_SetCounter(TIM4,0);
	TIM_Cmd(TIM4, ENABLE);

	// Main program loop
	while(1)
	{
		// Get the current system time
		float current_time = get_system_time_uncal();
	
		// Handle event scheduler events.  This mostly handles tranmission of COM packets, but can also handle any periodic or non-periodic
		// timed event (i.e. flashing an LED, checking to make sure we are getting GPS data regularly, etc.)
		run_event_handlers();
	
		// If new data has arrived, use it to make new state estimates
		if( g_sensor_raw.new_gyro_data || g_sensor_raw.new_accel_data || g_sensor_raw.new_mag_data )
		{
			// Convert raw data to actual
			process_sensor_data( &g_sensor_raw, &g_sensor );
		
			// Run the EKF
			EKF_EstimateStates( &g_states, &g_sensor );				
		}
	
		// Check to see if we've received any GPS packets
		if( gGPSPacketReceived )
		{
			parseGPSPacket( g_GPS_packet, &GPS_data );
			copy_GPS_packet_to_raw( &GPS_data );
		
			gGPSPacketReceived = 0;
		
			// If we have new GPS course/speed or new GPS position
			if( g_sensor_raw.new_GPS_course_speed || g_sensor_raw.new_GPS_position )
			{
				// Handle time calibration
				if (g_PPS_received && GPS_data.mode == GPS_3D_FIX_MODE)
				{
					// Make sure it has been less than 1 sec since the PPS pulse arrived
					// Assumes pulse is synchronized to the nearest second
					if (unwrap_delta_T(get_system_time_uncal() - g_UTC_timing.t_sys_k) < 1.0f)
					{
						set_UTC_time_at_PPS(GPS_data.UTC_time);
						compute_time_calibration();
						g_data.r[DREG_HEALTH-DATA_REG_START_ADDRESS] &= ~HEALTH_GPS;
					}
					g_PPS_received = 0;
				}
				
				// Transmit GPS packets if needed
				if( g_config.r[CREG_COM_SETTINGS] & COM_GPS_DATA_ENABLED )
				{
					SendGlobalData(DREG_GPS_LATITUDE, ADDRESS_TYPE_DATA, 1, 6);
				}
			
				// Reset GPS flags
				g_sensor_raw.new_GPS_course_speed = 0;
				g_sensor_raw.new_GPS_position = 0;
			
				// Record the time we received this GPS data.  We'll check this periodically to make sure that we are still getting data.
				g_sensor_raw.GPS_last_received_time = current_time;
				
				// Make sure that the GPS fault flag is clear, since we've received GPS data
				g_data.r[DREG_HEALTH-DATA_REG_START_ADDRESS] &= ~HEALTH_GPS;
			
				// Copy new data to output
				copy_data_to_output();
			}
		}
	
		// Check to see if it has been too long since last receiving GPS data
		// If so, set the GPS fault flag in the health register
		if( unwrap_delta_T(current_time - g_sensor_raw.GPS_last_received_time) >= 2.0f )
		{
			g_data.r[DREG_HEALTH-DATA_REG_START_ADDRESS] |= HEALTH_GPS;
		}

		// If any of the sensors have new data, retrieve it
		if( gGyroReady )
		{
			gGyroReady = 0;
		
			float data_time = get_UTC_time();
			if( MPU9150_get_accel_gyro_data() )
			{
				gGyroStatesReceived = 1;

				g_sensor_raw.accel_y = (g_i2cRxBuf[1] << 8) | (g_i2cRxBuf[2] & 0x0FF);
				g_sensor_raw.accel_x = (g_i2cRxBuf[3] << 8) | (g_i2cRxBuf[4] & 0x0FF);
				g_sensor_raw.accel_z = -((g_i2cRxBuf[5] << 8) | (g_i2cRxBuf[6] & 0x0FF));

				g_sensor_raw.temperature = (float)((int16_t)(g_i2cRxBuf[7] << 8) | (int16_t)(g_i2cRxBuf[8] & 0x0FF));
				g_sensor_raw.temperature = g_sensor_raw.temperature/340.0f + 35.0f;
				g_sensor.temperature = g_sensor_raw.temperature;
			
				g_sensor_raw.temperature_time = data_time;

				g_sensor_raw.gyro_y = (g_i2cRxBuf[9] << 8) | (g_i2cRxBuf[10] & 0x0FF);
				g_sensor_raw.gyro_x = (g_i2cRxBuf[11] << 8) | (g_i2cRxBuf[12] & 0x0FF);
				g_sensor_raw.gyro_z = -((g_i2cRxBuf[13] << 8) | (g_i2cRxBuf[14] & 0x0FF));

				g_sensor_raw.new_gyro_data = 1;
				g_sensor_raw.new_accel_data = 1;
			
				g_sensor_raw.gyro_time = data_time;
				g_sensor_raw.accel_time = data_time;

				g_data.r[DREG_HEALTH-DATA_REG_START_ADDRESS] &= ~HEALTH_GYRO;
				g_data.r[DREG_HEALTH-DATA_REG_START_ADDRESS] &= ~HEALTH_ACCEL;
			}
			else
			{
				g_data.r[DREG_HEALTH-DATA_REG_START_ADDRESS] |= HEALTH_GYRO;
				g_data.r[DREG_HEALTH-DATA_REG_START_ADDRESS] |= HEALTH_ACCEL;
			}
		}
      
		// See if it is time to trigger a magnetometer measurement
		if( gMagStartMeasurement )
		{
			gMagStartMeasurement = 0;
		
			MPU9150_start_mag_conversion( );
			g_sensor_raw.mag_time = get_UTC_time();
		}

		// If the magnetometer is ready with new data, read it.
		if( gMagReady )
		{
			gMagReady = 0;
		
			if( MPU9150_get_mag_data( ) )
			{
				
				gMagStatesReceived = 1;
				g_sensor_raw.new_mag_data = 1;
			
				g_sensor_raw.mag_x = ((g_i2cRxBuf[1] << 8) | g_i2cRxBuf[0]);
				g_sensor_raw.mag_y = ((g_i2cRxBuf[3] << 8) | g_i2cRxBuf[2]);
				g_sensor_raw.mag_z = ((g_i2cRxBuf[5] << 8) | g_i2cRxBuf[4]);
				
				g_data.r[DREG_HEALTH-DATA_REG_START_ADDRESS] &= ~HEALTH_MAG;
			}
			else
			{
				g_data.r[DREG_HEALTH-DATA_REG_START_ADDRESS] |= HEALTH_MAG;
			}				
		}

		// Handle any new characters received by the UART (piped into memory via DMA)
		HandleUSART1Reception();
		HandleUSART2Reception();
	  
		// Check if a packet has been received over the UART.  If so, respond.
		if( gRXPacketReceived )
		{
			ProcessPacket( (USARTPacket*)&gRXPacketBuffer[gRXPacketBufferStart] );

			// Increment RX packet buffer pointer.  The RX and TX buffer is circular.  The buffer is
			// empty if the "start" and "end" pointers point to the same location.
			gRXPacketBufferStart++;
			if( gRXPacketBufferStart >= RX_PACKET_BUFFER_SIZE )
			{
				gRXPacketBufferStart = 0;
			}
				
			// If there are no more packets ready for processing, clear the gRXPacketReceived flag.
			// If there is another packet ready, leave the flag set so that the other RX packet will
			// be processed the next time through the main loop.
			if( gRXPacketBufferStart == gRXPacketBufferEnd )
			{
				gRXPacketReceived = 0;
			}
		}
		  
		// If there is a packet waiting to be transmitted and the USART is inactive, then start transmission
		if( gTXPacketBufferStart != gTXPacketBufferEnd)
		{
			if( TXPacketBufferReady )
			{
				TXPacketBufferReady = 0;
				SendNextPacket( );
				TXPacketBufferReady = 1;
			}						  
		}
	
	// ----- END OF MAIN LOOP ------
	}
	 
}

void updateGPS_Position( GPS_Struct *GPS_data )
{
	// If there is no GPS home position (ie. lat = 0 and lon = 0), set the new measured position to home
	if( g_states.GPS_lat_home == 0 )
	{
		g_states.GPS_lat_home = GPS_data->latitude;
		g_states.GPS_lon_home = GPS_data->longitude;
		g_states.GPS_alt_home = GPS_data->altitude;
	}

	// Compute position in meters relative to home lat/lon/alt.  This works for small distances (< 100 miles or so)
	g_sensor.GPS_position_up = g_sensor_raw.GPS_altitude - g_states.GPS_alt_home;
	float lon_scale = (3.14159f/180.0f)*6367499.0f*cos(GPS_data->latitude*3.14159f/180.0f);	
	g_sensor.GPS_position_east = lon_scale*(GPS_data->longitude - g_states.GPS_lon_home);			// Longitudinal distance per degree depends on latitude
	g_sensor.GPS_position_north = 110600.0f*(GPS_data->latitude - g_states.GPS_lat_home);			// Latitudinal distance per degree is constant
}

/*******************************************************************************
* Function Name  : copy_GPS_packet_to_raw
* Input          : GPS_Struct *GPS_data
* Output         : None
* Return         : None
* Description    : 

Takes the GPS data in GPS_data and copies it to the raw sensor data structure

*******************************************************************************/
void copy_GPS_packet_to_raw( GPS_Struct *GPS_data )
{
	// Retrieve GPS time - this will always be a part of the packet.
	g_sensor_raw.GPS_time = GPS_data->UTC_time;

	// GGA Packet
	if( GPS_data->GGA_packet_received )
	{
		GPS_data->GGA_packet_received = 0;

		g_sensor_raw.GPS_longitude = GPS_data->longitude;
		g_sensor_raw.GPS_latitude = GPS_data->latitude;
		g_sensor_raw.GPS_altitude = GPS_data->altitude;
		g_sensor_raw.GPS_Hdop = GPS_data->Hdop;
		g_sensor_raw.GPS_satellites_used = GPS_data->satellites_used;
		g_sensor_raw.GPS_satellites_in_view = GPS_data->satellites_in_view;

		updateGPS_Position(GPS_data);
		g_sensor_raw.new_GPS_position = 1;
		g_sensor_raw.new_GPS_satellite_summary = 1;
	}

	// GLL Packet
	if( GPS_data->GLL_packet_received )
	{
		GPS_data->GLL_packet_received = 0;

		g_sensor_raw.GPS_longitude = GPS_data->longitude;
		g_sensor_raw.GPS_latitude = GPS_data->latitude;

		updateGPS_Position(GPS_data);
		g_sensor_raw.new_GPS_position = 1;
	}

	// GSA Packet
	if( GPS_data->GSA_packet_received )
	{
		GPS_data->GSA_packet_received = 0;

		g_sensor_raw.GPS_Hdop = GPS_data->Hdop;
		g_sensor_raw.GPS_Vdop = GPS_data->Vdop;
		g_sensor_raw.GPS_mode = GPS_data->mode;

		g_sensor_raw.new_GPS_satellite_summary = 1;
	}

	// GSV Packet
	if( GPS_data->GSV_packet_received )
	{
		GPS_data->GSV_packet_received = 0;

		// Clear out empty array entries
		int loop_index;
		for( loop_index = GPS_data->satellites_in_view; loop_index < 12; loop_index++ )
		{
			GPS_data->satellite_data[loop_index][0] = 0;
			GPS_data->satellite_data[loop_index][1] = 0;
			GPS_data->satellite_data[loop_index][2] = 0;
			GPS_data->satellite_data[loop_index][3] = 0;
		}

		g_sensor_raw.GPS_SatID_1 = GPS_data->satellite_data[0][0];
		g_sensor_raw.GPS_SatID_2 = GPS_data->satellite_data[1][0];
		g_sensor_raw.GPS_SatID_3 = GPS_data->satellite_data[2][0];
		g_sensor_raw.GPS_SatID_4 = GPS_data->satellite_data[3][0];	
		g_sensor_raw.GPS_SatID_5 = GPS_data->satellite_data[4][0];
		g_sensor_raw.GPS_SatID_6 = GPS_data->satellite_data[5][0];
		g_sensor_raw.GPS_SatID_7 = GPS_data->satellite_data[6][0];
		g_sensor_raw.GPS_SatID_8 = GPS_data->satellite_data[7][0];
		g_sensor_raw.GPS_SatID_9 = GPS_data->satellite_data[8][0];
		g_sensor_raw.GPS_SatID_10 = GPS_data->satellite_data[9][0];
		g_sensor_raw.GPS_SatID_11 = GPS_data->satellite_data[10][0];
		g_sensor_raw.GPS_SatID_12 = GPS_data->satellite_data[11][0];

		g_sensor_raw.GPS_SatSNR_1 = GPS_data->satellite_data[0][3];
		g_sensor_raw.GPS_SatSNR_2 = GPS_data->satellite_data[1][3];
		g_sensor_raw.GPS_SatSNR_3 = GPS_data->satellite_data[2][3];
		g_sensor_raw.GPS_SatSNR_4 = GPS_data->satellite_data[3][3];
		g_sensor_raw.GPS_SatSNR_5 = GPS_data->satellite_data[4][3];
		g_sensor_raw.GPS_SatSNR_6 = GPS_data->satellite_data[5][3];
		g_sensor_raw.GPS_SatSNR_7 = GPS_data->satellite_data[6][3];
		g_sensor_raw.GPS_SatSNR_8 = GPS_data->satellite_data[7][3];
		g_sensor_raw.GPS_SatSNR_9 = GPS_data->satellite_data[8][3];
		g_sensor_raw.GPS_SatSNR_10 = GPS_data->satellite_data[9][3];
		g_sensor_raw.GPS_SatSNR_11 = GPS_data->satellite_data[10][3];
		g_sensor_raw.GPS_SatSNR_12 = GPS_data->satellite_data[11][3];

		g_sensor_raw.new_GPS_satellite_data = 1;
	}

	// RMC Packet
	if( GPS_data->RMC_packet_received )
	{
		GPS_data->RMC_packet_received = 0;

		g_sensor_raw.GPS_speed = GPS_data->speed;
		g_sensor_raw.GPS_course = GPS_data->course;

		g_sensor_raw.GPS_longitude = GPS_data->longitude;
		g_sensor_raw.GPS_latitude = GPS_data->latitude;

		updateGPS_Position(GPS_data);
		g_sensor_raw.new_GPS_position = 1;
		g_sensor_raw.new_GPS_course_speed = 1;				
	}

	// VTG Packet
	if( GPS_data->VTG_packet_received )
	{
		GPS_data->VTG_packet_received = 0;

		g_sensor_raw.GPS_speed = GPS_data->speed;
		g_sensor_raw.GPS_course = GPS_data->course;

		g_sensor_raw.new_GPS_course_speed = 1;
	}
}

/*******************************************************************************
* Function Name  : TIM4_IRQHandler
* Input          : None
* Output         : None
* Return         : None
* Description    : 

Called when Timer4 overflow occurs.  

This is used to control the rate gyro sample rate.

*******************************************************************************/
void TIM4_IRQHandler( void )
{
	static int flag = 0;
	static int iterations = 0;

	if( TIM_GetITStatus(TIM4, TIM_IT_Update ) != RESET )
	{		  
		gGyroReady = 1;

		// Control magnetometer sampling here
		// Gyros and accels are sampling at 500 Hz.  We want to sample the mag. at 25 Hz.
		// To sample the mag, we must first trigger a measurement on the IC.  Then, later, we need to 
		// read the measured data.
		if( iterations == 1 )
		{
			gMagStartMeasurement = 1;
		}
		else if( iterations == 10 )
		{
			gMagReady = 1;
		}
		else if( iterations >= 20 )
		{
			iterations = 0;
		}
		iterations++;

		// Clear pending interrupt bit
		TIM_ClearITPendingBit(TIM4, TIM_IT_Update);
	}
}
