/* ------------------------------------------------------------------------------
  File: util.c
  Author: CH Robotics
  Version: 1.0
  
  Description: Misc. utilities
------------------------------------------------------------------------------ */ 

#include <math.h>
#include "stm32f10x.h"

#include "util.h"

// Define macro to make sure that we don't overrun any buffers
#define BUF_INDEX_CHECK()		if(buf_index>=max_length) return max_length

/*******************************************************************************
* Function Name  : print_float 
* Input          : float value, uint8_t decimal_places, uint8_t max_length
* Output         : uint8_t* buffer
* Return         : Number of bytes written to 'buffer'
* Description    : 

Takes the float given in 'value' and prints it as a human-readable ASCII
string into the buffer pointed to by 'buffer'.

The input 'max_length' should specify the size of the buffer to prevent
buffer overflow.

The input 'decimal_places' specifies how many decimal places to print.

IMPORTANT NOTE: This function call assumes that the whole number part of the
float value (the part before the decimal) can be represented by an unsigned 32-bit integer.
Floating point values don't actually have this limitation, so there are cases
where this function will fail. Specifically, if 
abs(value) > 2^32, or 4,294,967,296

The maximum number of decimal places allowed for printing using this function
is also limited to 9.
						 
*******************************************************************************/
uint32_t print_float( float value, uint8_t decimal_places, char* buffer, uint8_t max_length )
{
	uint8_t buf_index = 0;

	// Limit the number of printable decimal places
	if( decimal_places > 9 )
	{
		decimal_places = 9;
	}

	// Is this a negative number?
	if( value < 0 )
	{
		value = -value;
		buffer[buf_index++] = '-';
	}

	BUF_INDEX_CHECK();

	// Convert the float to a uint32_t to get rid of the fractional part, and print the
	// value as an unsigned integer.
	uint32_t uint_value = (uint32_t)value;
	uint32_t bytes_printed = print_unsigned( uint_value, 0, &buffer[buf_index], max_length-buf_index );

	buf_index += bytes_printed;

	BUF_INDEX_CHECK();

	// If we don't need to add a decimal point, we are finished.  Return.
	if( decimal_places == 0 )
	{
		return buf_index;
	}

	// Print a decimal place
	buffer[buf_index++] = '.';
	BUF_INDEX_CHECK();

	// Get the order of the digits
	uint32_t magnitude = 1;
	int index;
	for( index = 0; index < decimal_places; index++ )
	{
		magnitude = magnitude*10;
	}
	float value_mod = value*magnitude;

	// Now truncate any remaining digits by converting value_mod to an unsigned 32-bit integer and doing a % operation on the order of the digits
	uint_value = ((uint32_t)value_mod)%magnitude;

	bytes_printed = print_unsigned( uint_value, decimal_places, &buffer[buf_index], max_length-buf_index );

	buf_index += bytes_printed;

	// We are finished.  Return.
	return buf_index;
}

/*******************************************************************************
* Function Name  : print_unsigned 
* Input          : uint32_t value, uint32_t length_override, uint8_t max_length
* Output         : uint8_t* buffer
* Return         : Number of bytes written to 'buffer'
* Description    : 

Takes the unsigned value given in 'value' and prints it as a human-readable ASCII
string into the buffer pointed to by 'buffer'.

The input 'max_length' should specify the size of the buffer to prevent
buffer overflow.

The input 'length_override' can be used to guarantee a minimum print size.
If length_override is greater than the number of digits needed to print the
number, then zeros are prepended to the string.  If 'length_override' is less
than the number of digits needed to print the number, then it is ignored.
						 
*******************************************************************************/
uint32_t print_unsigned( uint32_t value, uint32_t length_override, char* buffer, uint8_t max_length )
{
	uint8_t buf_index = 0;

	// Figure out the number of digits needed to show this number
	uint8_t digits = 1;
	uint32_t new_value = value/10;
	while( new_value > 0 )
	{
		digits++;
		new_value = new_value/10;
	}
	
	// Prepend as many zeros as needed
	while( length_override > digits )
	{
		length_override--;
		buffer[buf_index++] = '0';
		BUF_INDEX_CHECK();		
	}

	// Now start at the lowest-order digit and print it.
	// It is efficient to start at the lowest-order because we can do a simple
	// value%10 to get the remainder in the lowest digit, and then we can divide
	// by 10 to get the next digit.  The downside is that the lowest-order digit
	// is displayed last, so we have to write nearer the end of the buffer and move backwards.
	// If we would ever go beyond the bounds of the buffer, we just ignore the current digit
	// and move on.
	new_value = value;
	uint8_t digits_remaining = digits;
	while( digits_remaining > 0 )
	{
		uint8_t remainder = (uint8_t)(new_value%10);
		uint8_t new_index = buf_index + digits_remaining - 1;
	
		// Is this a good index?
		if( new_index < max_length )
		{
			buffer[new_index] = remainder + '0';
		}
	
		digits_remaining--;
		new_value = new_value/10;
	}

	// Everything printable is printed.
	if( buf_index + digits <= max_length )
	{
		return buf_index+digits;
	}
	
	return max_length;
}

/*******************************************************************************
* Function Name  : print_signed
* Input          : int32_t value, uint32_t length_override, uint8_t max_length
* Output         : uint8_t* buffer
* Return         : Number of bytes written to 'buffer'
* Description    : 

Takes the signed value given in 'value' and prints it as a human-readable ASCII
string into the buffer pointed to by 'buffer'.

The input 'max_length' should specify the size of the buffer to prevent
buffer overflow.

The input 'length_override' can be used to guarantee a minimum print size.
If length_override is greater than the number of digits needed to print the
number, then zeros are prepended to the string.  If 'length_override' is less
than the number of digits needed to print the number, then it is ignored.

If a negative sign is present, it is ignored when computing the length needed
in length_override.
						 
*******************************************************************************/
uint32_t print_signed( int32_t value, uint32_t length_override, char* buffer, uint8_t max_length )
{
	uint8_t buf_index = 0;

	// Add the negative sign if needed
	if( value < 0 )
	{
		value = -value;
		buffer[buf_index++] = '-';
	}

	BUF_INDEX_CHECK();

	// At this point, 'value' is unsigned (because we negated it above)
	// Print it like an unsigned number now
	return print_unsigned( (uint32_t)value, length_override, &buffer[buf_index], max_length-buf_index);
}

/*******************************************************************************
* Function Name  : print_hex
* Input          : unt32_t value, uint32_t length_override, uint8_t max_length
* Output         : uint8_t* buffer
* Return         : Number of bytes written to 'buffer'
* Description    : 

Takes the unsigned value given in 'value' and prints it as a human-readable ASCII hex
string into the buffer pointed to by 'buffer'.  This function does not
add '0x' to the beginning of the number.

The input 'max_length' should specify the size of the buffer to prevent
buffer overflow.

The input 'length_override' can be used to guarantee a minimum print size.
If length_override is greater than the number of digits needed to print the
number, then zeros are prepended to the string.  If 'length_override' is less
than the number of digits needed to print the number, then it is ignored.
						 
*******************************************************************************/
uint32_t print_hex( uint32_t value, uint8_t length_override, char* buffer, uint8_t max_length )
{
	uint8_t n_index;
	uint8_t buf_index = 0;

	uint8_t char_data[8];

	// We need to do this nibble by nibble (since one nibble is displayed per character)
	for( n_index = 0; n_index < 8; n_index++ )
	{
		uint8_t n = (value >> (4*(7-n_index))) & 0x0F;
	
		// If n < 10, then offset into the ASCII table to display the appropriate number
		if( n < 10 )
		{
			n += 48;
		}
		// If n >= 10, offset into the ASCII table to display the appropriate letter
		else
		{
			n += 55;
		}
	
		char_data[n_index] = n;
	}

	// The full 32-bit hex number has been written.  Now determine how many consecutive high-order zero digits there are
	uint8_t zero_digits = 0;
	for( n_index = 0; n_index < 8; n_index++ )
	{
		if( char_data[n_index] != '0' )
		{
			break;
		}
	
		zero_digits++;
	}

	uint8_t display_digits = 8 - zero_digits;

	// Make sure we display at least one digit
	if( display_digits < 1 )
	{
		display_digits = 1;
	}

	// If length_override is higher than the number of digits we want to show (display_digits), then fill in the appropriate number of zeros
	if( display_digits < length_override )
	{
		uint8_t zeros = length_override - display_digits;
	
		for( n_index = 0; n_index < zeros; n_index++ )
		{
			buffer[buf_index++] = '0';
			BUF_INDEX_CHECK();
		}
	}

	// Now copy out the non-zero low-order digits from the hex conversion
	for( n_index = 0; n_index < display_digits; n_index++ )
	{
		buffer[buf_index++] = char_data[8-display_digits+n_index];
		BUF_INDEX_CHECK();
	}

	return buf_index;
}



